package charite.christo.blast;
import charite.christo.*;
import java.io.*;
import java.util.*;
import static charite.christo.ChUtils.*;
import static charite.christo.ChConstants.*;

/**HELP

WU-BLAST can be downloaded from http://blast.wustl.edu/licensing/.
The WU-BLAST directory must be added to the PATH environment variable.
Further the environment variables  WUBLASTMAT and WUBLASTDB must be set.

<br>
<pre class="terminal">
export WUBLASTMAT=/local/bioinf/wu-blast/matrix
</pre>
<br>

WUBLASTDB points to the directory where the blast databases are
installed. After formating a blast-database it contains files with the
endings ".xpd", ".xps" and ".xpt".

<pre class="terminal">
export WUBLASTDB=/local/bioinf/db/wu-blast
</pre>
<br>
The WUBLASTDB is also evaluated by STRAP to determine the
available databases.
The databases are created from multiple fasta files using the command
<br>
<pre class="terminal">
xdformat -p  -C x multipleFastaFile
</pre>
<br>
Strap calls the  Blast-program with the method Runtime#exec(String[]).

A valid Blast command is for example:
<br>
 <pre class="terminal">
wu-blastall -p blastp -d uniref50.fasta -i ~/testBlast/t.fa -b 300 -m 7
</pre>

<i>SEE_CLASS:Blaster_local_NCBI</i>
@author Christoph Gille
*/
public class Blaster_local_Wu extends AbstractBlaster {
    private File fOut,fIn;
    private File dir,dirDatabases;
    private static int countInstance;
    private ChExec exec;
    public String getCommandLine() {return (isSystProprty(IS_LINUX) ? "nice " : "")+programName+" -p "+getBlastProgram()+" -d "+getDatabase()+" -i "+fIn+" -o "+fOut +" "+additionalOptions;}
    @Override public ByteArray computeIt() {
        //putln(" computeIt ");
        final String database=getDatabase();
        if (database==null) return null;
        mkdrsErr(dir=file(dirTmp()+"/"+programName+"/"+countInstance++));
        delFile((fOut=file(dir,"output.xml")));
        wrte(fIn=file(dir,"input.fa"),">query\n"+getQuerySequence()+"\n");
        delFileOnExit(fIn);
        delFileOnExit(fOut);
        delFileOnExit(dir);
        exec=new ChExec(ChExec.STDOUT | ChExec.STDERR).setCommandLine(getCommandLine()).dir(dir);
        exec.run();
        if (exec.exitValue()!=0) putln(exec.getStdout());

        return readAndDecompress(fOut,null);

    }
    @Override public String[] getAvailableMatrices() {
        final String BLASTMAT=ChEnvironment.get(shellVariableBLASTMAT);
        File dirMat=file(BLASTMAT+ (isAAQuery() ? "/aa" : "/nt"));
        log("The environment variable ",shellVariableBLASTMAT," has the value",BLASTMAT,"\n");
        if (dirMat==null||!dirMat.isDirectory()) {
            log(RED_ERROR," the environment variable "+shellVariableBLASTMAT+" does not point to a directory\n");
            return NO_STRING;
        }
        return dirMat.list();
    }
    @Override public String[] getAvailableDatabases(boolean isAmino, boolean withoutDelay) {
        final String BLASTDB=ChEnvironment.get(shellVariableBLASTDB);
        dirDatabases=file(BLASTDB);
        final String error=
            BLASTDB==null ? "Shell variable $"+shellVariableBLASTDB+" is not set" :
            dirDatabases==null || !dirDatabases.isDirectory() ? " No such directory:  $"+shellVariableBLASTDB+"="+BLASTDB : null;
        if (error!=null) {
            logStdout(RED_ERROR,error,"\n");
            return NO_STRING;
        }
        log("The environment variable ",shellVariableBLASTDB," has the value",BLASTDB,"\n");
        final List v=new ArrayList();
        for(String s:dirDatabases.list())
            if (s.endsWith(extensionOfDatabaseFiles))
                v.add(delSfx(extensionOfDatabaseFiles,s));
        log("Databases found in "+BLASTDB+": "+v+"\n");
        final String dd[]=strgArry(v);
        if (dd.length==0) { putln("Error: No database files with ending ",extensionOfDatabaseFiles); putln("found in ", dirDatabases); }
        return dd;
    }
    /**
       The variable ..BLASTDB points to the directory where databases are installed.
       STRAP reads this directory to provide a List of available databases.
    */
    public void setShellVariableBLASTDB(String db, String extensionOfDatabaseFiles) { shellVariableBLASTDB=db; this.extensionOfDatabaseFiles=extensionOfDatabaseFiles;}
    //    public String shellVariableBLASTMAT() { return "WUBLASTMAT";}
    private String additionalOptions, shellVariableBLASTDB="WUBLASTDB", programName="wu-blastall", extensionOfDatabaseFiles=".xpt";
    private final String  shellVariableBLASTMAT="WUBLASTMAT";
    public void setProgramName(String prg) { programName=prg;}
    public void setAdditionalOptions(String options) { additionalOptions=options;}
    public Blaster_local_Wu() {
        additionalOptions=" -b "+getNumberOfAlignments()+" -m 7  -W 2 -e 100 -Ff";
    }

}
