package charite.christo.protein;
import charite.christo.ByteArray;
import charite.christo.ChTokenizer;
import charite.christo.interfaces.HasModiCount;
import static charite.christo.ChUtils.*;
/**
   A matrix defines an affine transformation in 3D space.
   @author Christoph Gille
*/
public final class Matrix3D implements HasModiCount {
    public final static Matrix3D[] NONE={};
    private double scale=1,  xx=1, xy, xz, xo,   yx, yy=1, yz, yo,   zx, zy, zz=1, zo;
    private int biomolecule;

    /** Create a new unit matrix */
    public Matrix3D() {}
    public Matrix3D(ByteArray ba){parse(ba);}
    public Matrix3D(
                    double xx,double xy, double xz, double xo,
                    double yx,double yy, double yz, double yo,
                    double zx,double zy, double zz, double zo) {
        this.xx=xx;this.xy=xy;this.xz=xz; this.xo=xo;
        this.yx=yx;this.yy=yy;this.yz=yz; this.yo=yo;
        this.zx=zx;this.zy=zy;this.zz=zz; this.zo=zo;
    }
    public Matrix3D(Matrix3D m) {
        this(m.xx, m.xy, m.xz, m.xo,
             m.yx, m.yy, m.yz, m.yo,
             m.zx, m.zy, m.zz, m.zo);
    }
    /* ---------------------------------------- */
    /* >>> Get  >>> */
    public double[][] getRotation() { return new double[][] { {xx, xy, xz}, {yx, yy, yz}, {zx, zy, zz} };}
    public double[] getTranslation() { return new double[]{xo,yo,zo}; }
    /* <<< Get  <<< */
    /* ---------------------------------------- */
    /* >>> Biomolecule >>> */
    private String name,chains;
    Matrix3D setMoleculeAndChainsAndName(int m, String chains, String name) {
        this.biomolecule=m; this.chains=chains; this.name=name;
        modi++;
        return this;
    }
    public String getName() { return name;}
    public int getBiomolecule() { return biomolecule;}
    String getChains() { return chains;}
    /* <<< Biomolecule <<< */
    /* ---------------------------------------- */
    /* >>> Set >>> */
    /** Scale by f in all dimensions */
    public synchronized Matrix3D scale(double f) {
        scale*=f;
        xx*=f;xy*=f;xz*=f;xo*=f;
        yx*=f;yy*=f;yz*=f;yo*=f;
        zx*=f;zy*=f;zz*=f;zo*=f;
        modi++;
        return this;
    }
    /** rotate theta degrees about the x,y,z axis */
    public synchronized Matrix3D rotate(char axis, double theta) {
        final double ct=Math.cos(theta),  st=Math.sin(theta);
        double
            Nxx=xx, Nxy=xy ,Nxz=xz, Nxo=xo,
            Nyx=yx, Nyy=yy ,Nyz=yz, Nyo=yo,
            Nzx=zx, Nzy=zy ,Nzz=zz, Nzo=zo;
        if ('x'==axis) {
            Nyx=yx*ct + zx*st;
            Nyy=yy*ct + zy*st;
            Nyz=yz*ct + zz*st;
            Nyo=yo*ct + zo*st;
            Nzx=zx*ct - yx*st;
            Nzy=zy*ct - yy*st;
            Nzz=zz*ct - yz*st;
            Nzo=zo*ct - yo*st;
        } else if ('y'==axis) {
            Nxx=xx*ct + zx*st;
            Nxy=xy*ct + zy*st;
            Nxz=xz*ct + zz*st;
            Nxo=xo*ct + zo*st;
            Nzx=zx*ct - xx*st;
            Nzy=zy*ct - xy*st;
            Nzz=zz*ct - xz*st;
            Nzo=zo*ct - xo*st;
        } else if ('z'==axis) {
            Nyx=yx*ct + xx*st;
            Nyy=yy*ct + xy*st;
            Nyz=yz*ct + xz*st;
            Nyo=yo*ct + xo*st;
            Nxx=xx*ct - yx*st;
            Nxy=xy*ct - yy*st;
            Nxz=xz*ct - yz*st;
            Nxo=xo*ct - yo*st;
        }
        this.xx=Nxx; this.xy=Nxy; this.xz=Nxz;  this.xo=Nxo;
        this.yx=Nyx; this.yy=Nyy; this.yz=Nyz;  this.yo=Nyo;
        this.zx=Nzx; this.zy=Nzy; this.zz=Nzz;  this.zo=Nzo;
        modi++;
        return this;
    }
    /** Multiply this matrix by a second: M=M*R */
    public Matrix3D mult(Matrix3D rhs) {
        if (rhs==null) return this;
        final double
            lxx=xx*rhs.xx + yx*rhs.xy + zx*rhs.xz,
            lxy=xy*rhs.xx + yy*rhs.xy + zy*rhs.xz,
            lxz=xz*rhs.xx + yz*rhs.xy + zz*rhs.xz,
            lxo=xo*rhs.xx + yo*rhs.xy + zo*rhs.xz + rhs.xo,
            lyx=xx*rhs.yx + yx*rhs.yy + zx*rhs.yz,
            lyy=xy*rhs.yx + yy*rhs.yy + zy*rhs.yz,
            lyz=xz*rhs.yx + yz*rhs.yy + zz*rhs.yz,
            lyo=xo*rhs.yx + yo*rhs.yy + zo*rhs.yz + rhs.yo,
            lzx=xx*rhs.zx + yx*rhs.zy + zx*rhs.zz,
            lzy=xy*rhs.zx + yy*rhs.zy + zy*rhs.zz,
            lzz=xz*rhs.zx + yz*rhs.zy + zz*rhs.zz,
            lzo=xo*rhs.zx + yo*rhs.zy + zo*rhs.zz + rhs.zo;
        xx=lxx; xy=lxy; xz=lxz; xo=lxo;
        yx=lyx; yy=lyy; yz=lyz; yo=lyo;
        zx=lzx; zy=lzy; zz=lzz; zo=lzo;
        return this;
    }
    public synchronized Matrix3D newInverse() {
        Matrix3D m=new Matrix3D();
        m.xo=- xx*xo-yx*yo-zx*zo;
        m.xx=xx; m.xy=yx; m.xz=zx;
        m.yo= -xy*xo-yy*yo-zy*zo;
        m.yx=xy;m.yy=yy;m.yz=zy;
        m.zo= -xz*xo-yz*yo-zz*zo;
        m.zx=xz;m.zy=yz;m.zz=zz;
        modi++;
        return m;
    }
    /** Set to unit matrix */
    public synchronized Matrix3D unit() { scale=xx=yy=zz=1;xo=yo=zo=xy=xz=yx=yz=zx=zy=0;  modi++; return this;  }
    public synchronized Matrix3D setTranslation(double tx, double ty, double tz) {
        xo=tx; yo=ty; zo=tz;
        modi++;
        return this;
    }
    public synchronized Matrix3D setTranslation(double t[]) { return setTranslation(t[0],t[1],t[2]);}
    public synchronized Matrix3D setRotation(double r[][]) {
        if (r!=null) {
            xx=r[0][0];xy=r[0][1];xz=r[0][2];
            yx=r[1][0];yy=r[1][1];yz=r[1][2];
            zx=r[2][0];zy=r[2][1];zz=r[2][2];
            modi++;
        }
        return this;
    }
    /* <<< Set <<< */
    /* ---------------------------------------- */
    /* >>> Parse >>> */
    public boolean parsePlain(byte text[], int begin, int end) {
        if (sze(text)==0) return false;
        final ChTokenizer TOK=new ChTokenizer();
        TOK.setText(text, begin, end);
        final double d[]=new double[12];
        for(int i=0;i<12;i++) {
            if (!TOK.nextToken()) return false;
            if (Double.isNaN(d[i]=TOK.asFloat())) return false;
        }
        xx=d[0];
        xy=d[1];
        xz=d[2];
        yx=d[3];
        yy=d[4];
        yz=d[5];
        zx=d[6];
        zy=d[7];
        zz=d[8];
        xo=d[9];
        yo=d[10];
        zo=d[11];
        modi++;
        return true;
    }
    public boolean parse(ByteArray ba) {
        final byte T[]=ba.getBytes();
        final int E=ba.getEnd(), ends[]=ba.getLineEnds();
        int iRot=-1, iTrans=-1;
        for(int iL=0; iL<ends.length; iL++) {
            final int b0= iL==0 ? ba.getBegin() : ends[iL-1]+1,  e=prev(NO_SPC,T,ends[iL]-1,b0-1)+1, b=nxtE(NO_SPC,T,b0, e);
            if (e-b>0 && T[b]=='-') {
                if (e-b>8  && iRot<0   && equls("-ROTATION",T,b)) iRot=e;
                if (e-b>11 && iTrans<0 && equls("-TRANSLATION",T,b)) iTrans=e;
            }
            if (e-b>0 && T[b]=='<') {
                if (e-b>9  && equls("<rotation>",T,b)) iRot=b+10;
                if (e-b>12 && equls("<translation>",T,b)) iTrans=b+13;
            }
        }
        ChTokenizer TOK=null;
        if (iRot>0) {
            if (TOK==null) TOK=new ChTokenizer();
            TOK.setText(T, iRot, E);
            TOK.nextToken(); xx=TOK.asFloat();
            TOK.nextToken(); xy=TOK.asFloat();
            TOK.nextToken(); xz=TOK.asFloat();
            TOK.nextToken(); yx=TOK.asFloat();
            TOK.nextToken(); yy=TOK.asFloat();
            TOK.nextToken(); yz=TOK.asFloat();
            TOK.nextToken(); zx=TOK.asFloat();
            TOK.nextToken(); zy=TOK.asFloat();
            TOK.nextToken(); zz=TOK.asFloat();
        }
        if (iTrans>0)  {
            if (TOK==null) TOK=new ChTokenizer();
            TOK.setText(T,iTrans,E);
            TOK.nextToken(); xo=TOK.asFloat();
            TOK.nextToken(); yo=TOK.asFloat();
            TOK.nextToken(); zo=TOK.asFloat();
        }
        modi++;
        return !isNaN() && iRot>=0 && iTrans>=0;
    }
    /* <<< Parse <<< */
    /* ---------------------------------------- */
    /* >>>  Analyze >>> */
    public boolean isUnit() {
        return
            .9999<xx && xx<1.0001  && .9999<yy && yy<1.0001 && .9999<zz && zz<1.0001  &&
            -.0001<xy&&xy<0.0001 &&
            -.0001<xz&&xz<0.0001 &&
            -.0001<yz&&yz<0.0001 &&
            -.0001<xo&&xo<0.0001 &&
            -.0001<yo&&yo<0.0001 &&
            -.0001<zo&&zo<0.0001;
    }
    public double getDeterminant(boolean normalize) {
        /*
          xx xy xz    xx xy
          yx yy yz    yx yy
          zx zy zz    zx zy
        */
        final double d= xx*yy*zz + xy*yz*zx +  xz*yx*zy  - zx*yy*xz-  zy*yz*xx -  zz*yx*xy,  s=scale;
        return normalize&&s!=0 ? d/(s*s*s) : d;
    }
    final boolean isNaN() {
        return
            Double.isNaN(xx) || Double.isNaN(xy) || Double.isNaN(xz) ||
            Double.isNaN(yx) || Double.isNaN(yy) || Double.isNaN(yz) ||
            Double.isNaN(zx) || Double.isNaN(zy) || Double.isNaN(zz) ||
            Double.isNaN(xo) || Double.isNaN(yo) || Double.isNaN(zo);
    }
    /* <<< Analyze <<< */
    /* ---------------------------------------- */
    /* >>> Apply >>> */
    /** Transform nvert points from v into tv.  v contains the input
        coordinates in floating point.  Three successive entries in
        the array constitute a point.  tv ends up holding the transformed
        points as integers; three successive entries per point */
    public synchronized final float[] transformPoints(float v[], float tv[],int nvert) {
        if (nvert>v.length/3) nvert=v.length/3;
        if (nvert>tv.length/3) nvert=tv.length/3;
        final float
            lxx=(float)xx, lxy=(float)xy, lxz=(float)xz, lxo=(float)xo,
            lyx=(float)yx, lyy=(float)yy, lyz=(float)yz, lyo=(float)yo,
            lzx=(float)zx, lzy=(float)zy, lzz=(float)zz, lzo=(float)zo;
        for(int i=nvert*3; (i -=3) >=0;) {
            final float x=v[i], y=v[i+1], z=v[i+2];
            tv[i  ]=x*lxx + y*lxy + z*lxz + lxo;
            tv[i+1]=x*lyx + y*lyy + z*lyz + lyo;
            tv[i+2]=x*lzx + y*lzy + z*lzz + lzo;
        }
        return tv;
    }
    /* <<< Apply <<< */
    /* ---------------------------------------- */
    /* >>> Distance >>> */
    private static double squareDist(float x1,float y1, float z1,   float x2,float y2, float z2 ) {
        final float dx=x2-x1, dy=y2-y1, dz=z2-z1;
        return dx*dx+dy*dy*dz*dz;
    }
    private static double squareDistOfPointAndLine(float xP,float yP, float zP, float x1,float y1, float z1,   float x2,float y2, float z2) {
        final double square12=squareDist(x1,y1,z1, x2,y2,z2), square1P=squareDist(x1,y1,z1, xP,yP,zP);
        if (square12<1E-10) return squareDist(xP,yP,zP, x1,x2,xP);
        final double skalarProdukt= ((x2-x1)*(xP-x1) + (y2-y1)*(yP-y1) + (z2-z1)*(zP-z1));
        final double square1F=skalarProdukt*skalarProdukt/square12;
        return square1P-square1F;
    }
    private static void mostDistantPoints2(float xyz[], int len, int[] result) {
        final int L=mini(xyz.length/3, len);
        double maxDist=Double.MAX_VALUE;
        for(int i=L; --i>=0;) {
            final int i3=3*i;
            for(int j=L; --j>i;) {
                final int j3=3*j;
                final double d=squareDist(xyz[i3],xyz[i3+1],xyz[i3+2], xyz[j3],xyz[j3+1],xyz[j3+2]);
                if (d>maxDist) {
                    maxDist=d;
                    result[0]=i;
                    result[0]=j;
                }
            }
        }
    }
    public static void mostDistantPoints3(float xyz[], int len, int[] result) {
        result[0]=result[1]=result[2]=-1;
        final int L=mini(xyz.length/3, len);
        mostDistantPoints2(xyz,L,result);
        double maxDist=Double.MAX_VALUE;
        final float x0,y0,z0, x1,y1,z1;
        {
            final int i0=result[0], i1=result[1];
            x0=xyz[i0]; y0=xyz[i0]; z0=xyz[i0];
            x1=xyz[i1]; y1=xyz[i1]; z1=xyz[i1];
        }
        for(int i=L; --i>=0;) {
            final int i3=3*i;
            final double d=squareDistOfPointAndLine(xyz[i3],xyz[i3+1],xyz[i3+2], x0,y0,z0, x1,y1,z1);
            if (d>maxDist) {
                maxDist=d;
                result[3]=i;
            }
        }
    }
    /* <<< Distance <<< */
    /* ---------------------------------------- */
    /* >>> equals >>> */
    private int modi=0;
    public int getModificationCount() { return modi;}
    @Override public int hashCode() { return (int)(999999*(xx+yy+zz+ xy+xz+ yz +xo+yo+zo));  }
    @Override public boolean equals(Object o) {
        if (o instanceof Matrix3D) {
            final Matrix3D m=(Matrix3D)o;
            return m!=null&& xx==m.xx&&xy==m.xy&&xz==m.xz&&xo==m.xo&&yx==m.yx&&yy==m.yy&&yz==m.yz&&yo==m.yo&&zx==m.zx&&zy==m.zy&&zz==m.zz&&zo==m.zo;
        } else return false;
    }

    /* <<< equals <<< */
    /* ---------------------------------------- */
    /* >>> Export >>> */
    @Override public String toString() {
        return "<rotation>\n"+
            xx+"\t"+xy+"\t"+xz+"\n"+
            yx+"\t"+yy+"\t"+yz+"\n"+
            zx+"\t"+zy+"\t"+zz+"\n</rotation>\n<translation>\n"+
            xo+"\t"+yo+"\t"+zo+"\n</translation>\n";
    }
    public final static int FORMAT_PLAIN=1, FORMAT_JMOL=2, FORMAT_PYMOL=3, FORMAT_EXPLAIN=4;
    public String toText(int format, String linePrefix) {
        if (format==FORMAT_JMOL) {
            return "r=[ "+
                " [ "+(float)xx+", "+(float)xy+", "+(float)xz+" ], "+
                " [ "+(float)yx+", "+(float)yy+", "+(float)yz+" ], "+
                " [ "+(float)zx+", "+(float)zy+", "+(float)zz+
                " ] ];\n t={ "+(float)xo+", "+(float) yo+", "+(float) zo+" };\n";

        }

        if (format==FORMAT_PYMOL) {
            return
                xx+", "+xy+", "+xz+", "+xo+",  "+
                yx+", "+yy+", "+yz+", "+yo+",  "+
                zx+", "+zy+", "+zz+", "+zo+
                "  , 0, 0, 0, 1";
        }
        if (format==FORMAT_EXPLAIN) {
            final ByteArray sb=byteArr(SB_TIP);
            final String pfx=linePrefix!=null ? linePrefix : "";
            sb.append(pfx).append("X' = X *").append(xx,4,3).append(" +  Y *").append(xy,4,3).append(" +  Z *").append(xz,4,3).append(" + ").append(xo,4,3).append('\n');
            sb.append(pfx).append("Y' = Y *").append(yx,4,3).append(" +  Y *").append(yy,4,3).append(" +  Z *").append(yz,4,3).append(" + ").append(yo,4,3).append('\n');
            sb.append(pfx).append("Z' = Z *").append(zx,4,3).append(" +  Y *").append(zy,4,3).append(" +  Z *").append(zz,4,3).append(" + ").append(zo,4,3).append('\n');
            return sb.newString();
        }
        return
            (float)xx+" "+(float)xy+" "+(float)xz+"  "+
            (float)yx+" "+(float)yy+" "+(float)yz+"  "+
            (float)zx+" "+(float)zy+" "+(float)zz+"  "+
            (float)xo+" "+ (float)yo+" "+ (float)zo;
    }

}
