package charite.christo.protein;
import charite.christo.*;
import charite.christo.interfaces.*;
import java.io.*;
import java.util.*;
import static charite.christo.ChUtils.*;
import static charite.christo.ChConstants.*;
import static charite.christo.protein.Protein3dUtils.*;

/**HELP

WIKI:PyMOL <i>HOME_PAGE:Pymol</i> is a widely applied protein
structure viewer by Warren Delano.

<br><br><b>Residue selections:</b> Dragging the mouse over a part of
the amino acid sequence in the STRAP alignment creates a
residue selection in Pymol.

The STRAP cursor position refers to another Pymol selection named "cursor".

Users can refer to these selections in Pymol to show <b>[S]</b> or hide
<b>[H]</b> representations of residues like <I>sphere</I>,
<I>cartoon</I> or <I>stick</I>.

For example to show all atoms of the amino acid at the current cursor position as spheres
open the corresponding [S] menu of the selection "cursor" and select "spheres".

    <div class="figure">
      <table style="caption-side: bottom">
        <CAPTION><b>Figure: </b>Side pane of Pymol containing a list of loaded proteins and selections.
          [H]Hide style, [S]Show style, [L]Add Label, [C]Change Color </CAPTION>
        <tr><td><img src="http://www.bioinformatics.org/strap/images/pymolSelections.jpg" /></td></tr>
      </table>
    </div>

<br><br><b>Removing a protein from Pymol:</b> If a protein
would be removed in Pymol this would be unnoticed by STRAP.
Therefore it is recommended to use the STRAP user interface and not the Pymol user interface to remove proteins.

There are two possibilities:
<ol>
<li>The STRAP menu item <i>ITEM:charite.christo.strap.ProteinPopup#ACTION_close3d</i></li>
<li>The context menu of pymol in STRAP which can be opened by right clicking the pymol node
in the tree representation. The latter
is located at the left of the STRAP application frame and  needs to be expanded by dragging.
</li>
</ol>

<div class="figure">
<table style="caption-side: bottom">
<CAPTION><b>Figure: </b>Automated installation on Windows, Linux and Mac</CAPTION>
<tr><td><i>JCOMPONENT:Pymol#canInstall(null)!</i></td></tr>
</table>
</div>

<br><br><b>Pymol commands:</b>

Pymol commands  (see http://pymol.sourceforge.net/newman/ref/S1000comref.html) are typed directly into the Pymol-window.

Most important commands:

<ul>

<li><b>ray</b><br> Creates a ray-traced image of the current frame.</li>

<li><b>png my_file.png</b><br> Writes an image file.</li>

<li><b>bg white</b><br>White background.</li>

</ul>

<br><br><b>Related links:</b> http://www.rubor.de/bioinf/ &nbsp;

http://www.rubor.de/anlagen/PyMOL_Tutorial.pdf

<br><br><b>Bugs:</b> 
Pymol does not treat negative pdb residue numbers correctly.
Pymol selection is unspecific if the  chain identifier is a space character ' ';

<i>PACKAGE:charite.christo.strap.</i>
<i>SEE_CLASS:ResidueAnnotation</i>
<i>SEE_CLASS:AddAnnotation</i>
@author Christoph Gille, Kristian Rother

 h+="If PyMOLWin.exe does not work replace it by PyMOL.exe.<br>\n"+
                    "If this does not work either comment out all lines with \"#\".<br>\n"+
                    "This will  force STRAP to install Pymol.<br>"+
                    "You can also try  PyMOL.exe instead of PyMOLWin.exe";
*/
public class Pymol implements ProteinViewer,ChRunnable,java.awt.event.ActionListener {
    public final static String
        DEFINE_THIS_COLOR="$DEFINE_THIS_COLOR",
        PYMOL_COLOR_NAME="$PYMOL_COLOR_NAME";

    /* <<<  <<< */
    /* ---------------------------------------- */
    /* >>> Shared fields >>> */
    private static ChExec shared_exec;
    private static UniqueList<Pymol> shared_vView;
    /* <<< Shared fields <<< */
    /* ---------------------------------------- */
    /* >>> Constructor  >>> */

    public Pymol() {
        ProteinUtils.initCustomizableSettings(this);
        if (shared_vView==null) shared_vView=new UniqueList(Pymol.class);
    }
    /* <<< Constructor <<< */
    /* ---------------------------------------- */
    /* >>> Launch >>> */
    private static Runnable runAfterAllLoaded;
    private static boolean isWaitingForInput;
    private boolean showStdoutAndStderr, isRunning;
    private void addPath(String fExe, ChExec ex) {
        if (ex==null) return;
        if (showStdoutAndStderr) ex.showStdoutAndStderr(null);

        if (eqNz(fExe, downloadedBinary)) {
            final Map<String,String> m=new HashMap();
            final String parent=withoutLastComponent(fExe, isWindows()?'\\':'/');
            if (isWindows()) m.put("PATH", ChEnvironment.get("PATH")+";"+parent);
            // for(String s : new String[]{"HOME","LD_LIBRARY_PATH"/*, "DISPLAY"*/ }) m.put(s,ChEnvironment.get(s));
            m.put("PYMOL_PATH",isMacintosh() ?  dirBin()+"/MacPyMOL.app/pymol" : parent);
            ex.addToEnvironement(m);
        }
    }
    public void showStdoutAndStderr(boolean yesNo) { showStdoutAndStderr=yesNo;}
    private boolean launch() {
        final boolean isWin=isWindows(), isMac=isMacintosh(), isLinux=isSystProprty(IS_LINUX386);
        ChExec ex=this.shared_exec;
        if (ex!=null) ex.dispose();
        this.shared_exec=ex=null;
        assertEDT();
        final boolean logPanel=getNextProcessHasLogPanel();
        if (_canInstall==null) _canInstall=new ChFrame("Pymol").ad(canInstall());
        _canInstall.shw(ChFrame.PACK|ChFrame.CENTER|ChFrame.ALWAYS_ON_TOP);
        final List v=new ArrayList();
        if (isWin||isLinux||isMac) v.add(downloadedBinary);
        adAllUniq(Customize.getCustomizableForClass(getClass(), Customize.CLASS_ExecPath).getSettings(),v);
        v.add("pymol");

        for(String cmd0 : toStrgArryO(v) ) {
            final String cmd[]=splitTokns(0L,cmd0), fExe=cmd[0];//,cmd1=cmd.clone();
            putln(ANSI_GREEN+"fExe="+fExe+ANSI_RESET+"  ");
            final ChExec exProbe=new ChExec(0L).setCommandLineV(fExe,"-c");
            addPath(fExe, exProbe);
            exProbe.run();
            if (exProbe.exitValue()==0) {
                ex=new ChExec(ChExec.STDOUT|ChExec.STDERR|ChExec.LOG).setCommandLineV(cmd,"-p", "XXXXXXXXXX");
                ex.printlnStdout(this);
                addPath(fExe, ex);
                ChDelay.newThread(this,"EXEC_PYMOL",ex).start();
                for(int i=10; --i>=0 && !isWaitingForInput; ) sleep(555);
                if (!ex.hasFinished()) {
                    ChDelay.newThread(this,"J",null).start();
                    this.shared_exec=ex;
                    break;
                }
            }
        }
        final String error=ex==null || ex.couldNotLaunch() ? "Pymol not started" :
            ex.hasFinished() ? "Could start Pymol but the process already terminated" : null;
        if (error!=null) putln(RED_WARNING,"error=",error);

        return error==null;
    }
    /* <<< Launch <<< */
    /* ---------------------------------------- */
    /* >>> ChRunnable >>> */
    public Object run(String id, Object arg) {
        if (id=="EXEC_PYMOL") {
            final ChExec ex=((ChExec)arg);
            isRunning=true;
            ex.run();
            isRunning=false;
            dispose();
        }
        if (id=="XXXXXXXXXX") {
            sleep(3333);
            if (!isWaitingForInput) {
                isWaitingForInput=true;
                interpretNow("from pymol import cmd");

                if (_canInstall!=null) _canInstall.setVisible(false);
            }
        }
        if (id==ChRunnable.INTERPRET_LINE && arg!=null) {
            putln("Pymol INTERPRET_LINE ",arg);
            final String s=toStrg(arg).trim();
            if (s.startsWith("ALL_OBJECTS"))  allObjects=s.substring(11);
            if (s.indexOf("XXXXXXXXXX")>=0) {
                ChDelay.newThread(this,"XXXXXXXXXX",null).start();
            }
            if (s.startsWith("PyMOL: ") && s.indexOf("program termination.")>=0) {
                dispos(0L, shared_exec);
                dispose();
            }
            setTextStatic(s);
        }
        if (id=="J") {
            while(isRunning) {
                final String s=(String)ChUtils.get(0,jobs);
                if (s!=null) {
                    jobs.remove(0);
                    interpretNow(s);
                }
                sleep(33);
                if (shared_exec==null || !shared_exec.isRunning() ) jobs.clear();
            }
        }
        return null;
    }
    /* <<< ChRunnable <<< */
    /* ---------------------------------------- */
    /* >>> ActionEvent >>> */
    public void actionPerformed(java.awt.event.ActionEvent ev) {
        final String cmd=ev.getActionCommand();
        final Protein protein=getProtein();
        final ProteinAlignment align=protein!=null ? protein.getAlignment() : null;

        if (cmd==ChExec.ACTION_TERMINATED  && isRunning) {
            for(int i=align==null?0:align.countRows(); --i>=0;) {
                Protein p=align.getProtein(i);
                if (p==null) continue;
                for(ProteinViewer v : p.getProteinViewers()) {
                    if (v instanceof Pymol) new StrapEvent(v,StrapEvent.PROTEIN_VIEWER_CLOSED).run();
                }
            }
            isRunning=false;
            jobs.clear();
        }
        if (cmd=="DL") {
            final String fExe=downloadedBinary;
            synchronized(fExe) {
                if (0==sze(file(fExe))) {
                    final String zipFile=isWindows() ? "PyMOL099" : isMacintosh() ? "macpymol-0_99rc6" : "pymol-0_99rc6-bin-linux-x86-glibc23";
                    putln("DL fExe="+fExe);
                    putln("DL zipFile="+zipFile);
                    putln("DL zipFile vor InteractiveDownload");
                    InteractiveDownload.downloadFilesAndUnzip(Web.urls(urlOfBinariesForThisOS(),zipFile+".zip.gz"),dirBin());
                    putln("DL zipFile nach InteractiveDownload ");
                    final File fZip=file(dirBin()+"/"+zipFile+".zip");
                    if (ChZip.extractAll(fZip,dirBin())) {
                        if (sze(file(fExe))>0) {
                            makeExecutable(file(fExe));
                            makeShellScript();
                            launch();
                        }
                    } else {
                        error("Error: could not install PyMOL "+fZip+" "+sze(fZip));
                        renamFile(fZip, fZip+".error");
                    }
                }
            }
        }
    }
    /* <<< ActionEvent  <<< */
    /* ---------------------------------------- */
    /* >>> Protein >>> */
    private String proteinName;
    private Protein protein;
    public Protein getProtein() {return protein;}

  public boolean setProtein(long options, Protein p, ProteinViewer inSameView) {
        interpretNow("print");/* otherwise waitFor does not terminate */
        final File pdb=p!=null ? p.getFileMayBeWithSideChains() : null;
        if (pdb==null) return false;
        for(ProteinViewer v: p.getProteinViewers()) {
            if (v instanceof Pymol && v!=this) return false;
        }

        protein=p;
        if (shared_exec==null || !shared_exec.isRunning()) launch();
        if (shared_exec==null || !shared_exec.isRunning()) {
            putln(RED_WARNING+"Could not launch PyMOL\n");
            return false;
        }
        String fn=p.toString().replace('!','_');
        final int l=sze(fn);
        char chain=0;
        if (chrAt(l-2,fn)==':') { chain=chrAt(l-1,fn); fn=fn.substring(0,l-2);}
        fn=delSfx(".ent",delSfx(".pdb",fn));
        if (chain!=0) fn+="_"+chain;
        this.proteinName=fn;
        loadObject(pdb, this.proteinName);
        for(HeteroCompound h: p.getDnaAndRnaAndHeteros()) {
            final String n=h.getNameForPymol();
            final File f=h.getFile();
            if (f==null) continue;
            loadObject(f, h.getNameForPymol());
            sendToPymol("hide everything, "+n+"; show "+ (h.isNucleotideChain()?"cartoon, ":"sticks, ")+n);
        }
        ChDelay.runAfterMS(EDT,runAfterAllLoaded,3333);
        shared_vView.add(this);
        thread_setWndwStateT('I',0,0,"PyMOL Tcl/Tk GUI").start();
        return true;
    }

    /* <<< Protein  <<< */
    /* ---------------------------------------- */
    /* >>> CommandInterpreter >>> */
    private final static Vector<String> jobs=new Vector();
    private final static String supported[]={
        COMMANDselect, COMMANDselection_name,
        COMMANDset_rotation_translation, COMMANDcenter,
        COMMANDspheres,COMMANDlines,COMMANDsticks,COMMANDribbons,
        COMMANDcartoon, COMMANDsurface,
        COMMANDcolor,
        COMMANDlabel,
        COMMANDcenter_amino
    };

    public void interpret(long options, String command) {

        final String id0=this.selection_id, id=id0==null?"unnamed" :id0;
        final String word0=toStrgIntrn(toStrgTrim(frstTkn(command)));
        final int L0=sze(word0);
        final int para=nxtE(NO_SPC,command, L0,MAX_INT);
        final long color=rgbInScript(command);
        if (L0==0) return;
        String send=idxOfStrg(word0, allCommands())>=0 ? null : rplcToStrg(PROTEIN,this.proteinName,  rplcToStrg("COLOR","Color"+id,command));
        log3D(true,this,send);
        if (supports(word0,this)) {
            final boolean off=command.endsWith(" off");
            //  final String pfx=off ?"hide " : "hide everything, RESIDUES; show ";
            final String pfx=off ?"hide " : "show ";
            if (word0==COMMANDdots)         send=pfx+"dots,RESIDUES\n";

            else if (word0==COMMANDcenter_amino) send="reset;";
            else if (word0==COMMANDset_rotation_translation) { transform(matrixInScript(command)); send="";}
            else if (word0==COMMANDspheres) send=pfx+"spheres,RESIDUES\n";
            else if (word0==COMMANDmesh)    send=pfx+"mesh,RESIDUES\n";
            else if (word0==COMMANDsurface) send=pfx+"surface,RESIDUES\n";
            else if (word0==COMMANDlines)   send=pfx+"lines,RESIDUES\n";
            else if (word0==COMMANDsticks)  send=pfx+"sticks,RESIDUES\n";
            else if (word0==COMMANDribbons) send=pfx+"ribbon,RESIDUES\n";
            else if (word0==COMMANDcartoon) send=pfx+"cartoon,RESIDUES\n";
            else if (word0==COMMANDcolor)   send="color 0x"+rgbHex((int)(color&0xFFffFF))+" ,RESIDUES\n";
            else if (word0==COMMANDlabel)   send="label RESIDUES, \""+(off?"": delSfx('"', delPfx('"', command.substring(para))))+"\"";
            else if (word0==COMMANDselect) {
                send=selectCmd(command.substring(para));
            } else if (word0==COMMANDselection_name) {
                next_selection_id=command.substring(para);
                send="";
            }
        }
        if (sze(send)>0) {
            sendToPymol(rplcToStrg(RESIDUES,id, send));
        }
    }

    private void sendToPymol(final String cmd0) {
        if (cmd0==null) return;
        final String cmd=rplcToStrg(PROTEIN,this.proteinName,cmd0).trim();
        String startsWith=null;
        for(int i=0; i<3; i++) {
            final String s=i==0?"select cursor " : i==1 ? "select selected " : "cmd.enable(";
            if (cmd.startsWith(s)) startsWith=s;
        }
        if (startsWith!=null) {
            for(String s: strgArry(jobs)) {
                if (s.startsWith(startsWith) || "reset".equals(cmd) && "reset".equals(s))  jobs.remove(s);
            }
        }
        log3D(false, this,cmd);
        jobs.add(cmd);
    }
    private static String interpretNow(String s) {
        final ChExec ex=Pymol.shared_exec;
        if (ex!=null) {
            try {
                shared_exec.write(s);
                shared_exec.write("\r\n");
            } catch(IOException e){}
        }
        return "";
    }
    /* <<< CommandInterpreter <<< */
    /* ---------------------------------------- */
    /* >>> Disposable >>> */
    public void dispose() {
        interpretNow("delete "+this.proteinName);
        interpretNow("delete "+this.proteinName+"_*");
        shared_vView.remove(this);
        new StrapEvent(this,StrapEvent.PROTEIN_VIEWER_CLOSED).run();
    }
    /* <<< Disposable <<< */
    /* ---------------------------------------- */
    /* >>> Object Management >>> */
    private static String allObjects;
    public String[] getAllObjects() {
        if (shared_exec==null || !shared_exec.isRunning()) return NO_STRING;
        allObjects=null;
        /* ALL_OBJECTS ['hs_EscherichiaColi'] */
        interpretNow("print 'ALL_OBJECTS', cmd.get_names('objects')");
        for(int i=0; i<100 && allObjects==null; i++) sleep(20);
        return allObjects!=null ? splitTokns(0L, delPfx("[",delSfx("]",allObjects.trim())), ",'") : NO_STRING;
    }
    public void loadObject(File f, String n0) {
        final String path=toStrg(f);
        if (path!=null && n0!=null) {
            final String n=rplcToStrg(PROTEIN,this.proteinName, n0);
            interpretNow("if not '"+n+"' in cmd.get_names('objects'): cmd.load( '"+path.replace('\\','/')+"' , '"+n+"')");
        }
    }
    private void setTextStatic(final String s) {
        final Protein protein=getProtein();
        if (protein==null) return;
        final ProteinAlignment align=protein.getAlignment();
        if (s!=null && s.startsWith("You clicked") && align!=null) {
            final int
                i1=s.indexOf('/'),
                i2=s.indexOf('/',i1+1),
                i3=s.indexOf('/',i2+1),
                i4=s.indexOf('/',i3+1),
                L=sze(s);
            if (i1>0 && L>i1+1 && i3+1<L && i4+1<L) {
                final String s1=s.substring(i1+1);
                final int
                    numStart=nxtE(DIGT,s, i4+1,MAX_INT),
                    num=atoi(s,numStart),
                    insStart=nxt(NO_DIGT,s, numStart,MAX_INT);

                final char insCode=is(LETTR,s,insStart) ? chrAt(insStart,s) : 0;
                final char chain=is(LETTR_DIGT,s,i3+1)  ? chrAt(i3+1, s) : 0;
                for(int i=align.countRows(); --i>=0;) {
                    final Protein p=align.getProtein(i);
                    if (p==null) continue;
                    for(ProteinViewer pv : p.getProteinViewers()) {
                        final Pymol pymol=deref(pv, Pymol.class);
                        if (pymol!=null && s1.startsWith(pymol.proteinName+"/")) {
                            dispatchEvtPicked(this, num, chain, insCode);
                        }
                    }
                }
            }
        }
    }
    /* <<< Object Management <<< */
    /* ---------------------------------------- */
    /* >>> Customize >>> */
    public static String getCustomizableExamples() {
        return
            "show dots,RESIDUES\n"+
            "show spheres,RESIDUES\n"+
            "show mesh,RESIDUES\n"+
            "show surface,RESIDUES\n"+
            "show lines,RESIDUES\n"+
            "show sticks,RESIDUES\n"+
            "show ribbon,RESIDUES\n"+
            "show cartoon,RESIDUES\n"+
            "color magenta, RESIDUES\n"+
            DEFINE_THIS_COLOR+", color "+PYMOL_COLOR_NAME+",+ RESIDUES\n"+
            sATOM_TYPE_CA+" label (RESIDUES), \"%s\" % (resn)\n"+
            sATOM_TYPE_CA+" label (RESIDUES), \"$FIRST_Aaa\"";
    }
    public static String getCustomizableInitCommands() {
        return
            "hide everything,PROTEIN\n"+
            "show cartoon, PROTEIN\n"+
            "show spheres , PROTEIN and Het and not resn hoh and not resn hho and not  resn ohh and not resn h2o\n"+
            "show lines, (PROTEIN and name ca,c,n)\n"+
            "hide everything,PROTEIN and (resn hoh or resn hho or resn ohh or resn h2o) ";
    }
    public static String getCustomizableLateCommands() {
        return
            "select allWater , resn hoh or resn hho or resn ohh or resn h2o\n"+
            "select allHeteros , Het and not allWater\n"+
            "select allDNA , resn DA or resn DC or resn DG or resn DT  \n"+
            "select allRNA , resn A or resn C or resn G or resn T or resn U  \n"+
            "set selection_width=10\n"+
            "reset";
    }

    public static String getCustomizableExecPath() {
        return
            isWindows() ? "%ProgramFiles%/Delano%20Scientific/PyMOL/PyMOLWin.exe" :
            isMacintosh() ?
            "/Applications/PyMOL.app/Contents/MacOS/PyMOL\n"+
            "/Applications/MacPyMOL.app/Contents/MacOS/MacPyMOL\n"+
            "/Applications/PyMOLX11Hybrid.app/Contents/MacOS/PyMOL\n" :
            "";
    }
    /* <<< Customize <<< */
    /* ---------------------------------------- */
    /* >>> TransformsCoordinates  >>> */
    private final Map<String,Matrix3D>mapMatrix0=new HashMap();
    public void setMatrix0(String object, Matrix3D m) { mapMatrix0.put(rplcToStrg(PROTEIN,this.proteinName,object),m);}
    public void transform(Matrix3D m3d) {
        final Protein p=getProtein();
        //final Matrix3D m3d=p.getRotationAndTranslation();
        final String loaded[]=getAllObjects(), namePfx=this.proteinName+"_";
        final Matrix3D matrix=new Matrix3D();
        final HeteroCompound[] dna=p.getDnaAndRnaStructures(), het=p.getHeteroCompounds();
        for(String l:loaded) {
            boolean isHet=false;
            for(HeteroCompound h : dna) isHet=isHet||l.equals(h.getNameForPymol());
            for(HeteroCompound h : het) isHet=isHet||l.equals(h.getNameForPymol());
            if (isHet || l.equals(this.proteinName) || l.startsWith(namePfx)) {
                sendToPymol("matrix_reset "+l);
                final Matrix3D m0=mapMatrix0.get(l);
                if (m3d!=null || m0!=null) sendToPymol("cmd.transform_object('"+l+"',( "+matrix.unit().mult(m0).mult(m3d).toText(Matrix3D.FORMAT_PYMOL,null)+" ) )");
            }
        }
        sendToPymol("reset");
    }
    /* <<< TransformsCoordinates <<< */
    /* ---------------------------------------- */
    /* >>> Installation >>> */

    private void makeShellScript() {
        if (sze(file(downloadedBinary))==0) return;
        final String
            downlDir= dirBin()+(isWindows() ? "\\PyMOL099\\" : "/pymol/"),
            txt=
            isWindows() ? "set PYMOL_PATH="+downlDir+"\r\nset PATH=%PATH%;%PYMOL_PATH%\r\n"+downloadedBinary +" %* \r\n":
            isMacintosh() ? "#!/bin/bash\nexport PYMOL_PATH="+dirBin()+"/MacPyMOL.app/pymol\n"+downloadedBinary +" $* & \n":
            " export PYMOL_PATH="+downlDir+"\n"+downloadedBinary +" $* \n";
        for(int iD=2; --iD>=0;) {
            final String dir=iD==0?downlDir:dirHome()+"/Desktop/strap_utils";
            if (!file(dir).isDirectory()) continue;
            final File f=file(dir+"/Pymol."+ (isWindows() ? "bat" : isMacintosh() ? "command" : "sh"));
            wrte(f,txt);
        }
    }

    private final String downloadedBinary=
        dirBin()+(
                  isWindows() ? "\\PyMOL099\\PyMOL.exe" :
                  isMacintosh() ? "/MacPyMOL.app/Contents/MacOS/MacPyMOL" :
                  "/pymol/pymol.exe");
    private static ChFrame _canInstall;

    public /* !public */ static Object canInstall() {
        final Pymol p=new Pymol();
        return pnl("STRAP can install Pymol automatically (13MB)", new ChButton("DL")
                   .cp(KEY_AVOID_GARBAGE_COLLECT,p)
                   .t("Download & install").li(p));
    }
    /* <<< Installation <<< */
    /* ---------------------------------------- */
    /* >>> ProteinViewer methods >>> */

    public java.awt.Container getControlPanel() {  return getCtrlPanel(shared_exec); }
    public ProteinViewer[] getViewersSharingView() { return shared_vView.asArray(); }

    private final Object FLAGS=new Long(PROPERTY_NAMED_SELECTIONS|PROPERTY_MULTI_MOLECULE|PROPERTY_EXTERNAL_PROCESS|PROPERTY_SEQUENCE_CURSOR_DELAYED);
    public void setProperty(String id, Object value) {
    }
    public Object getProperty(String getID) {
        if (getID==GET_FRAME_TITLE) return "PyMOL Viewer";
        if (getID==GET_ATOM_SELECTION_EXAMPLE) return " /////CA or /////CB ";
        if (getID==GET_SUPPORTED_COMMANDS) return supported;
        if (getID==GET_FLAGS) return FLAGS;
        return null;
    }
    /* <<< ProteinViewer methods <<< */
    /* ---------------------------------------- */
    /* >>> ResidueSelection >>> */
    private Selection3D selectedResidues[];
    private String selection_id, next_selection_id;
    private String selectCmd(String expr) {
        final Selection3D atoms[]=toSelectionT(expr, getProtein());
        this.selectedResidues=atoms;
        final String sel=selectionToText(PYMOL,atoms);
        final String id=makeSelId(atoms, this.selection_id=this.next_selection_id);
        this.next_selection_id=null;
        final StringBuffer sb=new StringBuffer(99).append("select ").append(id)
            .append(" = ( /PROTEIN ").append(sel!=null?"and":"").append(sel).append(" )\r\n");
        return sb.toString();
    }
    /* <<< ResidueSelection <<< */
    /* ---------------------------------------- */
    /* >>>  >>> */

}

// http://biodesign.asu.edu/labs/assets/lindsay/molvis/MOL.html
/*
if 'pdb1a34_mtx42' in cmd.get_names('objects'): matrix_reset pdb1a34_mtx42

See mgsl 
*/
