package charite.christo.strap;
import charite.christo.*;
import charite.christo.blast.*;
import charite.christo.interfaces.*;
import charite.christo.protein.*;
import java.util.*;
import javax.swing.JComponent;
import static charite.christo.ChConstants.*;
import static charite.christo.ChUtils.*;

/**HELP

Identifying Blast hits that are common to two sequences helps to compare remote homologs.

The analysis requires two steps.

<ol>
<li>One  protein is selected in one, the other in the other protein list. </li>
<li>For each sequence a BLAST search is performed against a non-redundant set of sequences such as uniref50

</li>
<li>For two sequences the BLAST hits are compared and all common segments presented to the user by pressing the button <i>LABEL:ChButton#BUTTON_GO</i>.
</li>

</ol>

Alignment programs perform better when sequences are added that share similarity with the proteins
to be aligned. Using two additional steps one can improve multiple sequence alignments:

<ol>
<li>Chose hits that are similar to either sequence by activating the respective check-boxes.</li>
<li>Protein objects are created for each activated Blast segment. At this stage overlapping segments with the same sequence ID but different start and end positions are joined.</li>
<li>Finally a multiple sequence alignment is computed <i>BUTTON:DialogAlign!</i>.</li>
</ol>

<br>
<b>Related publications:</b> PUBMED:11159329 PUBMED:9367767
<br>
   @author Christoph Gille
*/
public class IntermediateSeq extends AbstractDialogJTabbedPane implements java.awt.event.ActionListener,Disposable,HasControlPanel,StrapListener,ChRunnable {
    private final Map<StrapProtein,Object> mapBlastResult=new WeakHashMap();
    private final List<BlastAlignment> vAlignments=new Vector();
    private final List<String> vDatabases=new Vector();
    private final ChCombo comboDatabases=new ChCombo(0L,vDatabases).li(this);
    private final ChCombo comboClass=StrapAlign.classChoice(SequenceBlaster.class).li(this);
    private final JComponent butLoad=new ChButton("Load all selected",ChButton.DISABLED).li(this);
    private final ChLabel BLAST_labStatus=new ChLabel("blasting, be patient ...");
    private final AsciiView BLAST_LOG=new AsciiView(999), LOG_NO_HITS=new AsciiView("protein pairs without  common blast hits\n\n");

    private ProteinList listProt1,listProt2;

    private void updateDatabases(){
        final SequenceBlaster blaster=comboClass.instanceOfSelectedClass(SequenceBlaster.class);
        final String dd[]=blaster!=null ? blaster.getAvailableDatabases(true, true) : null;
        final Thread
            runSetDB=ChJList.thread_setSelO(ChJList.SEL_STRSTR_IC, "uniref50", comboDatabases),
            runAvail=DialogBlast.thread_setAvailableDB(blaster, new java.util.List[]{null,vDatabases}, new JComponent[]{null,comboDatabases}),
            runReval=thread_revalAndRepaintC(comboDatabases);
        newThrd(runAvail,runSetDB, runReval).start();
        vDatabases.clear();
        adAll(dd,vDatabases);
        if (sze(vDatabases)==0) vDatabases.add("Error: no Database");
        runSetDB.run();
        runReval.run();
    }
    public void actionPerformed(java.awt.event.ActionEvent ev) {
        StrapAlign.errorMsg("");
        final Object q=ev.getSource();
        final String cmd=ev.getActionCommand();
        if (q==comboClass || q==comboDatabases) mapBlastResult.clear();
        if (q==comboClass) updateDatabases();
        if (cmd=="GO") {
            setEnabld(false,butLoad);
            ChDelay.newThread(this,"GO",null).start();
        }
        if (q==butLoad) {
            IntermediateSeq.this.vAlignments.clear();
            for(IntermediateSeqPair sp: childs(this,IntermediateSeqPair.class))
                for(BlastHitIntersection a: sp.vAllAlignments.asArray())
                    if (a.SELECTED[0]) {
                        BlastAlignment ba=new BlastAlignment();
                        ba.setMatchStartEnd(a.getMatchStart(),a.getMatchEnd());
                        ba.setMatchSeq(onlyLetters(a.getSeq('M',0)),0, MAX_INT);
                        ba.setBlastHit(a.getAlignment(0).getBlastHit());
                        IntermediateSeq.this.vAlignments.add(ba);
                    }
            union();
            final List<StrapProtein> vProteins=new Vector();
            for(int i=0; i<sze(vAlignments); i++) {
                BlastAlignment a=vAlignments.get(i);
                final String pn="xx"+a.getID()+'-'+a.getMatchStart()+'-'+a.getMatchEnd();
                if (StrapProtein.getProteinWithName(pn,null)==null){
                    final StrapProtein p=new StrapProtein(align);
                    p.setName(pn);
                    p.setResidueType(a.getMatchSeq());
                    p.setHeader(toStrg(a.getBlastHit().getDescription()));
                    adUniq(p, vProteins);
                }
            }
            if (sze(vProteins)==0) StrapAlign.errorMsg("You must select blast hits in the result tabs");
            final StrapView v=align.getAlignmentPanelWithFocus();
            if (v!=null) v.setCursorToColumn(v.getProtein(v.countRows()-1),1);
            align.addProteins(-1, toArry(vProteins, StrapProtein.class));
        }
    }
    private static ChButton TOG_LOG=toggl("log");
    public JComponent getControlPanel() { return TOG_LOG.cb();}
    @Override protected void finalize() throws Throwable {
        logFinalize(ANSI_CYAN+"IntermediateSeq",getClass());
        super.finalize();
    }

    // ----------------------------------------
    public IntermediateSeq() {
        listProt1=new ProteinList(align,0);
        listProt1.setSelI(0);
        listProt1.li(this);
        listProt2=new ProteinList(align,0);
        listProt2.setSelI(1);
        listProt2.li(this);
        getPanel();
        updateDatabases();
    }
    // ----------------------------------------
    final static ChTextField tfNumAlignments=new ChTextField("55").cols(4,true,true);
    public Object run(String id, Object arg) {
        final Object argv[]=arg instanceof Object[] ? (Object[])arg : null;
        if (id=="GO") {
            boolean show=false;
            int count=0;
            StrapProtein[] pp0=listProt1.getSelectedProteins(),pp1=listProt2.getSelectedProteins();
            if(pp0.length==0) pp0=pp1;
            if(pp1.length==0) pp1=pp0;
            if (pp0.length==0) {
                StrapAlign.errorMsg("Select at least one protein in the top and button protein list");
                BLAST_labStatus.t("error").fg(0xFFffFF);
                return null;
            }
            final BlastResult[][] results=new BlastResult[2][];
            for(int i2=0; i2<2; i2++) {
                final StrapProtein[] pp=i2==0 ? pp0 : pp1;
                results[i2]=new BlastResult[pp.length];
                for(int iP=0; iP<pp.length; iP++) {
                    final StrapProtein p=pp[iP];
                    BlastResult result=gcp(p,mapBlastResult,BlastResult.class);
                    if (result==null) {
                        show=true;
                        BLAST_labStatus.t(count++ +" / "+(pp0.length+pp1.length));
                        final SequenceBlaster blaster=HotswapUtil.instance(comboClass,SequenceBlaster.class,true);
                        if (blaster==null)  { BLAST_LOG.appendS("Error: could not instantiate blast service\n");  return null; }
                        blaster.setAAQuerySequence(p.getResidueTypeAsStringUC());
                        blaster.setDatabase(toStrg(comboDatabases));
                        blaster.setNumberOfAlignments(MAX_INT);
                        blaster.setSensitivity(10);
                        result=BlastResult.newInstance(blaster, "Intermed Seq "+p,null);
                    }
                    if (result!=null) {
                        mapBlastResult.put(p,newSoftRef(result));
                        results[i2][iP]=result;
                    }
                }
            }
            BLAST_labStatus.t("all done").fg(0xFF00);

            ChDelay.invokeAndWait(this,"COMPUTE",new Object[]{pp0,pp1,results[0],results[1]});
            setEnabld(true,butLoad);
        }
        if (id=="COMPUTE") {
            compute((StrapProtein[])argv[0], (StrapProtein[])argv[1], (BlastResult[])argv[2], (BlastResult[])argv[3]);
        }
        return null;
    }
    public void compute(StrapProtein[] pp1,StrapProtein[] pp2, BlastResult[] rr1, BlastResult[] rr2) {
        boolean success=false;
        for(int i=0; i<pp1.length; i++) {
            for(int j=0; j<pp2.length; j++) {
                final StrapProtein p1=pp1[i], p2=pp2[j];
                if (p1==p2)  {
                    LOG_NO_HITS.appendS("skip "+p1+" with itself\n").showInFrame("Protein pairs without  common blast hits",0);
                    continue;
                }
                if (rr1[i]==null) { LOG_NO_HITS.appendS("Error "+p1+": No Blast Result\n"); continue; }
                if (rr1[j]==null) { LOG_NO_HITS.appendS("Error "+p2+": No Blast Result\n"); continue; }
                final IntermediateSeqPair r=new IntermediateSeqPair(p1,p2,rr1[i],rr2[j]);
                if (sze(r.vHitViewShown)==0) {
                    LOG_NO_HITS.appendS(p1+" and "+p2+"\n");
                    LOG_NO_HITS.showInFrame("Protein pairs without  common blast hits",0);
                    r.dispose();
                } else {
                    success=true;
                    adTab("<sub>"+p1+"<br>"+p2+"</sub>",r, this);
                }
                StrapAlign.progressBar((i*pp1.length+j)*100/pp1.length/pp2.length);
            }
        }
        if (success) setSelectedIndex(getTabCount()-1);
    }
    private JComponent getPanel() {
        final JComponent

            butTry=new ChButton("GO").t(ChButton.GO).li(this).tt("Invokes comparison of BLAST hits."),
            pStep23=vB(
                       vBhB("#ETBem",
                            " ",
                            "1st step: Load those blast hits that are<br>activated with the check-boxes \"accept\".",
                            butLoad,
                            " ",
                            "2nd step: Align the sequences together with the loaded blast hits.",
                            StrapAlign.b(CLASS_DialogAlign)
                            )
                       ),
            pDetails=vBhB(
                          "Method:",
                          comboClass.panel(),
                          "Database:",
                          comboDatabases,
                          ChButton.actionShowResults().t("Log blast")
                          ),
            pStep1=vB(KOPT_TRACKS_VIEWPORT_WIDTH,
                      vBpnl("#ETBem",
                           "Find common blast hits.",
                           butTry,
                           hBL("Options ",toggl().actionCollapse(false,pDetails)),
                           pDetails
                           ),
                      " ",
                      hBL("Optional improve alignment:", toggl().actionCollapse(false, pStep23)),
                      pStep23
                      ),

            pCenter=pnl(new java.awt.GridLayout(1,2),scrllpn(0L,pStep1),
                        pnl(new java.awt.GridLayout(2,1),listProt1.scrollPane(),listProt2.scrollPane())),
            pMain=newCNSEW(pCenter,StrapAlign.headLine(this),null,null,null);
        setPrefSze(false, prefWidth(comboClass), prefHeight(butTry), comboClass);
        adMainTab(pMain, this,null);
        return this;
    }
    // ----------------------------------------
    static java.io.File getFileOfHits(StrapProtein pp[],String suffix) {
        final String ss[]=new String[pp.length];
        for(int i=pp.length;--i>=0;) ss[i]=pp[i].toString();
        sortArry(ss);
        final java.io.File dirTmp=file(dirTmp()+"/X-blast");
        mkdrs(dirTmp);
        return file(dirTmp,joinAsStrg(ss,"+")+suffix);
    }
    private void union() {
        for(int i0=sze(vAlignments);--i0>=0;) {
            BlastAlignment a0=vAlignments.get(i0);
            if (a0==null) continue;
            for(int i1=sze(vAlignments);--i1>=0;) {
                final BlastAlignment a1=vAlignments.get(i1);
                if (a1==null || a1==a0) continue;
                final BlastAlignment u=BlastHitIntersection.union(a0,a1);
                if (u!=null) {
                    vAlignments.set(i0,a0=u);
                    vAlignments.set(i1,null);
                }
            }
        }
        while(vAlignments.remove(null));
    }
    public @Override void dispose() {
        for(IntermediateSeqPair p: childs(this,IntermediateSeqPair.class)) p.dispose();
        mapBlastResult.clear();
        super.dispose();
    }
    public void handleEvent(StrapEvent ev) {
        final int type=ev.getType();
        if (type==StrapEvent.AA_SHADING_CHANGED) repaint();
    }
}
// http://bibiserv.techfak.uni-bielefeld.de/altavist/
//http://www.ncbi.nlm.nih.gov/entrez/query.fcgi?cmd=Retrieve&db=pubmed&dopt=Abstract&list_uids=16076885&itool=iconabstr&query_hl=8
