package charite.christo.strap;
import charite.christo.*;
import charite.christo.interfaces.*;
import charite.christo.protein.*;
import java.awt.*;
import java.io.File;
import java.util.*;
import java.util.List;
import static charite.christo.ChConstants.*;
import static charite.christo.ChUtils.*;
import static charite.christo.protein.ProteinModiCount.*;

/**HELP

Residue annotations, allow the assignment of information to specific
amino acids or nucleotides of proteins.

They are special types of residue selections.

A residue annotation has a list of entries each having a name, some
user typed text and a toggle button.

The user can change these entries and add new entries.

The entries can contain several types of information:

<ul>
<li>Rendering commands for 3D-viewers,</li>
<li>Expressions to select certain atoms for the rendering commands to act upon.</li>
<li>LaTeX-code for the PDF-output,</li>
<li>Hyperlinks, PUBMED-references, database references,</li>
<li>Notes, for example literature citations, experimental conditions.</li>
</ul>

More detailed information is found in <i>ITEM:ResidueAnnotationView</i>

A ResidueAnnotation can be created in different ways:
<ul>

<li>A residue selection can be created by dragging the mouse over the sequence. Right-click opens the context menu of the residue
selection.  With the menu item
<i>BUTTON:ResidueSelectionPopup#ACTION_convertToResidueAnnotation</i> a
residue annotation can be created that has the same amino acid
positions.</li>

Alternatively, use the context menu of the alignment panel to create an annotation for the protein under the mouse pointer.

</ul>

Residue annotations which are highlighted in the alignment can be selected in two different ways: By CTRL left-click or by dragging  a rectangular box.
Selected residue annotations have WIKI:Marching_ants.

<br><br><b>Sequence Features: </b>

Features like phosporylation sites and active sites are special
annotations which are retrieved from a computational service and are
therefore not saved to hard disk.  Their color cannot be changed.

<h2>Background Images</h2>

<i>INCLUDE_DOC:ChIcon</i>

<i>SEE_DIALOG:DialogResidueAnnotationChanges</i>
<i>SEE_DIALOG:DialogResidueAnnotationList</i>
<i>SEE_CLASS:ResidueAnnotationView</i>
<i>SEE_CLASS:ActivateDeactivate</i>
<i>SEE_CLASS:AddAnnotation</i>
*/
public final class ResidueAnnotation
    implements HasWeakRef, HasName, ResidueSelection,
    charite.christo.strap.interfaces.SelectorOfNucleotides,
    VisibleIn123,Colored,Disposable,
    TabItemTipIcon.Interface, HasRendererText, HasRendererOptions, HasRendererModiCount, HasImage, HasPropertyMap,
    ChRunnable, HasModiCount {

    public final static String
    NAME="Name",POS="Positions",GROUP="Group",ATOMS="Atoms",
        COLOR="Color", BG_IMAGE="BG_IMAGE", VIEW3D="3D_view", SPECIFIC_VIEW3D="Specific_3D_view",
        MAIN_KEYS[]={NAME,GROUP,POS},
        OTHER_KEYS[]={BG_IMAGE, CLASS_Texshade, VIEW3D, SPECIFIC_VIEW3D, ATOMS};
    public final static ResidueAnnotation NONE[]={};
    private static int maxLengthOfUnderlining=7;
    static void setMaxLengthOfUnderlining(int l) { maxLengthOfUnderlining=l;}

    private final Object WEAK_REF=newWeakRef(this);
    public Object getWeakRef() { return WEAK_REF;}

    private int modiCount;
    public int getModificationCount() { return this.modiCount;}
    /* ---------------------------------------- */
    /* >>> Instance  >>> */
    public ResidueAnnotation(StrapProtein p) {
        refProtein=weakRef(p);
        for(int i=0; i<2; i++) {
            final String v=i==0?"new annotation" : i==1?"new selection group" : "";
            final String k=i==0?NAME : i==1?GROUP : POS;
            if (getEntryWithKey(k)==null) addE(k,v);
        }
        vAll.add(WEAK_REF);
    }
    private final long created=System.currentTimeMillis();
    public long getWhenCreated() { return created;}
    public void dispose() {
        final Protein protein=getProtein();
        if (protein!=null) protein.removeResidueSelection(this);
        rendererText=" Discarded: "+getName();
        vAll.remove(WEAK_REF);
    }
    /**  ResidueAnnotation-Objects belong to a certain protein */
    public StrapProtein getProtein() { return (StrapProtein)deref(refProtein); }
    private Object refProtein;
    public boolean setProtein(Protein p) { refProtein=weakRef(p);  return true; }
    /* <<< Instance <<< */
    /* ---------------------------------------- */
    /* >>> Name and ID >>> */
    /**  A ResidueAnnotation-object has a name. */
    public String getName() {return getValue(NAME);}
    private ByteArray baID;
    public String getID() {
        if (baID==null) baID=new ByteArray(99);
        return baID.setEnd(0).filter(LETTR_DIGT, getProtein(), 0 , MAX_INT, '_').append('_').filter(LETTR_DIGT, getName(), 0 , MAX_INT, '_').newString();
    }
    /* <<< Name and ID  <<< */
    /* ---------------------------------------- */
    /* >>> Selected Nucleotides and Aminoacids >>> */
    private final static int IS_SEL_NT=0, IS_SEL=1, IS_SEL_V=2, REND_TEXT=3, REND_TEXT_JLIST=4;
    private final long cacheMC[]=new long[6];
    private transient Object rCached;
    private Object[] cached() {
        Object[] r=(Object[])deref(rCached);
        if (r==null) rCached=newSoftRef(r=new Object[6]);
        return r;
    }
    public void setSelectedAminoacids(boolean bb[],int offset) {
        final StrapProtein p= getProtein();
        setValue(POS,Protein.selectedPositionsToText(bb,0,p));
        setEnabled(true,ResidueAnnotation.POS);
    }
    /** e.g. s.setSelectedAminoacids("3-5"); */
    public void setSelectedAminoacids(String selection) { setValue(POS,selection); setEnabled(true,POS);}
    public void setSelectedNucleotides(String selection) { setValue(POS,selection);setEnabled(false,POS);}
    private int selectedAAOffset;
    public int getSelectedAminoacidsOffset() {
        getSelectedAminoacids();
        final int offset=this.selectedAAOffset;
        return offset;
    }
    public boolean[] getSelectedAminoacids() {
        final Object[] cached=cached();
        parseResidueAnnotationIfNecessary(cached);
        boolean bb[]= (boolean[])cached[IS_SEL];
        return bb!=null ? bb : NO_BOOLEAN;
    }
    private int isSelectedNtOffset;
    public int getSelectedNucleotidesOffset() {
        final Object[] cached=cached();
        parseResidueAnnotationIfNecessary(cached);
        return isSelectedNtOffset;
    }
    public boolean[] getSelectedNucleotides() {
        final Object[] cached=cached();
        parseResidueAnnotationIfNecessary(cached);
        return (boolean[])cached[IS_SEL_NT];
    }

    /* <<< Selected Nucleotides and Aminoacids <<< */
    /* ---------------------------------------- */
    /* >>> Rendering >>> */
    /**  Rendering style  */
    public int getStyle() {  return style>=0 ? style : iconImage==null ? VisibleIn123.LOWER_HALF_BACKGROUND : VisibleIn123.IMAGE; }
    /**  Rendering style  */
    public void setStyle(int i) {style=i;}
    private int style=-1;
    /** sets the color */
    public void setColor(Color c) {
        if (c!=this.color) {
            this.color=c;
            modiCount++;
            Protein.incrementModificationCount(MC_RESIDUE_SELECTION_LABELS, getProtein());
            StrapAlign.dispatchLater(StrapEvent.RESIDUE_SELECTION_CHANGED_COLOR,111);
        }
    }
    public Color getColor() { return color; }
    void setColorForWhiteBG(Color c) { colorPaper=c; }
    Color getColorForWhiteBG() { return colorPaper!=null ? colorPaper : color; }
    private Color color=Color.PINK, colorPaper;
    private int shownWhere=0xffffffff;
    public int getVisibleWhere() { return style!=VisibleIn123.HIDDEN && isEnabled() ? shownWhere : 0;}
    public void setVisibleWhere(int where) { shownWhere=where;}
    /* <<< Rendering  <<< */
    /* ---------------------------------------- */
    /* >>> Image Icon >>> */
    private Image iconImage;
    public Image getImage(Component observer) {  return iconImage; }

    private final static Object[] provideStrings={null,null}, provideWC={null,null};
    public Object run(String id, Object arg) {
        final Object argv[]= arg instanceof Object[] ? (Object[])arg : null;
        if (id==DialogStringMatch.PROVIDE_STRINGS) {
            provideStrings[0]=getName();
            provideStrings[1]=vEntries;
            return provideStrings;
        }
        if (id==ChRunnable.SET_ICON_IMAGE) {
            Image im=(Image)arg;
            if (iconImage!=im) {
                StrapAlign.dispatchLater(StrapEvent.RESIDUE_SELECTION_CHANGED_COLOR,444);
                iconImage=im;
                icon=null;
                this.modiCount++;
            }
        }
        if (id==PROVIDE_WORD_COMPLETION_LIST){
            provideWC[0]=getRendererText(HasRendererText.JLIST);
            provideWC[1]=getName();
            return provideWC;
        }
        return null;
    }
    private String imageId, lastImage;
    String getLastImage() { return lastImage;}
    void findIconImage() {
        lastImage=getLastValue(BG_IMAGE);
        final String newImageId=lastImage!=null ? lastImage : null;
        if (imageId!=newImageId) {
            imageId=newImageId;
            icon=null;
            iconImage=null;
            requestImage(imageId,this);
        }
    }
    private javax.swing.ImageIcon icon;
    public javax.swing.ImageIcon getIcon() {
        final Image im=iconImage;
        if (im==null) return null;
        if (icon==null) {
            Image imSmall=Toolkit.getDefaultToolkit().createImage(new java.awt.image.FilteredImageSource(im.getSource(),   new java.awt.image.ReplicateScaleFilter(ICON_HEIGHT, ICON_HEIGHT)));
            if (imSmall!=null) icon=new javax.swing.ImageIcon(imSmall);
        }
        return icon;
    }
    /* <<< Image Icon <<< */
    /* ---------------------------------------- */
    /* >>> Tooltip  >>> */
    private TabItemTipIcon titi;
    private boolean titiShift;
    private int titi_mc;
    public TabItemTipIcon getTabItemTipIcon(AWTEvent ev) {
        final boolean shift=isShift(ev);
        final int mc=getModificationCount();
        if (titi==null || titiShift!=shift || titi_mc!=mc) {
            titi_mc=mc;
            titiShift=shift;
            final Color color=getColorForWhiteBG();
            final boolean[] sel=getSelectedAminoacids();
            final int offset=getSelectedAminoacidsOffset();
            final ByteArray sb=new ByteArray(99).append("<span style=\"background-color: ");
            toHexB(color!=null ? color.getRGB()&0xFFffFF : 0, 6, sb).append(";\"> &nbsp; </span>&nbsp;");
            final Protein p=getProtein();
            final String pN=p==null?"":p.toString(), n=getName(), f=getFeatureName();
            if (f!=null) sb.append(" feature ").append(f).append(" ");
            else sb.append("annotation ").append(getName());
            sb.append("<small>");
            for(Entry e : getEntries()) {
                final String k=e.getKey();
                if (k==POS || k=="Balloon" || shift && (k=="Remark" || k=="Note")) {
                    final String v=e.getValue();
                    if (v!=null) sb.append(f==null ? "<br>" :"").append(htmlEncod(v, null));
                }
            }
            if (simRegion!=null && shift) {
                sb.append("<pre>");
                simRegion.getLocalAlignment(offset+firstTrue(sel)-5, offset+lastTrue(sel)+5, sb);
                sb.append("\n</pre>");
            } else sb.append("<br>");
            if (cached()[IS_SEL_V]!=null) sb.append("Positions ").append(getValue(POS)).append("<br>");
            sb.append("</small>");
            final String tab="<sub>"+
                (pN!=null ? pN.substring(0,mini(10,sze(pN))) : null)+
                "<br>"+
                (n!=null ? n.substring(0,mini(10,sze(n))) : null)+
                "</sub>";
            titi= new TabItemTipIcon(tab,null, sb.newString(), getIcon());

        }
        return titi;
    }
    private String rendererText;
    public void setRendererText(String s) { rendererText=s; }
    public long getRendererOptions() {
        final Protein p=getProtein();
        return p==null || !cntains(this, p.getAllResidueSelections()) ? HasRendererOptions.STRIKE_THROUGH : 0L;
    }

    public CharSequence getRendererText(int options) {
        final Object[] cached=cached();
        final boolean jList=(options&HasRendererText.JLIST)!=0;
        final int cacheID=jList ? REND_TEXT_JLIST : REND_TEXT;
        final Protein p=getProtein();
        long hc=hashCd(rendererText)+getModificationCount();
        if (p!=null) hc+=p.getModificationCount(MC_PROTEIN_LABELS)+p.getModificationCount(MC_RESIDUE_SELECTIONS);
        String s=(String)cached[cacheID];
        if (s==null || cacheMC[cacheID]!=hc) {
            cacheMC[cacheID]=hc;
            final String rt=rendererText!=null ?  rendererText : getName();
            cached[cacheID]=s= jList ? getProtein()+"/"+rt : rt;
        }
        return s;
    }
    private static boolean _rendererInit;
    {
    if (!_rendererInit) {
        _rendererInit=true;
        TabItemTipIcon.set("",VIEW3D,"Script lines, suitable for different protein viewers",iicon(IC_JMOL),VIEW3D);
        TabItemTipIcon.set("",SPECIFIC_VIEW3D,"Script lines for one particular protein viewers",iicon(IC_JMOL),SPECIFIC_VIEW3D);
    }
    }
    final long hashCodeShift=hashCode()<<32;
    public long getRendererModiCount() {
      final StrapProtein p=(StrapProtein)getProtein();
      long mc=hashCodeShift+modiCount;
      if (p!=null) { p.getResidueAnnotations(); mc+=p.getModificationCount(MC_RESIDUE_ANNOTATIONS_V)<<16; }
      return mc;
    }

    /* <<< Tooltip <<< */
    /* ---------------------------------------- */
    /* >>>  Get Entry >>> */
    String getLastValue(String key) {
        final Entry ee[]=getEntries();
        for(int i=ee.length; --i>=0;) {
            if (ee[i].isEnabled() && key==ee[i].getKey()) {
                final String v=ee[i].getValue();
                if (v.length()>0) return v;
            }
        }
        return null;
    }
   /**  Return the 1st value with the key. */
    public String getValue(String key) {
        final Entry e=key!=null ? getEntryWithKey(key) : null;
        return e!=null ? e.getValue(): null;
    }

    public boolean isAmino() { return isEnabled(ResidueAnnotation.POS);}

     /** Returns the status of the i-th key-value pair or null. */
    boolean isEnabled(String key) {
        for(Entry e:vEntries.asArray()) {
            if (key==e.getKey() && e.isEnabled()) return true;
        }
        return false;
    }
    /* <<< Get Value <<< */
    /* ---------------------------------------- */
    /* >>> Set Value >>> */
    /** Adds a new key-value-pair */
    public void setValue(String key,CharSequence textValue) {
        if (key==null) return;
        final String value=toStrg(textValue);
        boolean found=false;
        for(Entry e:getEntries()){
            if (key.equals(e.getKey())) {
                setValue(value,e,true);
                found=true;
            }
        }
        if (!found && value!=null) addE(key,value);
        if (BG_IMAGE==key) findIconImage();
        Protein.incrementModificationCount(MC_RESIDUE_SELECTIONS,getProtein());
    }

    boolean setValue(final String value, Entry e, boolean alsoTextField) {
        if (value==null || e==null) removeEntry(e);
        else if (!value.equals(e.getValue())) {
            final String k=e.getKey();
            String v=value.trim();
            if (k==NAME || k==GROUP) {
                v=rplcNoFilenameChars(v);
            }
            e._value=v;
            if (alsoTextField) {
            final ChTextField tf=deref(e.TEXTFIELD[0],ChTextField.class);
            if (tf!=null) tf.t(v);
            }
            this.modiCount++;
            final Protein p=getProtein();
            Protein.incrementModificationCount(MC_RESIDUE_SELECTIONS,p);
            if (k==NAME) Protein.incrementModificationCount(MC_RESIDUE_SELECTION_LABELS, p);
            if (k==GROUP) Protein.incrementModificationCount(MC_RESIDUE_ANNOTATION_GROUPS, p);
            return true;
        }
        return false;
    }
    public void setEnabled(boolean b) {
        setEnabled(b,NAME);
    }
    public void setEnabled(boolean b, String key) {
        final Protein p=getProtein();
        for(Entry e:getEntries()) {
            if (key==e.getKey() && e.isEnabled()!=b) {
                e.setEnabled(b);
                Protein.incrementModificationCount(MC_RESIDUE_SELECTIONS,p);
                if (key==NAME) Protein.incrementModificationCount(MC_RESIDUE_SELECTION_LABELS, p);
                if (key==GROUP) Protein.incrementModificationCount(MC_RESIDUE_ANNOTATION_GROUPS,p);
            }
        }
    }
    /* <<< Set Value <<< */
    /* ---------------------------------------- */
    /* >>> Get Entry Objects >>> */
    private Entry onlyEntry;
    void setOnlyEntry(Entry e) {onlyEntry=e;}
    private final UniqueList<Entry> vEntries=new UniqueList(Entry.class);
    public Entry[] getEntries() {
        Entry oe=this.onlyEntry;
        return oe!=null ? new Entry[]{oe}:vEntries.asArray();
    }
    public void removeEntry(Entry entry) { vEntries.remove(entry); modiCount++; }
    public final static long ENTRY_NOT_SAVED=1<<1;
    public void addE(String key0, String value) { addE(0L, key0,value,true, MAX_INT); }
    public void addEnS(String key0, String value) { addE(ENTRY_NOT_SAVED, key0,value,true, MAX_INT); }
    public void addE(long options,String key0, String value, boolean isEnabled, int row) {
        if (key0==null || value==null) return;
        final String key=key0.intern();
        if (cntains(key,MAIN_KEYS)) {
            final Entry e=getEntryWithKey(key);
            if (e!=null) {
                e.setEnabled(isEnabled);
                setValue(value, e,true);
                e.options=options;
                return;
            }
        }
        final int L=sze(vEntries);
        vEntries.add(row<0? L : mini(L,row), newEntry(options, key,value,isEnabled));
        modiCount++;
        if (BG_IMAGE==key) findIconImage();
    }
    Entry getEntryWithKey(String key) {
        for(Entry e:getEntries()) if (key==e.getKey()) return e;
        return null;
    }
    Entry getEntryAt(int i) { return vEntries.get(i);}

    /* <<< Entry List  <<< */
    /* ---------------------------------------- */
    /* >>> Parse >>> */
    /**
       Example
       texshade=  X\feature...
       ^
       i1234
       value is substring(i+4);
       X/x  is substring(i+3);
    */
    private void parseLines(String[] lines) {
        if (sze(lines)==0) return;
        for(String s:lines){
            if (s.startsWith("charite.christo.strap.BG_Image")) s=BG_IMAGE+s.substring(30);
            final int i=s.indexOf('=');
            if (i<2 || s.length()<i+4) continue;
            final String k=s.substring(0,i).intern(),v=s.substring(i+4);
            if (k==COLOR) {
                color=str2color(v);
                if (color==null) color=Color.PINK;
            } else {
                addE(0L, k,v,(s.charAt(i+3)=='X'), MAX_INT);
            }
        }
    }

    private static Boolean allEnabled;
    static void setEnabledAll(Boolean b) {
        allEnabled=b;
        Protein.incrementModificationCount(MC_RESIDUE_SELECTIONS, null);
    }
    public boolean isEnabled() {
        if (getFeatureName()==null) return isEnabled(NAME);
        final boolean ret=allEnabled==Boolean.TRUE ? true : allEnabled==Boolean.FALSE ? false : isEnabled(NAME);
        return ret;

    }
    private int _range[]={0,0};
    int[] getRange() {
        getSelectedAminoacids();
        return _range;
    }
    private void parseResidueAnnotationNT(String sSet, Object[] cached) {
        final StrapProtein protein=getProtein();
        if (protein==null) {
            cached[IS_SEL_NT]=NO_BOOLEAN;
            isSelectedNtOffset=0;
        }
        else {
            final boolean bb[]=parseSet(sSet,protein.countNucleotides(),-1, null);
            final int first=firstTrue(bb);
            cached[IS_SEL_NT]=trimArray(bb,first);
            isSelectedNtOffset=first<0 ? 0 : first;
        }
    }
    /* <<< Parse <<< */
    /* ---------------------------------------- */
    /* >>> Parse >>> */
    private int  aa_modificationCount;
    private void parseResidueAnnotationIfNecessary(Object[] cached) {
        final String txt=getValue(ResidueAnnotation.POS);
        final StrapProtein p=(StrapProtein)getProtein();
        if (txt!=null && p!=null) {
            final int modi=p.getModificationCount(MC_RESIDUE_TYPE)+this.modiCount+(maxLengthOfUnderlining<<10);
            final boolean isA=isAmino();
            if (cached[IS_SEL]==null || !isA && cached[IS_SEL_NT]==null || aa_modificationCount!=modi) {
                aa_modificationCount=modi;
                if (isA) {
                    cached[IS_SEL]=parseResidueAnnotationAA(txt);
                    isSelectedNtOffset=0;
                } else {
                    parseResidueAnnotationNT(txt,cached);
                    cached[IS_SEL]=p.nucleotideSelection2aminoAcidSelection( (boolean[])cached[IS_SEL_NT],isSelectedNtOffset);
                }
            }

        }
    }

    private boolean[] parseResidueAnnotationAA(String residueRange) {
        final StrapProtein p=getProtein();
        final int nR=p.countResidues();
        final ReferenceSequence refS;
        final String rr;
        {
            final String s=residueRange.trim();
            final int ref=s.indexOf("REFERENCE=");
            if (ref>=0) {
                final String uniprot=wordAt(ref+10,s);
                refS=p.getReferenceSequence(uniprot,(ByteArray)null, true, (Runnable)null);
                rr=s.substring(0,ref)+substrg(s,  ref+10+uniprot.length(),MAX_INT);
            } else {
                refS=null;
                rr=s;
            }
        }
        final StrapProtein refP=refS!=null ? refS.getReferenceProtein() : null;
        final boolean bb[]=
          refS==null ? p.residueSubsetAsBoolean(false, rr,1, _range) :
          refP!=null ? refS.uniprotIdx2Idx(refP.residueSubsetAsBoolean(false, rr,1, null)) :
            null;

        if (refP!=null) { _range[0]=firstTrue(bb); _range[1]=lastTrue(bb)+1; }
        final int firstTrue=firstTrue(bb);
        if (firstTrue<0) return NO_BOOLEAN;
        this.selectedAAOffset=firstTrue;
        final boolean selAA[]=trimArray(bb,firstTrue);
        final String fn=getFeatureName();
        final Object[] cached=cached();
        if (fn!=null && selAA.length>maxLengthOfUnderlining && countTrue(selAA)==selAA.length) {
            final boolean lastResIsSelected=selAA.length+selectedAAOffset>=nR;
            final boolean[] selAA_d=new boolean[lastResIsSelected ? 2 : selAA.length];
            selAA_d[0]=selAA_d[1]=selAA_d[0]=selAA_d[selAA_d.length-1]=selAA_d[selAA_d.length-2]=true;
            cached[IS_SEL_V]=selAA_d;
        } else cached[IS_SEL_V]=null;
        return selAA;
    }
    public static void loadResidueAnnotation(StrapProtein p,File  file) {
        if (sze(file)==0) return;
        final String txt[]=readLines(file);
        if (txt==null) return;
        int n=0, lineNumber[]=new int[txt.length];
        for(int i=0;i<txt.length;i++) {
            txt[i]=rplcToStrg(CLASS_Hyperrefs,"Note",txt[i]);
            if (txt[i].startsWith(NAME+"=")) lineNumber[n++]=i;
        }
        for(int i=0;i<n;i++) {
            final ResidueAnnotation s=new ResidueAnnotation(p);
            final int
                from=lineNumber[i],
                to= (i==n-1) ? txt.length : lineNumber[i+1];
            final String lines[]=new String[to-from];
            System.arraycopy(txt,from,lines,0,to-from);
            s.parseLines(lines);
            p.addResidueSelection(s);
        }
    }
    /* <<< Parse <<< */
    /* ---------------------------------------- */
    /* >>> Save >>> */
    private static String line(Entry e) {
        return
            e==null ? null :
            e.getKey()+"="+
            ( e.isEnabled() ? "  X" : "  x")+
            e.getValue()+"\n";
    }
    public static File save(StrapProtein p, File dir) {

        final ResidueAnnotation[] aa=p.getResidueAnnotations();
        final StringBuffer sb=new StringBuffer();
        for(ResidueAnnotation a:aa) {
            if (!isAnno(a)) delFile(a.featureFile());
        }
        for(ResidueAnnotation a:aa) {
            final File f=a.featureFile();
            if (f!=null) {
                sb.setLength(0);
                for(Entry e: a.getEntries()) {
                    if (cntains( e.getKey(), MAIN_KEYS)) continue;
                    if (0!=(e.options&ENTRY_NOT_SAVED)) continue;
                    sb.append(line(e));
                    //putln("save  "+e.getKey()+"  "+e.options);
                }
                if (sze(sb)>0) wrte(f,sb);
            }
        }
        final File f=Protein.strapFile(dir,p,Protein.mANNO);
        for(ResidueAnnotation a:aa) {
            if (isAnno(a)) {
                sb.append(line(a.getEntryWithKey(NAME)));
                for(Entry e: a.getEntries()) if (NAME!=e.getKey()) sb.append(line(e));
                final Color c=a.getColor();
                if (c!=null) {
                    final int rgb=c.getRGB();
                    sb.append(line(a.newEntry(0L, COLOR, ((rgb>>16)&255)+","+ ((rgb>>8)&255)+","+(rgb&255)+","+c.getAlpha(),true)));
                }
            }
        }
        if (sze(sb)==0) delFile(f);
        return wrte(f,sb) ? null : f;
    }
    /* <<<  Save <<< */
    /* ---------------------------------------- */
    /* >>> Entry Instance >>> */
    private Entry newEntry(long options, String key,String value, boolean status) {
        final Entry e=new Entry();
        e.resAnno=WEAK_REF;
        e.options=options;
        e._status=status;
        final String kShort=StrapPlugins.mapL2S(key);
        e._key=kShort!=null?kShort:key;
        e._value=value.trim();
        return e;
    }
    public final static class Entry implements ChRunnable {
        private long options;
        private boolean _status;
        private String _key, _value;
        private Object resAnno;
        final Object TEXTFIELD[]={null};
        private Entry(){}
        public String toString() { return "RA.E@"+_key;}
        public void setEnabled(boolean b) {
            if (_status!=b) {
                _status=b;
                final ResidueAnnotation a=getResidueAnnotation();
                if (a!=null) a.modiCount++;
            }
        }
        public ResidueAnnotation getResidueAnnotation() { return deref(resAnno,ResidueAnnotation.class);}
        public String getValue() { final String v=this._value; return v!=null ? v : "";}
        public boolean setValue(final String value) {
            final ResidueAnnotation a=getResidueAnnotation();
            if (a!=null) return a.setValue(value,this,true);
            return false;
        }
        public boolean isEnabled() { return _status;}
        public String getKey() { return _key;}
        /*
        private Class _keyClass;
        public Class getKeyClass() {
            if (_keyClass==null) {
                _keyClass=HotswapUtil.getClass(getKey());
                if (_keyClass==null) _keyClass=Object.class;
            }
            return _keyClass==Object.class||_keyClass==String.class?null:_keyClass;
        }
        */
        public Object run(String id, Object arg) {
            if (id==DialogStringMatch.PROVIDE_STRINGS || id==PROVIDE_WORD_COMPLETION_LIST) return getValue();
            return null;
        }
    }
    /* <<<  Entry Class <<< */
    /* ---------------------------------------- */
    /* >>> Client Property >>> */
    private Map _clientObjects;
    public Map getPropertyMap(boolean create) {return _clientObjects==null&&create?_clientObjects=new HashMap() : _clientObjects;}
    /* <<< Client Property <<< */
    /* ---------------------------------------- */
    /* >>> Sequence Features >>> */
    private String featureName, featureSrcDb="", featureSeqName="";
    private int whereLoadedFrom;
    private File featureFile;
    private File featureFile() {
        if (this.featureFile==null) {
            final File dir=StrapAlign.isViewer() ? file(dirSettings()+"/strap") : dirWorking();
            final String fn=rplcChars(charClass(NO_FILENM),'_',getProtein()+"_"+getValue(POS)+"_"+featureName+"_"+featureSrcDb+"_"+featureSeqName+".seqFeatureAnno");
            this.featureFile=file(dir+"/"+DIR_ANNOTATIONS+"/"+fn);
        }
        return this.featureFile;
    }
    public void readFeatureAnnotations() {
        parseLines(readLines(featureFile()));
    }
    public void setFeatureName(String name, int whereLoadedFrom) {
        featureName=name!=null ? name.intern() : null;  this.whereLoadedFrom=whereLoadedFrom;
    }
    public String getFeatureName() { return featureName;}
    public int getWhereFeatureLoadedFrom() { return whereLoadedFrom;}
    private String cssClass;
    static String toCssClass(String s) {
        String css=s!=null ? s.replaceAll("[^0-9A-Za-z_]","_") : "";
        if (css.length()==0) return "annotation";
        return  (is(LETTR,css,0)) ? css : "x"+css;
    }
    public String getCssClass() {
        if (cssClass==null)  cssClass=toCssClass(getFeatureName());
        return cssClass;
    }
    void setFeatureSrc(String db, String seqName) { featureSrcDb=db; featureSeqName=seqName; }
    String getFeatureSrcDb() { return featureSrcDb;}
    private ReferenceSequence simRegion;
    private float simRegionScore=Float.NaN;
    float getFeatureAlignmentScore() {  return simRegionScore;}
    void setReferenceSequence(ReferenceSequence s, int start, int end) {
        if (s!=null) simRegionScore=s.getAlignmentScore();
        simRegion=s;
    }
   private CharSequence srcText;
    public CharSequence getSrcText() { return srcText;}
    public void setSrcText(CharSequence t) {  srcText=t;}

    /* <<< Sequence Features <<< */
    /* ---------------------------------------- */
    /* >>> DnD >>> */
    public boolean verifyDropRows(boolean drop, String data, int row, Point screen) {
        boolean success=false;
        if (data.indexOf('\t')>=0) {
            final Entry ee[]=getEntries();
            for(String line : splitLnes(0L, data)) {
                final String fields[]=splitTokns(SPLIT_ALLOW_EMPTY_TOKENS, line,"\t");
                if (sze(fields)<3+2) continue;
                final String k=fields[1], v=fields[2], bool=fields[3];
                final int hc=atoi(fields,4), srcRow=atoi(fields,5);
                if (!drop) {
                    for(Entry e : ee) success=success|| k.equals(e.getKey());
                    success=success
                        || idxOfStrg(k, MAIN_KEYS)>=0
                        || idxOfStrg(k, OTHER_KEYS)>=0
                        || idxOfStrg(k, AddAnnotation.customizeKeys().getSettings() )>=0;
                } else {
                    success=true;
                    int mvE=-1;
                    if (hc==hashCode()) {
                        for(int iE=ee.length; --iE>=0;) {
                            if (k.equals(ee[iE].getKey()) && v.equals(ee[iE].getValue())) {
                                mvE=iE;
                                if (iE==srcRow) break;
                            }
                        }
                    }
                    if (mvE>=0) {
                        final Point start=dndScreenLocation();
                        final int diff=start==null||screen==null?0 : screen.y-start.y;
                        if (mvE==row) row+=diff>EX/2?1:diff<EX/2?-1:0;
                        mvElements(mvE,mvE+1, row, vEntries);

                    } else addE(0L, k, v, isTrue(bool), row);
                    modiCount++;
                }
            }
        }

        if (success) {
            revalAndRepaintC(getPnl(gcp(KEY_GET_VIEW, this)));
            StrapAlign.dispatchLater(StrapEvent.RESIDUE_SELECTION_CHANGED, 111);

        }
        return success;
    }
    /* <<< DnD  <<< */
    /* ---------------------------------------- */
    /* >>> static utils >>> */
    public static boolean[] getSelectedAminoacidsDisplayed(ResidueSelection s) {
        if (s instanceof ResidueAnnotation) {
            final ResidueAnnotation a=(ResidueAnnotation)s;
            final boolean sa[]=a.getSelectedAminoacids();
            final boolean saViewed[]=(boolean[])a.cached()[IS_SEL_V];
            return saViewed!=null ? saViewed : sa;
        }
        return s!=null ? s.getSelectedAminoacids() : NO_BOOLEAN;
    }

    static boolean isAnno(Object o) {return o instanceof ResidueAnnotation && ((ResidueAnnotation)o).getFeatureName()==null;}
    static boolean isEnabledAnno(Object o) {return isAnno(o) && ((ResidueAnnotation)o).isEnabled();}
    static final java.util.List vAll=new ArrayList(999);
}

