package charite.christo.strap;
import charite.christo.*;
import charite.christo.interfaces.*;
import charite.christo.protein.*;
import java.awt.*;
import java.util.*;
import java.util.List;
import javax.swing.*;
import static charite.christo.ChUtils.*;
import static charite.christo.ChConstants.*;
import static java.awt.event.MouseEvent.*;
import static charite.christo.strap.StrapUtils.*;
/**HELP
   <h2>Sequence Features</h2>

   Sequence features are annotations of signal peptides, disulfide bonds,
   glycosylated or phosphorylated amino acids at sequence level.  In
   Strap they are loaded from the protein file, from BioDAS* servers, from Expasy or from the
   Catalytic Site Atlas at the EBI.  They are underlined in the alignment
   panel and in the alignment output.  They have a context menu
   (right-click), a balloon message and they can be dragged to other proteins or 3D-views.

   <br><br> <h3>The Table</h3>

   The table lists the sequence features of all or of all selected proteins.  Each
   table row gives the color, the type such as "Metal binding", and the number of features.  The following
   check-boxes allow to view or to hide the sequence features in the alignment panel, the 3D-backbone, the PDF
   output etc.  The last columns contain buttons for navigation to the selected positions in the alignment
   pane.  By clicking the first cell of a table row, all sequence features of the given type are selected as
   indicated by WIKI:Marching_ants.

   <h3>By sequence similarity</h3>

   Sequence features my be indirectly inferred from a homologous protein ("By similarity").  In this
   case the sequence positions are assigned using a sequence alignment method

   If the alignment is poor, the sequence features is deaktivated and not shown.

   Sequence features are special types of residue annotations with the difference that they are not
   saved when the project is saved.

   <h3>Colors</h3>

   Each feature type has a color for black background (screen), and one for white background (paper).
   This table can be customized <i>CUSTOMIZE:SequenceFeatures#customizeColors()!</i>.

   <h3>Deactivated Features</h3>

   Deactivated features are hidden unless the respective choice-menu item is activated. Reasons why
   features are deaktivated:

   <ul>
   <li>Poor alignment with the homologous sequence, where the feature is indirectly inferred from.</li>

   <li>The Sequence positions do not fall into the shown sequence fragment. This is often the case for
   N-terminal signal sequences in Pfam alignments showing only the conserved part.</li>

   <li>Secondary structure when the the structure information is already obtained from the protein
   file. Files with secondary structure information: PDB, Pfam (Stockholm-Format), DSSP.</li>

   <li>Phospho-tyrosine falling on a non-'Y' residue. This is controlled by the third column of the
   customizable table which contains a set of allowed letters or a dash if all letters are allowed
   <i>CUSTOMIZE:SequenceFeatures#customizeColors()!</i> </li>

   </ul>

   <h3>Mouse-control</h3>

   The following mouse actions  apply to the sequence features in the sequence feature table and to the alignment panel.
   <ul>

   <li>Right click opens a context menu for the feature.</li>

   <li>CTRL left-click  selects and deselects the sequence features.</li>

   <li>Dragging the mouse copies sequence features to other proteins or to 3D-views (See  WIKI:Drag_and_drop).</li>

   <li>Double clicking a feature opens a table with detailed data.</li>

   <li>Moving the mouse provides a balloon text with information on the sequence feature under the mouse.</li>

   </ul>

   <h3>Examples</h3>
   The following Swissprot entries have many sequence features and can be used for testing/demonstration:
   SWISS:Q9Y4K3 &nbsp; SWISS:Q9H9D4 &nbsp;  SWISS:Q05516 &nbsp;  SWISS:P28825 &nbsp;  SWISS:P08045 &nbsp;  SWISS:Q13105
   <br>
   However, for many other protein files no sequence features is recorded yet.
   @author Christoph Gille
   <h3>Problems</h3>
   There are rare cases where mapping goes wrong because the sequence of Uniprot entries changed slightly.
   Example: In earlier versions PRC3_HUMAN and UNIPROT:P25787 was lacking the initial Methionine.
   <br><br><u>Related documents</u>
   <ul>
   <li>http://www.uniprot.org/docs/humsavar</li>
   <li>BIOWIKI:GffFormat</li>
   </ul>
*/
public class SequenceFeatures extends javax.swing.table.DefaultTableModel implements Comparator,ChRunnable,ChProcessEvent, TooltipProvider {
    final static int
    MAX_DAYS=33,
        SRC_CSA=1<<1, SRC_PROTEIN_FILE=1<<2, SRC_GFF_EXPASY=1<<3, SRC_DAS=1<<4,
        OPT_ID_BY_HOMOLOGY=1<<5;

    final static String ALI_SCORE=" Alignment-Score=";
    private static String[] featureNames=NO_STRING;
    private final static String TT_default=
        "The sequence features such as \"Phosphorylated Tyrosine\""+
        "will be underlined in the alignment panel.";
    private static ByteArray _log;
    static ByteArray log() {
        if (_log==null) _log=new ByteArray(99);
        return _log;
    }

    /* ---------------------------------------- */
    /* >>> Instance  >>> */
    private final ChEventAdapter LI=new ChEventAdapter(this);
    private static SequenceFeatures _inst;
    public static SequenceFeatures instance() {if (_inst==null) _inst=new SequenceFeatures(); return _inst;}
    static ChButton getButton() { return _inst!=null ? _inst.butDefault : null; }
    /* <<< Instance <<< */
    /* ---------------------------------------- */
    /* >>> Table >>> */
    private final static String longestHeader=ChLabel.PFX_VERTICAL+"3D-backbone";
    public SequenceFeatures() {
        super(
              ("\n"+
              "Feature\n"+
              "#\n"+
              ChLabel.PFX_VERTICAL+"Sequence\n"+
               longestHeader+"\n"+
              ChLabel.PFX_VERTICAL+"PDF\n"+
              ChLabel.PFX_VERTICAL+"HTML\n"+
              ChLabel.PFX_VERTICAL+"Pymol\n"+
              ChLabel.PFX_VERTICAL+"Jalview\n"+
              ChLabel.PFX_VERTICAL+"Web link\n"+
              ChLabel.PFX_VERTICAL+"Go to first\n"+
              ChLabel.PFX_VERTICAL+"To previous\n"+
              ChLabel.PFX_VERTICAL+"Go to next\n"+
               ChLabel.PFX_VERTICAL+"Go to last").split("\n"),0);
    }
    //DefaultTableModel
    @Override public boolean isCellEditable(int row, int col){ return col>=BOOL_START; }
    private ChButton BUT_WEB,BUT_WEB_EDITOR, butExpasy, butDefault;
    private static ChButton butFindUniprot;
    static JComponent butFindUniprot() {
        if (butFindUniprot==null) {
            butFindUniprot=new ChButton()
                .li(instance().LI).t("Find UniProt IDs")
                .tt("Find UniProt identifiers of selected proteins by sequence identity or by PDB ID");
        }
        return butFindUniprot;

    }
    private StrapProtein[] proteins() {
        final StrapAlign a=StrapAlign.getInstance();
        return onlySelected() ? a.getSelectedProteins() : a.getProteins();
    }
    @Override public Object getValueAt(int row, int col){
        final String ft[]=featureNames;
        final String f=ft.length>row ? ft[row] : "error";
        if (col==0) return new Object[]{ChRenderer.KEY_TOGGLE_SELECTED,screenColor(featureName2color(f))};
        if (col==1) return f;
        if (col==BOOL_START+BOOLEANS) {
            for(StrapProtein p : proteins())
                for(ResidueAnnotation a :p.getResidueAnnotations())
                    if (a.getLastValue(CLASS_Hyperrefs)!=null)  return BUT_WEB;
            return null;
        }

        if (NAVI_START<=col && col<NAVI_START+4) {
            final Color color=paperColor(featureName2color(f));
            bbNext[col-NAVI_START].setForeground(color!=null && color!=C(0xFFffFF) ? color : C(0));
            return bbNext[col-NAVI_START];
        }
        final int iB=col-BOOL_START;
        int count=0;
        for(StrapProtein p : proteins()) {
            for(ResidueAnnotation a :p.getResidueAnnotations()) {
                if (a.getFeatureName()!=f) continue;
                count++;
                if (col>=BOOL_START) {
                    if (
                        (ib2vw(iB) &  a.getVisibleWhere())!=0 ||
                        iB==PYMOL &&  a.isEnabled(CLASS_Pymol) ||
                        iB==PDF   && a.isEnabled(CLASS_Texshade)
                        ) return Boolean.TRUE;
                }
            }
        }
        if (col==2) return intObjct(count);
        return Boolean.FALSE;
    }
    private final static int ib2vw(int F) {
        return
            F==WIRE   ? VisibleIn123.STRUCTURE :
            F==HTML   ? VisibleIn123.HTML :
            F==SEQ    ? VisibleIn123.SEQUENCE :
            F==JALVIEW? VisibleIn123.JALVIEW: 0;
    }
    private void getFeaturesAtRow(int rows[], List v) {
        final String ft[]=this.featureNames;
        for(StrapProtein p : proteins()) {
            for(ResidueAnnotation a :p.getResidueAnnotations()) {
                for(int r : rows) {
                    if (ft.length>r && r>=0 && a.getFeatureName()==ft[r]) v.add(a);
                }
            }
        }
    }
    private final static List<ResidueAnnotation.Entry> v_setValueAt=new ArrayList();
    @Override public void setValueAt(Object o, int row, int col){
        final String ft[]=featureNames;
        final int iB=col-BOOL_START;
        if (iB<0 || iB>=BOOLEANS ||  row>=ft.length) return;
        final String fName=ft[row];
        final StrapProtein pp[]=proteins();
        if (pp.length==0) StrapAlign.errorMsg("SequenceFeatures Error: no protein");
        final boolean on=o==Boolean.TRUE;
        for(StrapProtein p : pp) {
            for(ResidueAnnotation a :p.getResidueAnnotations()) {
                if (fName!=a.getFeatureName()) continue;
                if (on) a.setEnabled(true);
                final int vw0=ib2vw(iB);
                for(int i=0;i<2;i++) {
                    final int vw=i==0 ?  vw0  : vw0==VisibleIn123.SEQUENCE ? VisibleIn123.SB : 0;
                    if (vw!=0) a.setVisibleWhere( on ? (a.getVisibleWhere()|vw) : (a.getVisibleWhere() &~ vw));
                }
                final String key=
                    iB==PYMOL ? CLASS_Pymol :
                    iB==PDF ? CLASS_Texshade :
                    null;
                if (key!=null) {
                    for(ResidueAnnotation.Entry e : a.getEntries()) {
                        if (key==e.getKey() && e.isEnabled()) v_setValueAt.add(e);
                    }
                    for(int iV= sze(v_setValueAt); --iV>=0;) v_setValueAt.get(iV).setEnabled(on);
                    v_setValueAt.clear();
                }
            }
        }
        StrapAlign.dispatchLater(StrapEvent.RESIDUE_SELECTION_CHANGED_COLOR,333);
    }
    static void updateTable(int delay) {
        final ChJTable t=_inst!=null ? _inst.jTable : null;
        if (t!=null) { ChDelay.revalidate(t,delay); ChDelay.repaint(t,delay); }
    }
    private int tt_last_row;
    private String tt_last;
    public String getToolTipText(JComponent component, AWTEvent ev) {
        final Point point=point(ev);
        if (point==null) return null;
        if (jTable.columnAtPoint(point)>1) return null;
        final int row=jTable.rowAtPoint(point); // !!!
        if (row==tt_last_row && tt_last!=null) return tt_last;
        final StringBuffer sb=new StringBuffer(999);
        final Set setRemarks=new HashSet();
        final String[] nn=this.featureNames;
        if (row>=0 && row<nn.length) {
            for(StrapProtein p:proteins()) {
                int isFirstA=-1;
                setRemarks.clear();
                for (ResidueAnnotation a:p.getResidueAnnotations()) {
                    if (a.getFeatureName()!=nn[row] || !a.isEnabled() || a.getVisibleWhere()==0) continue;
                    if (isFirstA<0) { isFirstA=sze(sb); sb.append("<br>").append(p).append("  "); }
                    final int from=a.getSelectedAminoacidsOffset();
                    final int to=from+a.getSelectedAminoacids().length;
                    final String attribute=a.getValue("Remark");
                    sb.append(from);
                    if (to-from>1) sb.append('-').append(to-1);
                    if (sze(sb)-isFirstA>110) { sb.append(" ...."); break;}
                    if (sze(attribute)>3 && !setRemarks.contains(attribute)) {
                        setRemarks.add(attribute);
                        htmlEncod(attribute,sb.append("<small>(")).append(")</small>");
                    }
                    sb.append("   ");
                }
            }
        }
        return tt_last=addHtmlTagsAsStrg(sb);
    }
    /* <<< Table <<< */
    /* ---------------------------------------- */
    /* >>> GUI >>> */
    private final static int
    ROW_HEIGHT=3*EX/2,
        SEQ=0, WIRE=1, PDF=2, HTML=3, PYMOL=4, JALVIEW=5,  BOOLEANS=6, BOOL_START=3, NAVI_START=BOOL_START+1+BOOLEANS;
    private final JComponent labMsg=new JLabel(" ");
    private JComponent pnl;
    private ChJTable jTable;
    private ChFrame frame, frameWeb;
    private ChTextArea taWeb;
    private ChButton cbNotAll, cbArrows, bbNext[]=new ChButton[4];
    private ChCombo comboAll;
    private final JProgressBar PRG_BARS[]={new JProgressBar(), new JProgressBar(), new JProgressBar(), new JProgressBar()};
    private final ChRenderer renderer=new ChRenderer();
    static boolean onlySelected() { return _inst!=null && isSelctd(_inst.cbNotAll);}
    public void showPanel() {
        DAS.readRegistry((ChRunnable)this);
        if (pnl==null) {
            for(int i=bbNext.length; --i>=0;) {
                final String
                    t=i==0 ? "|<" : i==1 ? "<" : i==2 ? ">" : ">|",
                    tt="Go to "+ (i==0 ? "first": i==1 ? "previous" : i==2 ? "next" : "last")+" selected amino acid";
                final ChButton b=bbNext[i]=new ChButton(t,ChButton.MAC_TYPE_ICON|ChButton.PAINT_IN_TABLE).li(LI).tt(tt);
                b.cp(ChRenderer.KEY_KEEP_COLOR,"").cp(ChRenderer.KEY_EDITOR_COMPONENT, b.cln());
            }
            BUT_WEB_EDITOR=new ChButton(null,ChButton.NO_FILL|ChButton.NO_BORDER).i(IC_FIREFOX16).li(LI);
            BUT_WEB=new ChButton(null,ChButton.NO_FILL|ChButton.NO_BORDER|ChButton.ICON_SIZE).i(IC_FIREFOX16)
            .cp(ChRenderer.KEY_EDITOR_COMPONENT,BUT_WEB_EDITOR);
            jTable=new ChJTable(this, ChJTable.NO_REORDER);
            jTable.setRowHeight(ROW_HEIGHT);
            LI.addTo("Mkm",jTable);
            addActLi(LI,jTable);
            final Comparator[] comparators=new Comparator[BOOLEANS+BOOL_START+1];
            for(int i=comparators.length; --i>=0;) comparators[i]=jTable;
            ChJTable.setRowSorter(comparators,jTable);
            rtt(jTable);

            jTable.setDefaultRenderer(Object.class,renderer);
            jTable.getTableHeader().setDefaultRenderer(renderer);
            jTable.getTableHeader().setPreferredSize(new ChLabel(longestHeader).getPreferredSize());

            jTable.setDefaultEditor(Object.class,renderer);
            cbArrows=toggl("Show only arrow heads for  long underlinings").s(true).li(LI);
            cbNotAll=toggl("Act not on all but only on selected proteins")
                .tt("Select proteins by left-click the protein-label in the alignment").li(LI);
            comboAll=new ChCombo("Show those and only those features that are activated","Show all features, including those with poor alignment","Hide all features").li(LI)
                .tt("Features with poor alignment to the reference sequence are deactivated but can be activated by the user");
            butExpasy=new ChButton("Expasy").li(LI)
                .tt("GFF files are loaded from Expasy. These files contain lines of the form"+
                    "<pre class=\"data\">P25787	UniProtKB	Modified residue	24	24	.	.	.	Note=Phosphotyrosine</pre>"+
                    "As long as one job is runnting the button remains deactivated.<br>"+
                    "The process can be interrupt by closing this frame."
                    );
            butDefault=new ChButton("Load sequence features of "+StrapAlign.LABEL_ALL_OR_SELECTED_P).li(LI).tt(TT_default);
            StrapAlign.setNumOfProteins(butDefault);
            //for(JComponent c : PRG_BARS) if (c!=null) c.setPreferredSize(dim(0,0));
            final JComponent
                hb=pnl(Customize.newButton(customizeColors()).options(ChButton.NO_FILL|ChButton.NO_BORDER|ChButton.ICON_SIZE).magnifyUnderMouse(IC_CUSTOM), panHS(SequenceFeatures.class)),
              panAboveT=vB(hB("#", cbNotAll.cb()),hB("#",C(0xFFffFF), comboAll)),

                pTop=newCNSEW(pnl("<h2>Sequence Features</h2>"),null,panAboveT, hb,pnl(new ChButton("Log").li(LI))),
                b1=new ChButton("FROM_FILE").li(LI).t("Protein files")
                .tt("Some protein files contain sequence features lines like <pre class=\"data\">FT   METAL       141    141       Copper.\n\nSITE     1 CAT  3 ASP A  32  HIS A  64  SER A 221\n\nMODRES 1IVO ASN A   32  ASN  GLYCOSYLATION</pre>"),
                b3=new ChButton("FROM_DAS").li(LI).t("BioDAS...")
                .tt("The user can chose from a large number of BioDAS Servers"),
                pMoreAdvanced=vB(
                                 pnl(monospc("Spice-Browser* on selected protein"),new ChButton("SPICE").li(LI).t(ChButton.GO)),
                                 pnl(monospc("Dasty on selected protein"),new ChButton("DASTY").li(LI).t(ChButton.GO))
                                 ),
                pLoadFrom=vBpnl(pnl("Load from ",b1,butExpasy,b3, "#", ChButton.actionShowResults().t("Log blast")), cbArrows.cb()),
                pLoadDefault=vBhB(TT_default,
                                  butDefault,WATCH_MOVIE+MOVIE_Sequence_Features_in_3D),
                pAdvanced=vB(
                             hBL("Other options ",toggl().actionCollapse(false,pMoreAdvanced).actionCollapse(true,pLoadFrom),"#",pnl(pLoadFrom)),
                             pMoreAdvanced
                             ),
                togA=toggl().actionCollapse(false,new Object[]{ pAdvanced, comboAll}).actionCollapse(true, new Object[]{butDefault,pLoadDefault}),
                pSouth0=vB( hB(pLoadDefault,"#","Options ", togA),pAdvanced ),
                pSouth=newCNSEW(pSouth0, vB(PRG_BARS),labMsg,null,null);

            repaintOnMouseEnterAndExit(false,labMsg);
            pnl=newCNSEW(scrllpn(SCRLLPN_SEARCH_BUTTON,jTable),pTop,pSouth, null,null);
            instance().LI.addTo("Mkm",pnl);
            frame=new ChFrame("feature").ad(pnl).size(EM*70,EX*30).shw(CLOSE_ON_CtrlW|ChFrame.PACK);

            for(int i=NAVI_START+4;--i>=0;) {
                if (i==1) continue;
                if (i>=NAVI_START) jTable.setColumnWidthC(false, i,bbNext[0]);
                else jTable.setColumnWidth(false, i, i==2 ? 5*EM : EX*2);

            }
            enlargeFont(jTable.getTableHeader(),.8);
        }
        setSettings(instance(),customizeCSS_IE(), customizeColors());
        frame.shw(ChFrame.CENTER);
    }
    /* <<< Instance <<< */
    /* ---------------------------------------- */
    /* >>> Sort >>> */
    public int compare(Object o1, Object o2) {
        if (o1 instanceof ResidueAnnotation && o2 instanceof ResidueAnnotation) {
            final ResidueAnnotation a1=(ResidueAnnotation)o1, a2=(ResidueAnnotation)o2;
            final String f1=a1.getFeatureName(), f2=a2.getFeatureName();
            if (f1!=null && f2!=null) {
                final int diffLen=a1.getSelectedAminoacids().length-a2.getSelectedAminoacids().length;
                if (diffLen!=0) return -diffLen;
                final int diffOff=a1.getSelectedAminoacidsOffset()-a2.getSelectedAminoacidsOffset();
                if (diffOff!=0) return -diffOff;
                final int diffName=f1.compareTo(f2);
                if (diffName!=0) return diffName;
            }
        }
        return 0;
    }
    /* <<< Sort <<< */
    /* ---------------------------------------- */
    /* >>> AWTEvent >>> */
    private int lastRow;
    private final Map<String,NextAndPrevSelection> mapNextPrev=new HashMap();
    private final static List<ResidueAnnotation> vTMP=new ArrayList();
    public AWTEvent processEvent(JComponent component, AWTEvent ev) {
        final int id=ev.getID(), modi=modifiers(ev);
        final String cmd=actionCommand(ev);
        final Object q=ev.getSource();
        final StrapAlign align=StrapAlign.getInstance();
        final StrapProtein ppSel[]=align.getSelectedProteins();
        final StrapView view=align.getAlignmentPanelWithFocus();
        vTMP.clear();

        if (q==butFindUniprot) {
            setEnabld(false,butFindUniprot);
            ChDelay.newThread(this,"FIND_UP", ppSel.length>0?ppSel:align.getProteins()).start();
        }
        if (cmd=="<" || cmd==">" || cmd=="|<" || cmd==">|") {
            final int row=ChJTable.rowIndexToModel(renderer.editedRow(),jTable);
            final String name=get(row,featureNames, String.class);
            if (name!=null) {
                NextAndPrevSelection nextPrev=mapNextPrev.get(name);
                if (nextPrev==null) mapNextPrev.put(name,nextPrev=new NextAndPrevSelection());
                vTMP.clear();
                getFeaturesAtRow(new int[]{row}, vTMP);
                final ResidueSelection[]
                    ss=toArry(vTMP,ResidueAnnotation.class),
                    ss2=cbNotAll.s() ? ResSelUtils.onlyThoseWithProteins(true,ss, ppSel) : ss;
                nextPrev.goNextMatch(cmd,ss2);
            }
        }
        if (component!=null && component==jTable && cmd==null) {
            final int row=y(ev)/ROW_HEIGHT;
            if (isPopupTrigger(ev)) {
                if (0<=row && row<getRowCount()) {
                    final int selected[]=ChJTable.selectedRowsConverted(jTable);
                    getFeaturesAtRow(idxOf(row, selected,0,MAX_INT)>=0 ? selected : new int[]{row} ,(List)vTMP);
                    StrapAlign.showContextMenu('F', toArry(vTMP,ResidueAnnotation.class));
                }
            }
            if (id==MOUSE_MOVED && view!=null && lastRow!=row) {
                getFeaturesAtRow(new int[]{row},vTMP);
                final ResidueAnnotation ss[]=toArry(vTMP,ResidueAnnotation.class);
                final StrapProtein pp[]=align.getVisibleProteins();
                final int iiP[]=new int[ss.length];
                Color color=null;
                for(int i=ss.length; --i>=0;) {
                    iiP[i]=idxOf(sp(ss[i]),pp);
                    if (color==null) color=ss[i].getColor();
                }
                view.highlightProteins(iiP,new Color[]{color});
                lastRow=row;
            }
            if (id==MOUSE_DRAGGED && !isCellEditable(row, jTable.columnAtPoint(point(ev)))) {
                final List v=getDndFilesV(jTable);
                v.clear();
                getFeaturesAtRow(new int[]{row},v);
                exportDrg(jTable, ev, TransferHandler.COPY);
            }
        }
        if (id==MOUSE_EXITED) lastRow=-1;
        if (id==MOUSE_ENTERED && view!=null) { lastRow=-1;  view.highlightProteins(null,null);}
        if (cmd!=null) {
            StrapAlign.errorMsg("");
            if (cmd=="SPICE" || cmd=="DASTY") {
                if (ppSel.length!=1) error("For Spice-Browser* or Dusty please select exactly one protein");
                else {
                    if (cmd=="SPICE") toSpiceViewer(ppSel[0]);
                    if (cmd=="DASTY") {
                        final String uniprot=ppSel[0].getUniprotID();
                        if (uniprot==null) error("No uniprot ID associated with "+ppSel[0]);
                        else visitURL(Hyperrefs.toUrlString("DASTY:"+uniprot.substring(uniprot.indexOf(':')+1),Hyperrefs.WEB_LINK), modi);
                    }
                }
            }
            if (q==jTable && cmd==ACTION_SELECTION_CHANGED) {
                StrapAlign.getSelectedObjectsV().clear();
                getFeaturesAtRow(ChJTable.selectedRowsConverted(jTable), StrapAlign.getSelectedObjectsV());
                StrapAlign.dispatchLater(StrapEvent.OBJECTS_SELECTED,11);
            }
            if (q==frame && cmd==ACTION_WINDOW_CLOSING) setEnabld(true,butExpasy);
            if (q==cbNotAll || q==comboAll || q==cbArrows) {
                if (q==cbArrows) ResidueAnnotation.setMaxLengthOfUnderlining(cbArrows.s() ?6:99999);
                Protein.incrementModificationCount(ProteinModiCount.MCA_SEQUENCE_FEATURES_V,null);
                StrapAlign.dispatchLater(StrapEvent.RESIDUE_SELECTION_CHANGED,111);
            }
            final boolean fromFile=cmd=="FROM_FILE" || q==butDefault;
            if (fromFile || q==butExpasy || q==butDefault) {
                final StrapProtein pp[]=ppSel.length>0 ? ppSel: align.getVisibleProteins();
                if (fromFile) ChDelay.newThread(this,"VAR", ExpasyGff.readFeaturesInProteinFile(null,MAX_INT, pp) ).start();
                final long opts=
                    q==butExpasy  ? SRC_GFF_EXPASY :
                    q==butDefault ? SRC_GFF_EXPASY | SRC_DAS | SRC_CSA :
                    0;
                if (opts!=0) {
                    if (0!=(opts&(SRC_GFF_EXPASY|SRC_DAS))) {
                        thread_loadFeaturesBG(opts| OPT_ID_BY_HOMOLOGY,pp, null).start();
                    }
                    thread_loadFeaturesBG(opts&~OPT_ID_BY_HOMOLOGY,pp, thread_setEnabld(true,q)).start();
                    setEnabld(false,q);

                }
            }
            if (cmd=="FROM_DAS") StrapDAS.show(true);
            if (cmd=="Log") new AsciiView(log()).showInFrame("Log"); /* TODO update problem */
            if (q instanceof Customize && q==customizeColors()) {
                mapFeatureColors.clear();
                customizeCSS_IE=null;
                StrapAlign.dispatchLater(StrapEvent.RESIDUE_SELECTION_CHANGED_COLOR,999);
                ChDelay.repaint(this,999);
            }
            if (q==BUT_WEB_EDITOR) {
                final int row=ChJTable.rowIndexToModel(renderer.editedRow(),jTable);
                final String[] nn=featureNames;
                if (row<0 || row>=nn.length) return null;
                if (frameWeb==null) frameWeb=new ChFrame("feature URLs").ad(new JScrollPane(taWeb=new ChTextArea(""))).shw(CLOSE_ON_CtrlW);
                final Set vURLs=new HashSet();
                int max=0;
                final StringBuffer sb=new StringBuffer().append(nn[row]).append('\n');
                for(int pass=2; --pass>=0;) {
                    for(StrapProtein p:proteins()) {
                        nextFeature:
                        for (ResidueAnnotation a:p.getResidueAnnotations()) {
                            if (a.getFeatureName()!=nn[row] || !a.isEnabled() || a.getVisibleWhere()==0) continue;
                            final int L=sze(p.getName());
                            if (pass>0) { max=maxi(max,L); continue;}
                            boolean firstUrl=true;
                            for(ResidueAnnotation.Entry e : a.getEntries()) {
                                if (CLASS_Hyperrefs!=e.getKey()) continue;
                                final String url=e.getValue();
                                if (firstUrl) {
                                    if (vURLs.contains(url)) continue nextFeature;
                                    vURLs.add(url);
                                    for(int j=max-L+1; --j>=0;) sb.append(' ');
                                    sb.append(sp(a)).append(' ');
                                    firstUrl=false;
                                }
                                sb.append(' ').append(url);
                            }
                            sb.append('\n');
                        }
                    }
                }
                taWeb.t(sb).tools().underlineRefs(0);
                frameWeb.shw(0);
            }
        }
        return ev;
    }
    /* <<< AWTEvent <<< */
    /* ---------------------------------------- */
    /* >>> Color >>> */
    private final static Map<String,FeatureColor> mapFeatureColors=new HashMap();
    public static void cssColors(StrapProtein pp[], StringBuffer sb) {
        final Map<String,Color> map=new HashMap();
        for(StrapProtein p:pp) {
            for(ResidueAnnotation a:p.getResidueAnnotations()) {
                final String n=a.getFeatureName();
                if (n!=null) {
                    map.put(n,a.getColorForWhiteBG());
                }
            }
        }
        for(Map.Entry<String,Color> e : entryArry(map)) {
            final Color color=e.getValue();
            if (color==null) continue;
            final int rgb=color.getRGB();
            sb.append('.').append(e.getKey()).append(" { background-color: #");
            toHex(rgb,6,sb).append("; color: #");
            toHex(rgb,6,sb).append("; } \n");
        }
    }
    private static Customize customizeColors, customizeCSS_IE;
    public static Customize customizeCSS_IE() {
        if (customizeCSS_IE==null) {
            parseColors();
            final Object oo[]=toArry(new HashSet(mapFeatureColors.values()),Object.class);
            StringBuffer sbIE=null;
            for(int i=oo.length;--i>=0;) {
                final FeatureColor fc=(FeatureColor)oo[i];
                final String cssClass=ResidueAnnotation.toCssClass(fc.name);
                if (sbIE==null) sbIE=new StringBuffer(999).append("PRE { margin: 0mm;}\n PRE A, ");
                else sbIE.append(", ");
                sbIE.append(" .").append(cssClass);
            }
            sbIE.append(" { background-color: transparent;  } ");
            final String hlpIE="Workaround for IE bugs.";
            customizeCSS_IE=new Customize("cssInternetExplorer",new String[]{"/* "+hlpIE+" */",sbIE.toString()},hlpIE,Customize.LIST);
        }
        return customizeCSS_IE;
    }
    public static Customize customizeColors() {
        if (customizeColors==null) {
            final String
                help="# WIKI:Web_colors of features for black and white background.",
                rsc="charite/christo/strap/SequenceFeatures.rsc",
                txt0[]=readLines(CLASSLOADER.getResourceAsStream(rsc)),
                txt[]=txt0!=null ? txt0 : new String[]{"#Error not found: "+rsc};
            addActLi(instance().LI,customizeColors=new Customize("featureColors",txt,help,Customize.LIST|Customize.FLAG_NOTIFY_IMMEDIATELY));
        }
        return customizeColors;
    }
    static class FeatureColor {
        String name,nameUS,  screen, paper, rule="-";
        Color cS, cP;
        FeatureColor(String n) { name=n;}
    }
    private static Color screenColor(FeatureColor fc) {
        if (fc.cS==null) {
            final long rgb=HtmlUtil.hexToInt(fc.screen,0);
            fc.cS=rgb>=0 ? new Color((int)rgb) : C(0xFFffFF);
        }
        return fc.cS;
    }
    private static Color paperColor(FeatureColor fc) {
        if (fc.cP==null) {
            final long rgb=HtmlUtil.hexToInt(fc.paper,0);
            fc.cP=rgb>=0 ? new Color((int)rgb) : C(0x808080);
        }
        return fc.cP;
    }
    private static String nameUS(FeatureColor fc) {
        if (fc.nameUS==null && fc.name!=null) fc.nameUS=fc.name.replace(' ','_').replace('-','_');
        return fc.nameUS;
    }
    public static FeatureColor featureName2color(String f) {
        parseColors();
        FeatureColor fc=mapFeatureColors.get(f);
        if (fc==null) {
            mapFeatureColors.put(f,fc=new FeatureColor(f));
            mapFeatureColors.put(nameUS(fc),fc);
        }
        return fc;
    }
    private static void parseColors() {
        if (sze(mapFeatureColors)==0) {
            StringBuffer err=null;
            for(String s :customizeColors().getSettings()) {
                final StringTokenizer st=new StringTokenizer(s);
                final String fn=st.hasMoreTokens() ? st.nextToken().replace('_',' ') : null;
                if (fn==null || fn.startsWith("#")) continue;

                final String screen=st.hasMoreTokens() ? st.nextToken() : null;
                if (chrAt(0,screen)!='#') continue;
                final FeatureColor fc=new FeatureColor(fn);
                fc.screen=screen;
                fc.paper=st.hasMoreTokens() ? st.nextToken() : null;
                if (chrAt(0,fc.paper)!='#') continue;
                fc.rule=st.hasMoreTokens() ? st.nextToken().intern() : "-";
                mapFeatureColors.put(fn,fc);
                mapFeatureColors.put(nameUS(fc),fc);

            }
            error(err);
        }
    }
    static void setFN(String n, int src, ResidueAnnotation a) {
        if (a!=null && n!=null) {
            a.setFeatureName(n,src);
            final FeatureColor featureC=featureName2color(n);
            a.setColor(screenColor(featureC));
            a.setColorForWhiteBG(paperColor(featureC));
        }
    }
    /* <<< Color <<< */
    /* ---------------------------------------- */
    /* >>> Layout Underlining >>> */
public final static int  F_NO_SEC_STRU=1, F_SEC_STRU=2;
    private final static Set<String> setFeatureNames=new HashSet();
    static int layoutFeatures(StrapProtein pp[]) {
        assertEDT();
        int maxLine=0, max_nR=0;
        for(StrapProtein p : pp) max_nR=maxi(max_nR,p.countResidues());
        setFeatureNames.clear();
        final int
            features[]=new int[max_nR],
            SELECTED=0, ALL=1, NONE=2,
            activate=_inst!=null && _inst.comboAll!=null ? _inst.comboAll.i() : SELECTED;
        ResidueAnnotation.setEnabledAll(activate==ALL ? Boolean.TRUE : activate==NONE ? Boolean.FALSE : null);
        for(StrapProtein p : pp) {
            Arrays.fill(features,0);
            final ResidueAnnotation[] aa=p.getResidueAnnotations();
            if (cntains(null,aa)) putln(RED_ERROR,"SequenceFeatures aa contains null");
            ResidueAnnotation[] ff=null;
            while(true) {
                int countF=0;
                for(ResidueAnnotation a : aa) {
                    if (a==null) {
                        putln(RED_ERROR,"SequenceFeatures a==null");
                        continue;
                    }
                    final String f=a.getFeatureName();
                    if (f==null || activate==NONE ||activate!=ALL && !a.isEnabled()) continue;
                    if (ff!=null) ff[countF]=a;
                    countF++;
                }
                if (ff==null) ff=new ResidueAnnotation[countF];
                else break;
            }
            sortArry(ff,instance());
            String lastFn=null;
            int lastFrom=0,lastTo=0;
            for(ResidueAnnotation a : ff) {
                if (a==null) continue;
                final int from=a.getSelectedAminoacidsOffset();
                final boolean sel[]=ResidueAnnotation.getSelectedAminoacidsDisplayed(a);
                final int to=mini(features.length,from+sel.length );
                final String fn=a.getFeatureName();
                if (fn==null) continue;
                if (fn==lastFn && lastFrom==from && lastTo==to) {
                    a.setStyle(VisibleIn123.HIDDEN);
                    continue;
                }
                lastFn=fn;
                lastFrom=from;
                lastTo=to;
                int foundLine=-1;
                nextLine:
                for(int iLine=0; iLine<32; iLine++) {
                    final int bitMask=1<<iLine;
                    for(int iA=from; iA<to; iA++) {
                        if (iA<0 || iA<from || iA<features.length && sel[iA-from] && (features[iA]&bitMask)!=0)
                            continue nextLine;
                    }
                    foundLine=iLine;
                    for(int iA=from; iA<to; iA++) features[iA]|=bitMask;
                    break;
                }
                if (foundLine>=0) {
                    a.setStyle(VisibleIn123.LINES+ (foundLine<<VisibleIn123.BIT_SHIFT_LINE));
                    setFeatureNames.add(fn);
                    if (maxLine<foundLine) maxLine=foundLine;
                }
                setFeatureNames.add(fn);
            }
        }
        updateTable(33);
        featureNames=strgArry(setFeatureNames);
        instance().setNumRows(featureNames.length);
        return maxLine+1;
    }
    /* <<< Layout Underlining <<< */
    /* ---------------------------------------- */
    /* >>> Threading  >>> */
    public Object run(String id, Object arg) {
        final Object argv[]=arg instanceof Object[] ? (Object[]) arg : null;
        if (id=="FIND_UP") {
            StrapProtein pp[]=((StrapProtein[])arg).clone();
            for(int i=pp.length; --i>=0;) {
                if (pp[i]==null || pp[i].getUniprotID()!=null || sze(FindUniprotID.ids(false,pp[i]))>0) pp[i]=null;
            }
            pp=rmNullA(pp, StrapProtein.class);
            StrapDAS.pdb2uniprotP(pp, SequenceFeatures.instance());
            FindUniprotID.load(pp, SequenceFeatures.instance());
            final String title="Assigned uniprot IDs";
            final ByteArray sb=new ByteArray(999).append(title).append("\n\n");
            for(StrapProtein p:pp) {
                if (p==null) continue;
                sb.append(p).append("  getUniprotID=").append(p.getUniprotID())
                    .append("  by-sequence=").join(FindUniprotID.ids(false,p)," ")
                    .append('\n');
            }
            new AsciiView(sb).showInFrame(title);
            setEnabld(true,butFindUniprot);
        }

        if (id==ChRunnable.PROGRESS) {
            final Object src=argv[0];
            final int iB=
                src==FindUniprotID.class ? 0 :
                src==CsaWeb.class ? 1 :
                src==ExpasyGff.class ? 2 :
                src==StrapDAS.class ? 3 :
                -1;
            if (iB>=0) {
                setJPrgrss(PRG_BARS[iB],argv);
                setPrefSze(true, EX,EX*3/2, PRG_BARS[iB]);
            }
        }
        if (id=="LOAD") {
            final long type=atol(argv[0]), opts=atol(argv[1]);
            final String title[]=(String[])argv[2];
            final Protein pp[]=(Protein[])argv[3];
            final Runnable run=(Runnable)argv[4];
            final ChRunnable log=instance();
            if (type==SRC_DAS)         StrapDAS.load(opts,title,pp, log);
            if (type==SRC_CSA)           CsaWeb.load(opts,      pp, log);
            if (type==SRC_GFF_EXPASY) ExpasyGff.load(opts,      pp, log);
            inEDT(run);
            StrapAlign.dispatchLater(StrapEvent.RESIDUE_SELECTION_ADDED, 555);
        }
        if (id==ChRunnable.APPEND) drawMsg(toStrg(arg), labMsg);
        if (id=="VAR") {
            ByteArray ba=null;
            final ResidueAnnotation aa[]=(ResidueAnnotation[])arg;
            for(ResidueAnnotation a : aa ) {
                for(ResidueAnnotation.Entry e : a.getEntries()) {
                    if ("Remark"!=e.getKey()) continue;
                    final String v=e.getValue();
                    if (v.startsWith("FTId=VAR")) {
                        final String u=Hyperrefs.toUrlString(v,Hyperrefs.WEB_LINK);
                        final java.io.File f=download_TS(url(u), CacheResult.isEnabled()?MAX_DAYS:0);
                        ba=readBytes(f,ba);
                        if (ba!=null) {
                            int i=strstr("<td>Disease",ba);
                            if (i<0) i=strstr("<td>Polymorphism",ba);
                            if (i<0) i=strstr("<td>Unclassified",ba);
                            if (i>0) {
                                final String type=ChUtils.wordAt(i+4, (CharSequence)ba, LETTR).toUpperCase();
                                final char c=chrAt(0,type);
                                setFN(c=='U' ? type+"_"+type : type, SRC_PROTEIN_FILE, a);
                                if (idxOfStrg(type,this.featureNames)<0) featureNames=ad(type,featureNames,0);
                                updateTable(999);
                            }
                            final int from=strstr(STRSTR_AFTER,"<td><b>From ",ba);
                            if (from>0) {
                                final int colon=strstr(",",ba,from, mini(ba.length(),from+30));
                                if (colon>0) a.setValue(ResidueAnnotation.NAME,wordAt(colon+2,ba,LETTR_DIGT));
                            }
                        }
                    }
                }
            }
        }
        return null;
    }

    public static Thread thread_loadFeaturesBG(long options, Protein pp[], Runnable run) {
        return threadM("loadFeaturesBG", SequenceFeatures.class, new Object[]{longObjct(options), pp, run});
    }
    public static void loadFeaturesBG(long options, Protein pp[], Runnable run) {
        StrapDAS.pdb2uniprotP(pp, instance());

        final boolean byHomology=0!=(OPT_ID_BY_HOMOLOGY&options);
        if (byHomology) {
            Protein[] ppNoID=new Protein[pp.length];
            for(int iP=pp.length; --iP>=0;) {
                if (pp[iP]!=null && pp[iP].getUniprotID()==null) ppNoID[iP]=pp[iP];
            }
            ppNoID=rmNullA(ppNoID, Protein.class);
            FindUniprotID.load(ppNoID, instance());
        }

        for(int type : new int[]{SRC_CSA, SRC_DAS, SRC_GFF_EXPASY}) {
            if (0==(type&options)) continue;
            final Object arg[]={longObjct(type), longObjct(options), DAS.IMPORTANT_FEATURES, pp, run};
            ChDelay.newThread(instance(), "LOAD", arg).start();
        }
    }
    // longObjct(pass>0?options:(options&~OPT_ID_BY_HOMOLOGY))
    public static Thread thread_loadFeatures(long options, Protein pp[]) {
        return threadM("loadFeatures",SequenceFeatures.class, new Object[]{longObjct(options),pp});
    }

    public static void loadFeatures(long options, Protein pp[]) {

    }
    /* <<< Threading <<< */
    /* ---------------------------------------- */
    /* >>>  Score >>> */
    static void keepFeaturesWithBestScore(StrapProtein p) {
        final ResidueAnnotation[] aa=p.getResidueAnnotations();
        log().append("\n --- "+p+": Keep only features whith max alignment score --- \n");
        for(int i0=aa.length; --i0>=0;) {
            final ResidueAnnotation a0=aa[i0];
            if (a0==null) continue;
            final float score0=a0.getFeatureAlignmentScore();
            final String name=a0.getFeatureName();
            //final boolean isSecStru=name=="Turn" || name=="Helix" || name=="Beta_strand";
            final boolean[] sel0=a0.getSelectedAminoacids();
            final int from0=a0.getSelectedAminoacidsOffset(), to0=from0+lastTrue(sel0);
            final int where0=a0.getWhereFeatureLoadedFrom();
            for(int i=aa.length; --i>=0;) {
                final ResidueAnnotation a=aa[i];
                if (a==null || a==a0 || a.getFeatureName()!=name) continue;
                final int where=a.getWhereFeatureLoadedFrom();
                final float score=a.getFeatureAlignmentScore();
                final boolean[] sel=a.getSelectedAminoacids();
                if (where0==where && score0<=score) continue;
                final int from=a.getSelectedAminoacidsOffset(), to=from+lastTrue(sel);
                if (/* !isSecStru && */ from==from0 && to==to0 ||
                    from0<=from && from<to0 || from0<=to && to<=to0 ||  from<=from0 && from0<=to   || from<=to0 && to0<=to) {
                    log().append("   DELETE: "+name+" "+score+"  ["+from+"-"+to+"]   BECAUSE  OF "+score0+"  ["+from0+"-"+to0+"]\n");
                    p.removeResidueSelection(a);
                    p.addInvalidFeatureLine(a.getSrcText());
                    aa[i]=null;
                }
            }
        }
    }
    /* <<< Score <<< */
    /* ---------------------------------------- */
    /* >>> Enable / Disable >>> */
    static boolean isDeactivated(String s) { return s=="Sequence_conflict" || s=="mature_protein_region"; }
    static void disableSecondaryStructure(ResidueAnnotation[] aa) {
        for(ResidueAnnotation a : aa) {
            final String n=a!=null ? a.getFeatureName() : null;
            final StrapProtein p=sp(a);
            boolean changed=false;
            if (n!=null && a.isEnabled() && p!=null && p.getResidueSecStrType()!=null) {
                if ("helix".equalsIgnoreCase(n) || "beta_strand".equalsIgnoreCase(n) || "strand".equalsIgnoreCase(n) || "turn".equalsIgnoreCase(n)) {
                    a.setEnabled(false);
                    a.addEnS("Disabled","Secondary structure already knwown");
                    changed=true;
                }
            }
            if  (changed) {
                StrapAlign.dispatchLater(StrapEvent.RESIDUE_SELECTION_CHANGED_COLOR, 999);
            }
        }
    }
    /* <<< Enable / Disable <<< */
    /* ---------------------------------------- */
    /* >>> Translations >>> */
    private final static String[][] RPLC_FEATURE={
        {"GLYCOSYLATION SITE","Glycosylation"},
        {"CARBOHYD","Glycosylation"},
        {"ACT_SITE","Active_site"},
        {"Catalytic residue","Active_site"},
        {"BINDING","Binding_site"},
        {"COMPBIAS","Compositional_bias"},
        {"DISULFID","Disulfide_bond"},
        {"alpha-helix","Helix"},
        {"INIT_MET","Initiator_methionine"},
        {"METAL","Metal_binding"},
        {"Metal contact","Metal_binding"},
        {"MOD_RES","Modified_residue"},
        {"NP_BIND","Nucleotide_binding"},
        {"PROPEP","Propeptide"},
        {"REGION","Region"},
        {"SIGNAL","Signal_peptide"},
        {"STRAND","Beta_strand"},
        {"TOPO_DOM","Domain"},
        {"DOMAIN","Domain"},
        {"Polypeptide_domain","Domain"},
        {"TRANSMEM","Transmembrane"},
        {"Non cytoplasmic region","Non-cytoplasmic"},
        {"Natural_variant_site","Natural_variant"},
        {"protein protein contact","Protein_contact"},
        {"CONFLICT","Sequence_conflict"},

        {"?","Phosphoserine"},
        {"?","Phosphotyrosine"},
        {"?","Phosphothreonine"},
        {"?","Selenomethionine"},
        {"?","Alternative_sequence_site"},
        {"?","Mutated_variant_site"},
        {"?","Theoretical_tryptic_peptide"},
        {"?","Tryptic_peptide"},
        {"?","Solvent_accessibility"},
        // {"?","Cytoplasmic region"},
        // {"?","Low_complexity"},
        // {"?","Turn"},
        {"?","Peptide"},
        {"?","Swissmodel"},
        {"?","Polypeptide_conserved_region"},
        {"?","Merops_feature"},
        {"?","Merops_identifier"},
        {"?","Merops_clan"},
        {"?","Merops_family"},
        {"?","Variant"},
        {"","Repeat"},
        {"","Polypeptide_structural_domain"},
    };
    private static boolean equalsLetter(String s1, String s2) {
        if (s1==null || s2==null || s1.length()!=s2.length()) return false;
        for(int i=s1.length(); --i>=0;) {
            final char  c1=s1.charAt(i), c2=s2.charAt(i);
            if (is(LETTR,c1)!=is(LETTR,c2)) return false;
            if (is(LETTR,c1) && (c1|32) != (c2|32)) return false;
        }
        return true;
    }

    static void addFeature(ResidueAnnotation a, StrapProtein p) {
        if (a==null || p==null) return;
        final boolean bb[]=a.getSelectedAminoacids();
        if (countTrue(bb)>5) a.setVisibleWhere(a.getVisibleWhere()&~VisibleIn123.STRUCTURE);
        p.addResidueSelection(a);
    }
    static String translateFeatureName(String f, boolean found[]) {
        //  NON_TER CROSSLNK DOMAIN MUTAGEN  SIMILAR SITE   VAR_SEQ
        if (f==null) return null;
        String ret=null;
        if (f.indexOf("phosphorylated")>=0) {
            ret=
                f.indexOf("tyrosine")>=0 ?  "Phosphotyrosine" :
                f.indexOf("threonine")>=0 ? "Phosphothreonine":
                f.indexOf("serine")>=0 ? "Phosphoserine" :
                // id="UNIPROTKB_P56817_DISULFID_330_380" label="UNIPROTKB_P56817_DISULFID_330_380">
                null;
        }
        if (ret==null && (f.indexOf("glycosylat")>=0 || f.indexOf("glycosolat")>=0) ) ret="Glycosylation";
        //        if (ret==null && f.indexOf("Swissmodel")>=0) ret="Swissmodel";
        for(String[] rplc : RPLC_FEATURE) {
            if (ret!=null) break;
            if (equalsLetter(rplc[0],f) || equalsLetter(rplc[1],f)) ret=rplc[1];
        }
        if (ret!=null && found!=null) found[0]=true;
        return ret!=null? ret : f;
    }
    public static String toShortFeatureName(String fn) {
        if (sze(fn)<4) return fn;
        final int space=fn.indexOf(' ');
        return space<0 || space==sze(fn)-1 ?  fn.substring(0,1) :  fn.substring(0,1)+fn.substring(space+1,space+2);
    }
    /* <<< Translations <<< */
    /* ---------------------------------------- */
    /* >>> Spice >>> */
    public static void toSpiceViewer(StrapProtein p) {
        if (p==null) return;
        final StringBuffer sb=new StringBuffer("http://www.dasregistry.org/runspice.jsp?");
        final String pdb=p.getPdbID(), uniprot=p.getUniprotID();

        if (pdb!=null) sb.append("pdb=").append(pdb.substring(pdb.indexOf(':')+1));
        else if (uniprot!=null) sb.append("uniprot=").append(uniprot.substring(uniprot.indexOf(':')+1));
        else {
            error("Error: cannot run Spice on "+p+"\nbecause it has neither a PDB ID nor a Uniprot ID");
            return;
        }
        visitURL(sb.append(VISIT_URL_JAVAWS),0);
    }
}
