package charite.christo.strap;
import charite.christo.*;
import charite.christo.interfaces.*;
import charite.christo.protein.*;
import charite.christo.strap.interfaces.*;
import java.io.*;
import java.awt.*;
import java.util.*;
import java.util.List;
import static charite.christo.protein.ProteinModiCount.*;
import static charite.christo.ChUtils.*;
import static charite.christo.ChConstants.*;
/**
   The class StrapProtein describes the protein objects in Strap and is important for users who write plugins for STRAP.
   It is an extension of <i>JAVADOC:ProteinNT</i>  because it allows proteins to be read
   from nucleotide files.
   In addition it allows gaps which is the basis for sequence alignments.
   <h3>get-methods in StrapProtein</h3>
   Methods starting with the prefix getResidue return information
   associated with amino acids.
   They come in two versions.
   Without parameter they return an array with at least as many elements as there are amino acids
   or null if the information is not available.
   With a residue index as parameter they return the information for the particular residue.
   If the information is not available -1 or NaN or null is returned.
   <h3>What you should never do</h3>
   The java-language allows to change elements within arrays but you
   should not change elements in arrays which are obtained by using a get method.
   To change information one should always use the appropriate setters
   such as <i>JAVADOC:Protein#setResidueType(byte[])</i>.
   Never use the field .length of arrays  to obtain the number of amino acids.
   Use the method  <i>JAVADOC:Protein#countResidues()</i>  instead.
   <h3>Here are some frequently used get-methods:</h3>
   <ul>
   <li><i>JAVADOC:Protein#countResidues()</i> number of amino acids in the protein.</li>
   <li><i>JAVADOC:Protein#getResidueType()</i> returns the residue letters. This array of bytes contains one letter for each residue. It may have more elements than the protein has amino acids. The superfluous bytes are invalid.</li>
   <li><i>JAVADOC:Protein#getResidueSecStrType()</i> returns a byte array representing the
   secondary structure type.</li>
   <li><i>JAVADOC:Protein#getResidueCalpha()</i> provides the c-alpha coordinates</li>
   <li><i>JAVADOC:StrapProtein#getResidueColumn()</i> returns the column (horizontal positions) in the alignment.</li>
   <li><i>JAVADOC:StrapProtein#setResidueGap(int[])</i> Sets the number of white space before a white space.</li>
   </ul>
   <h3>Caching information:</h3>
   Since STRAP allows to work with hundreds of proteins simultaneously
   computation must be kept to a minimum.
   Computing one and the same several times must be avoided.  Each
   method has an input and a return value.
   When a method is called a second time with the same input it can return the previous result.
   Therefore a computational intensive  method should have a memory where the last result is kept.
   If the input is an array it would be difficult to test whether all
   elements in this array are still the same. For that reason  methods with the suffix modificationCount
   return a version number of the information.

   The modification count is a non-negative value that increases (by 1 or perhaps by more) as this return value changed.
   The initial value is unspecified.

   This version number is incremented when elements of the array change.

   A method simply needs to determine whether the  modificationCount has changed since it was called last time.

   For example a method which depends on the amino acid sequence obtained with <i>JAVADOC:Protein#countResidues()</i>
   can retrieve the version with <i>JAVADOC:Protein#getResidueType()</i>.
   Depending on whether this version number changed it can repeat the calculation or just return the last result.
   @author Christoph Gille
*/
public class StrapProtein extends ProteinNT implements ChRunnable, HasRendererComponent,HasRendererModiCount, HasImage, HasDndFiles  {
    private final static boolean log1=false;
    public final static String METHOD_PARSE="StrapProtein.parse ";
    public final static StrapProtein NONE[]={};
    private final static byte[] TRIPLET=new byte[3];
    /* ---------------------------------------- */
    // >>> Constructor >>> */

    public StrapProtein(ProteinAlignment a) {
        super();
        align_=a;
        vHeteroDna.addHook(UniqueList.HOOK_CHANGE,this,"H_CHANGED");
        vProteinViewers.addHook(UniqueList.HOOK_CHANGE,this,"V3D_CHANGED");

    }
    public StrapProtein() { this(null); }
    /**
       creates Protein from a protein file
       The file type may be PDB, SWISSPROT, EMBL, FASTA
       The method is not thread safe because it uses a static byte buffer for file data.
    */
    public static StrapProtein newInstance(File proteinFile, long options) {
        if (sze(proteinFile)==0) return null;
        final StrapProtein p=new StrapProtein(null);
        p.setName(proteinFile.getName());
        p.setFile(proteinFile);
        p.parse(proteinFile,(ByteArray)null,options);
        StrapAlign a=StrapAlign.getInstance();
        if (a!=null) p.setAlignment(a);
        return p;
    }
    public static StrapProtein newInstance(File proteinFile) { return newInstance(proteinFile,0); }
    public boolean[] getResiduesInAlignmentcolumnrangeAsBoolean(int fromColumn,int toColumn) {
        final int fromIdx=maxi(0,column2nextIndex(fromColumn));
        final int toIdx=column2thisOrPreviousIndex(toColumn);
        final boolean bb[];
        if (fromIdx<toIdx) Arrays.fill(bb=new boolean[toIdx+1],fromIdx,toIdx,true);
        else bb=NO_BOOLEAN;
        return bb;
    }
    /* @Override protected void finalize() throws Throwable { logFinalize("YELLOW StrapProtein",getName());  super.finalize(); } */

    /* <<< Constructor <<< */
    /* ---------------------------------------- */
    /* >>> Alignment Object >>> */
    private ProteinAlignment align_;
    /**  Get The alignment object. */
    @Override public final ProteinAlignment getAlignment() { return align_;}

    public Protein[] getProteinsSameRow() {
        final ProteinAlignment a=getAlignment();
        final StrapView v=a instanceof StrapAlign ? ((StrapAlign)a).getAlignmentPanelWithFocus() : null;
        if (v!=null) {
            final StrapProtein pp[]=v.getProteinsInRow(v.findRow(this));
            if (sze(pp)>0 && cntains(this,pp)) return pp;
        }
        return new StrapProtein[]{this};
    }

    public final void setAlignment(ProteinAlignment a) { align_=a;}
    private boolean isInAlignment=true;
    void setIsInAlignment(boolean b) { isInAlignment=b;}
    boolean isInAlignment() { return isInAlignment;}
    private ValueOfProtein valueOfProtein;
    void setValueOfProtein(ValueOfProtein v) { valueOfProtein=v;}
    ValueOfProtein getValueOfProtein() { return valueOfProtein;}
    /* <<< Alignment Object <<< */
    /* ---------------------------------------- */
    /* >>> Cache >>> */
    private final static int V_GROUPS=60+0, A_SEL=60+1, CACHE_ResCol=60+2, COL_2_i=60+3, g_SEQ=60+4, g_SS=60+5, SEL_AMIN=60+6, SEL_NUC=60+7, g_SOLV=60+8, g_STRG_UC=60+9, g_STRG=60+10, g_EXACT=60+11, RES_COL_W=60+12,
        COL2NXT=60+13, REND_TEXT=60+14, CHILDS=60+15, CHILDS_BUF=60+16;

    /* <<< Cache <<< */
    /* ---------------------------------------- */
    /* >>> Gaps >>> */
    private int _residueGap[], rGap_mc;
    public void setResidueGap(int rg[]) {
        _residueGap=rg;
        rGap_mc++;
        final ProteinAlignment a=getAlignment();
        if (a instanceof StrapAlign) ((StrapAlign)a).setChanged();
    }
    public void setResidueGap(int iA,int gap) {
        final int gaps[]=getResidueGap();
        if (iA>=0 && iA<gaps.length) {
            gaps[iA]=maxi(0,gap);
            setResidueGap(gaps);
        }
    }
    public int addResidueGap(int iA,int gap) {
        final int gg[]=getResidueGap();
        if (iA<0 || gg.length<=iA) return 0;
        final int add=gg[iA]+gap<0 ? -gg[iA] : gap;
        if (add!=0) {
            gg[iA]+=add;
            setResidueGap(gg);
        }
        return add;
    }
    public int[] getResidueGap() {
        final int n=countResidues();
        if (_residueGap==null) _residueGap=new int[n];
        if (_residueGap.length<n || _residueGap.length>n+99) _residueGap=chSze(_residueGap,n);
        return _residueGap;
    }
    public final int getResidueGap(int i) { return i<countResidues() ? get(i,getResidueGap()) : MIN_INT;}
    int sumUpGaps(int ias) {
        if (ias<0) return 0;
        final int gaps[]=getResidueGap(), N=mini(ias,countResidues()-1)+1;
        int g=-1;
        for(int i=N; --i>=0; ) g+=gaps[i];
        return g+N;
    }
    public int wideColumn2nextIndex(int column) {
        final int gaps[]=getResidueGap(), N=mini(gaps.length,countResidues());
        int g=-1;
        for(int i=0; i<N; i++) {
            g+=gaps[i]+1;
            if (g>=column) return i;
        }
        return -1;
    }

    /* <<< Gaps <<< */
    /* ---------------------------------------- */
    /* >>> Columns >>> */
    /**
       Maps residue indices upon horizontal text positions of the alignment.
       @return an integer-array of all horizontal positions for each residue
    */

    public int[] getResidueColumn() {
        final Object[]cached=cached();
        final int nR=countResidues();
        getResidueType();
        final ProteinAlignment align=getAlignment();
        final Gaps2Columns gap2col=align instanceof StrapAlign ? ((StrapAlign)align).gap2col() : null;
        final boolean isWide= (!isInAlignment ||  gap2col==null ||  gap2col.isWide());
        final int wide2narrow[]=gap2col!=null ? gap2col.wide2narrowColumn() : null;
        final int gaps[]=getResidueGap();
        int mc=getModificationCount(ProteinModiCount.MC_RESIDUE_TYPE)+rGap_mc;
        if (gap2col!=null) mc+=gap2col.getModificationCount();

        int cc[]=(int[])cached[CACHE_ResCol];
        if (cc==null || cc.length<nR || cacheEmpty(cached, CACHE_ResCol, mc)) {
            if (log1) debug("!RC"+rGap_mc);
            cached[CACHE_ResCol]=cc=isWide ? getResidueColumnWide() : Gaps2Columns.computeColumns(wide2narrow,gaps,nR,cc);

        }
        return cc!=null ? cc : NO_INT;
    }
    private int[] getResidueColumnWide() {
        final Object[]cached=cached();
        final int nR=countResidues();
        final int resGaps[]=getResidueGap();
        int cc[]=(int[])cached[RES_COL_W];
        if (cc==null || cc.length<nR || cacheEmpty(cached, RES_COL_W,  MODI_COUNTS[MC_RESIDUE_TYPE]+rGap_mc)) {
            cached[RES_COL_W]=cc=redim(cc,nR,33);
            for(int po=-1,ia=0;ia<nR;ia++) cc[ia]= po+=resGaps[ia]+1;
        }
        return cc;
    }
    /**
       Maps residue indices to horizontal text positions of the alignment.
    */
    public final int getResidueColumn(int i){ return i<countResidues() ? get(i,getResidueColumn()) : MIN_INT; }
    /**
       Get the horizontal position of the last residue..
    */
    public int getMaxColumn() { return maxi(-1,getResidueColumn(countResidues()-1));}
    /* <<< Columns <<< */
    /* ---------------------------------------- */
    /*  >>> col to idx >>> */
    /**
       An array to map horizontal residue positions to residue indices. At gap positions the array contains -1.
    */
    public int[] columns2nextIndices() {
        final Object[] cached=cached();
        final int cc[]=getResidueColumn();
        final int nR=countResidues();
        if (nR==0) return NO_INT;
        int[] ii=(int[])cached[COL2NXT];
        if (ii==null || cacheEmpty(cached,COL2NXT,  CACHE_MC[CACHE_ResCol]+MODI_COUNTS[MC_RESIDUE_TYPE])) {
            final int nI=getMaxColumn();
            ii= redim(ii,nI+1,99);
            Arrays.fill(ii,0,nI+1,-1);
            for(int iAa=0;iAa<nR;iAa++) {
                final int c=cc[iAa];
                if (c>=0 && c<ii.length) ii[c]=iAa;
            }
            int last=-1;
            for(int i=nI;i>=0;i--) {
                if (ii[i]<0) ii[i]=last;
                else last=ii[i];
            }
            cached[COL2NXT]=ii;
        }
        return ii;
    }
    /**
       Get  indeces of residues at  horizontal alignment positions col.
       @return The residue index or -1 for an alignment position (column).
    */
    public int[] columns2indices() {
        return (int[])getGappedSequence(cached())[COL_2_i];
    }
    /**
       Get  index of residue at the horizontal alignment position col.
       @param col the alignment position (column)
       @return The residue index or -1 for an alignment position (column).
    */
     public int column2index(int col) {
        final int ii[]=columns2indices();
        return ii!=null && col<ii.length && 0<=col && col<=getMaxColumn() ? ii[col] : -1;
        //return get(columns2indices(),col);
    }
    /**
       Get  index of residue at the horizontal alignment position col.
       if h points to a gap return the previous index.
       @param col  the alignment position (column)
       @return The residue index or -1 for an alignment position (column).
    */
    public int column2thisOrPreviousIndex(int col) {
        final int colMax=getMaxColumn();
        final int ii[]=columns2indices();
        if (col>colMax || col>=ii.length) return countResidues()-1;
        if (col<0) return 0;
        int c=col;  while(c>=0 && ii[c]<0) c--;
        return c<0 ? -1 : ii[c];
    }
    /**
       Get  index of residue at the horizontal alignment position h or the next index if h points to  a gap.
       @param h the alignment position (column)
       @return the index of the residue at alignment position
    */
    public int column2nextIndex(int h) {
        final int ii[]=columns2nextIndices(), maxC=getMaxColumn();
        return h>=0 && h<=maxC && ii.length>h ? ii[h]:-1;
    }
    /* <<< col to idx <<< */
    /* ---------------------------------------- */
    /* >>> gapped sequence >>> */
    /** To be run not in EDT */
    /** The amino acid sequence with gaps (0x20). Array may be longer than getMaxColumn().In this case the end is marked with 0x00.  */
    public byte[] getGappedSequenceTS() { return getGappedSequence();}
    public byte[] getGappedSequence() { return (byte[])getGappedSequence(cached())[g_SEQ];}

    public static byte[][] getGappedSequences(StrapProtein pp[], int colFrom, int colTo, char gap) { return StrapAlign.getGappedSequences(pp,colFrom,colTo,gap);}
    public byte[] getGappedSequences() {
        return (byte[])getGappedSequence(cached())[g_SEQ];
    }

    private Object[] getGappedSequence(Object[]cached) {
        final int n=countResidues();

        final byte aa[]=getResidueType();
        final byte ss[]=getResidueSecStrType();
        final int cc[]=getResidueColumn();
        final ProteinAlignment align=getAlignment();
        final Gaps2Columns gap2col=align instanceof StrapAlign ? ((StrapAlign)align).gap2col() : null;
        final float[] solvAcc=getResidueSolventAccessibility();
        if (cc==null || aa==null || cc.length<n || aa.length<n) return cached;
        final int MC[]=MODI_COUNTS;
        int mc=MC[MC_RESIDUE_TYPE]+MC[MC_RESIDUE_SECSTRU]+CACHE_MC[CACHE_ResCol] + rGap_mc;

        if (gap2col!=null) mc+=gap2col.isWide()?1000:0;
        /*synchronized(this)*/ {
            byte gs[]=(byte[])cached[g_SEQ];
            if (gs==null || setModificationCount(MC_GAPPED_SEQUENCE,mc)) {
                final boolean hasSecStru=sze(ss)>0, hasSolvAcc=sze(solvAcc)>0;
                final int RESERVE=11;
                if (n==0) {
                    cached[g_SEQ]=NO_BYTE;
                    cached[COL_2_i]=NO_INT;
                    if (hasSecStru) cached[g_SS]=NO_BYTE;
                    if (hasSolvAcc) cached[g_SOLV]=NO_FLOAT;
                    return cached;
                }
                final int maxCol=mini(cc[n-1],4*1000*1000), maxCol1=maxCol+1;

                cached[g_SEQ]=gs=redim((byte[]) cached[g_SEQ],maxCol1,RESERVE);
                if (cc[0]>maxCol) { debug("Error in StrapProtein#getGappedSequence "); debug(getName()); stckTrc(); return cached;}
                Arrays.fill(gs,cc[0], maxCol,(byte)'-');
                Arrays.fill(gs,maxCol1,gs.length,(byte)0);
                Arrays.fill(gs,0,cc[0],(byte)' ');

                final byte gss[];
                if (hasSecStru) {
                    gss=redim((byte[])cached[g_SS],maxCol1,RESERVE);
                    if (cc[0]>maxCol) { if (isMyComputer()) putln("Error in getGappedSequence: cc[0]="+cc[0]+" maxCol="+maxCol);}
                    else {
                        Arrays.fill(gss,cc[0],maxCol,(byte)'-');
                        Arrays.fill(gss,maxCol1,gss.length,(byte)0);
                        Arrays.fill(gss,0,cc[0],(byte)' ');
                    }
                } else gss=null;
                cached[g_SS]=gss;
                final float[] gappedSolv=hasSolvAcc ? redim((float[]) cached[g_SOLV],maxCol1,RESERVE) : null;
                cached[g_SOLV]=gappedSolv;
                final int[] c2i=redim((int[])cached[COL_2_i],maxCol1,RESERVE);
                Arrays.fill(c2i,0,maxCol1,-1);

                try {
                    for(int i=0;i<n;i++) {
                        final int col=cc[i];
                        gs[col]=aa[c2i[col]=i];
                    }
                    if (hasSolvAcc||hasSecStru) {
                        for(int i=0;i<n;i++) {
                            final int col=cc[i];
                            if (hasSecStru  && ss.length>i) gss[col]=ss[i];
                            if (hasSolvAcc) gappedSolv[col]=solvAcc[i];
                        }
                    }
                } catch(IndexOutOfBoundsException iobex) {
                    putln(RED_ERROR+" maxCol1="+maxCol1+"\n gss="+sze(gss)+"\n gs="+sze(gs));
                    stckTrc(iobex);
                }
                if (maxCol1<gs.length) gs[maxCol1]=0;

                cached[COL_2_i]=c2i;
                cached[g_STRG_UC]=cached[g_STRG]=cached[g_EXACT]=null;
            }
        }
        return cached;
    }
    /**  The amino acid sequence.  Gaps are character '-'. */
    public byte[] getGappedSequenceExactLength() {
        final Object[]cached=cached();
        final byte gg[]=getGappedSequence();
        if (gg==null) return NO_BYTE;
        final int maxCol=getMaxColumn();
        byte[] newSS=(byte[])cached[g_EXACT];
        if (newSS==null) { System.arraycopy(gg,0,newSS=new byte[maxCol+1],0,maxCol+1); cached[g_EXACT]=newSS; }
        return newSS;
    }
    /**  The amino acid sequence.  Gaps are character '-'. */
    public String getGappedSequenceAsString() {
        final Object[]cached=cached();
        final byte gg[]=getGappedSequenceExactLength();
        final String s=(String)cached[g_STRG];
        return s!=null ? s : (String)(cached[g_STRG]=new String(gg));
    }
    /**  The amino acid sequence.  Gaps are character '-'. */
    public String getGappedSequenceAsStringUC() {
        final Object[]cached=cached();
        final String s=getGappedSequenceAsString();
        final String uc=(String)cached[g_STRG_UC];
        return uc!=null ? uc : (String)(cached[g_STRG_UC]=s.toUpperCase());
    }
    byte[] getGappedSecStru() {
        return (byte[])getGappedSequence(cached())[g_SS];
    }
    float[] getGappedSolventAccessibility() { return (float[])getGappedSequence(cached())[g_SOLV];}
    /* <<< Get gapped Sequence <<< */
    /* ---------------------------------------- */
    /* >>> Set gapped Sequence >>> */
    /**
       The original residue types are kept but the gaps are inferred from the
       gapped sequence seq.
       A letter denotes an amino acid, any other character a gap.
    */
    public void inferGapsFromGappedSequence(byte[] seq) {
        if (seq==null) return;
        final int nR=countResidues(), gaps[]=new int[nR];
        for(int lastPos=-1,iAa=0,pos=0; pos<seq.length && iAa<nR; pos++) {
            final int c=seq[pos]|32;
            if ('a'<=c && c<='z') {
                gaps[iAa++]=pos-lastPos-1;
                lastPos=pos;
            }
        }
        setResidueGap(gaps);
    }
    /** set the gapped sequence */
    public void setGappedSequence(byte[] seq, int len) {
        setResidueType(onlyLetters(seq,0,len));
        inferGapsFromGappedSequence(seq);
    }
    /** See <i>JAVADOC:StrapProtein#setGappedSequence(byte[])</i>*/
    public void setGappedSequence(CharSequence seq) { setGappedSequence(toStrg(seq).getBytes(),MAX_INT);}
   /**
       Reads a line such as X2XXX3X1XXXX2XXXXX
       The numbers are gaps and the letters residue positions
    */
    boolean parseGaps(final byte z[],final int l)  {
        if (z==null) return false;
        {
            final int count=countLetters(z,0,l);
            final int gaps[]=new int[count];
            int po=0, ga=0;
            for(int i=0;i<l;i++) {
                final int c=z[i];
                if (is(LETTR,c)) {
                    gaps[po++]=ga;
                    ga=0;
                    continue;
                }
                if ('0'<=c && c<='9' || c=='-') {
                    int gap=z[i]-'0';
                    while(i+1<l && is(DIGT,z[i+1])) gap=10*gap+z[++i]-'0';
                    ga+=gap;
                }
            }
            setResidueGap(gaps);
        }
        /* may be toLowerCase */
        final int nRes=countResidues();
        final byte aa[]=getResidueType();
        for(int i=0,iAa=0; i<l && iAa<nRes; i++) {
            if (z[i]<='z' && z[i]>='a') aa[iAa++] |=32;
        }
        return true;
    }
    /* <<< Set gapped Sequence <<< */
    /* ---------------------------------------- */
    /* >>> Protein Name >>> */
    private final String[] name2base_ext() {
        final String name=getName(), sfx=ChUtils.getSfx(name,'.'), base=withoutLastComponent(name,'.');
        boolean success=false;
        if (sze(sfx)>0) {
            success=2*sze(sfx)<sze(base);
            for(String s : getCustomSettings(Customize.proteinFileExtensions)) {
                success=success | chrAt(strstr(STRSTR_AFTER,sfx,s),s)==' ';
            }
        }
        return  new String[]{success ? base:name,success?sfx:""};
    }
    private String origName;
    String getOriginalName() { return origName;}
    void setOriginalName(String s) {  origName=s;}
    /* >>> Protein Name >>> */
    /* ---------------------------------------- */
    /* >>> Sort >>> */
    final static int O_WEB_TOK=0, O_WEB_TOKENS=1, O_WEB_TOK1=2, B_PROCESSED=4, B_INTO_SAME_LINE=5, B_REGISTERD=6;
    final Object OBJECTS[]=new Object[7];
    private int tokenIdx=MIN_INT;
    int getWebTokenIdx() {
        final Object t=OBJECTS[O_WEB_TOK], tt[]=(Object[])OBJECTS[O_WEB_TOKENS];
        if (tokenIdx==MIN_INT && t!=null && tt!=null) tokenIdx=idxOf(t,tt);
        return tokenIdx;
    }
    private int preferedOrder;
    int getPreferedOrder() { return preferedOrder;}
    void setPreferedOrder(int t) { preferedOrder=t;}
    private int iRow=-1;
    void setRow(int i) { iRow=i;}
    int getRow() { return iRow;}

    /* <<< Sort <<< */
    /* ---------------------------------------- */
    /* >>> DnD >>> */
    private Map<File, String> mapDragfileModi;
    public File[] getDndFiles() {
        final int dragOptions=StrapAlign.getDragOptions();
        final boolean
            isEntireMSF=(dragOptions&StrapAlign.DRAG_MSA)!=0,
            isAllChains=(dragOptions&StrapAlign.DRAG_ALL_CHAINS)!=0,
            isOrigFile3D=(dragOptions&StrapAlign.DRAG_ORIG)!=0,
            writeSingleFA=isInMsfFile() && !isEntireMSF,
            writeAllPdbChains=getResidueCalpha()!=null && isAllChains && getFileWithAllChains()!=null;
        final int bioMolecule=(dragOptions>>>8)&255;
        getAtomCoordinates();
        final byte[] sequence=getResidueTypeExactLength();
        final int MC[]=this.MODI_COUNTS, thisMC=MC[MC_RESIDUE_TYPE]+MC[MC_ATOM_COORD];
        if (mapDragfileModi==null) mapDragfileModi=new HashMap();
        final String modi_options=thisMC+"_"+dragOptions;
        final String base=name2base_ext()[0], ext=name2base_ext()[1];

        final List<File> v=new ArrayList();
        int countB=0;
        final Matrix3D mmDrag[]=new Matrix3D[getBioMatrices().length];
        for(Matrix3D m:getBioMatrices()) if (m!=null && m.getBiomolecule()==bioMolecule) mmDrag[countB++]=m;
        {
            for(String m : FILE_MASKS) {
                final File fDst=strapFile(dirDndData(),this,m), fSrc=strapFile(dirWorking(),this,m);
                delFile(fDst);
                mkParentDrs(fDst);
                if (m==mICON && countB==0 &&  !writeAllPdbChains) {
                    final String txt0=readStrg(fSrc), txt= sze(txt0)>0 && !looks(LIKE_EXTURL,txt0) ? dirWorking()+"/"+txt0 : null;
                    if (sze(file(txt))>0) wrte(fDst,txt);
                    else if (fSrc.lastModified()>fDst.lastModified()) cpy(fSrc,fDst);
                }
            }

            if (countB==0 &&  !writeAllPdbChains) save(dirDndData(),null);
        }
        if (countB>0) {
            loadSideChainAtoms(null);
            final ByteArray ba=new ByteArray(countResidues()*666);
            for(int iM=mmDrag.length; --iM>=0;) {
                if (mmDrag[iM]==null) continue;
                final File fB=file(dirDndData()+"/"+base+".biomol_"+bioMolecule+(iM<10?"_0"+iM :"_0" )+".pdb");
                if (sze(fB)==0 || !modi_options.equals(mapDragfileModi.get(fB))) {
                    new ProteinWriter1().toText(this,  new Matrix3D[]{mmDrag[iM]}, ProteinWriter.SIDE_CHAIN_ATOMS | ProteinWriter.COMPLETE_PDB, ba.setEnd(0));
                    wrte(fB,ba);
                }
                v.add(fB);
            }
            final File fB=file(dirDndData()+"/"+base+".biomol_"+bioMolecule+"_ALL.pdb");
            if (sze(fB)==0 || !modi_options.equals(mapDragfileModi.get(fB))) {
                new ProteinWriter1().toText(this,null, ProteinWriter.PDB | ProteinWriter.SEQRES | ProteinWriter.HELIX_SHEET, ba.setEnd(0));
                try {
                    final OutputStream fo=fileOutStream(fB);
                    ba.write(fo);
                    for(Matrix3D m3d : mmDrag) {
                        if (m3d==null) continue;
                        new ProteinWriter1().toText(this,new Matrix3D[]{m3d}, ProteinWriter.PDB | ProteinWriter.ATOM_LINES | ProteinWriter.SIDE_CHAIN_ATOMS, ba.setEnd(0));
                        ba.write(fo);
                    }
                    closeStrm(fo);
                    v.add(fB);
                } catch(IOException iox) { stckTrc(iox);}
            }
        } else {
            File fExpr=null;
            StringBuffer sbHeteros=null;
            final String outName=
                isEntireMSF && getFile()!=null ?  getFile().getName() :
                base+(writeAllPdbChains ? "_all":"")+(writeSingleFA ? ".fa":ext);
            final File fOut=file(dirDndData()+"/"+outName);
            if (!modi_options.equals(mapDragfileModi.get(fOut))) delFile(fOut);
            final Matrix3D m3d=getRotationAndTranslation();
            if (writeAllPdbChains) {
                if (isOrigFile3D || (m3d==null || m3d.isUnit()) )  v.add(getFileWithAllChains());
                else {
                    if (sze(fOut)==0) {
                        final long oldOptions=StrapProtein.getParserOptions();
                        StrapProtein.setParserOptions(ProteinParser.SIDE_CHAIN_ATOMS);
                        final StrapProtein p=newInstance(getFileWithAllChains());
                        p.setRotationAndTranslation(m3d);
                        wrte(fOut,ProteinWriter1.mergePdbWithOriginal(p));
                        StrapProtein.setParserOptions(oldOptions);
                    }
                    v.add(fOut);
                }
            } else if (writeSingleFA) {
                if (sze(fOut)==0) {
                    final ByteArray ba=new ByteArray(countResidues()+999).append('>').append(this).append('|');
                    String r=" refs=";
                    for(String s: getSequenceRefs()) {ba.append(r).append(s).append(','); r="";}
                    final int[] firstLast=getFirstAndLastResidueIndex();
                    if (firstLast[0]>0) ba.append(" FIRSTINDEX=").append(firstLast[0]);
                    wrte(fOut,ba.append('\n').append(sequence).append('\n'));
                }
                v.add(fExpr=fOut);
            } else {
                if (!isOrigFile3D && getResidueCalpha()!=null && !(m3d==null || m3d.isUnit())) {
                    if (sze(fOut)==0) wrte(fOut,ProteinWriter1.mergePdbWithOriginal(this));
                    v.add(fExpr=fOut);
                } else {
                    v.add(fExpr=getFile());
                    for(HeteroCompound dna : getDnaAndRnaAndHeteros())  {
                        if (dna.getFile()!=null) {
                            if (sbHeteros==null) sbHeteros=new StringBuffer(99).append('\n');
                            sbHeteros.append(dna.getFile()).append(' ');
                        }
                    }
                }
            }
            if (fExpr!=null) wrte(file(dirDndData()+"/"+fExpr.getName()+".proteinName"), getName()+(sbHeteros!=null ? sbHeteros : ""));
        }
        while(v.remove(null));
        final File ff[]=toArry(v,File.class);
        for(File f:ff) mapDragfileModi.put(f,modi_options);
        return ff;
    }
    /* <<< DnD <<< */
    /* ---------------------------------------- */
    /* >>> Save >>> */
    public void save(File dir,StringBuffer error) {
        if (isTransient) return;
        if (!StrapAlign.isViewer() || dir==dirDndData()) {
            final int gaps[]=getResidueGap();
            if (gaps!=null && (dir==dirDndData() ||  rGap_mc>0)) {
                final byte buffer[]=new byte[12];
                final int n=countResidues();
                final byte[] residueType=getResidueType();
                final File f=strapFile(dir,this,mGAPS);
                OutputStream w=null;
                try {
                    w=fileOutStream(f);
                    for(int i=0;i<n;i++) {
                        final int g=i<gaps.length ? gaps[i] : 0;
                        if (g>0) {
                            itoa(g,buffer);
                            for(int iB=0;iB<buffer.length && buffer[iB]!=0; iB++) w.write(buffer[iB]);
                        }
                        w.write((int)'X');
                    }
                } catch(IOException e) { cnwError(f,error); }
                closeStrm(w);
            }
            if (getNucleotides()!=null) {
                final int n=countNucleotides();
                final File f=strapFile(dir,this,mDNA);
                final boolean[] isTrans=areNucleotidesCurrentStrandTranslated();
                OutputStream w=null;
                try {
                    w=fileOutStream(f);
                    w.write(isReverse()   ? 'R' :' ');
                    w.write(isComplement()? 'C' :' ');
                    for(int i=0;i<n;i++) {
                        w.write(isTrans==null || (i<isTrans.length && isTrans[i]) ? 'X' : 'x');
                    }
                } catch(IOException e) { cnwError(f,error); }
                closeStrm(w);
            }
            {
                final Matrix3D m3d=getRotationAndTranslation();
                final File f=strapFile(dir,this,mTRANS);
                if (m3d==null) delFile(f);
                if (m3d!=null && !wrte(f,m3d.toString())) cnwError(f,error);
            }
            {
                final int idx=getResidueIndexOffset();
                final File f=strapFile(dir, this, Protein.m1stIdx);
                if (idx==0) delFile(f);
                else if (!wrte(f,String.valueOf(idx))) cnwError(f,error);
            }
            {
                final String pdbId=getInferredPdbID();
                final File f=strapFile(dir, this, Protein.mASSOCPDB);
                if (pdbId==null) delFile(f);
                else if (!wrte(f,pdbId)) cnwError(f,error);
            }
        }

        cnwError(ResidueAnnotation.save(this,dir), error);
    }

    static void cnwError(File f,StringBuffer sb) {
        if (sb!=null && f!=null) sb.append("Error: Could not write ").append(f).append('\n');
    }
    public boolean isTransient() { return isTransient;}
    private boolean isTransient;
    /** Non-transient proteins are not saved */
    public void setTransient(boolean b) { isTransient=b;}
    /* <<< Save <<< */
    /* ---------------------------------------- */
    /* >>> Parser >>> */
    /**
       Interpretes the text and extracts the protein data and sets the fields such as residueType in the protein.
       @param parsers An array of parsers to do the job. Try all until one succeeds.
       @param txt the protein text in fasta or pdb or swissprot format.
       @return true on success
    */
    public boolean parse(ByteArray txt, ProteinParser parsers[], long mode) {
        boolean ret=false;
        final long time=System.currentTimeMillis();
        for(ProteinParser parser:parsers) {
            if (parser!=null && parser.parse(this,mode,txt)) {
                setProteinParserClass(parser.getClass());
                setParsingTime(System.currentTimeMillis()-time);
                ret=true;
                break;
            }
        }
        if ( (mode&ProteinParser.SEQUENCE_FEATURES)!=0) ExpasyGff.readFeaturesInProteinFile(txt, 2, this);
        return ret;
    }
     boolean parse(File proteinFile, ByteArray text, long mode) {
        if (proteinFile==null && text==null) return false;
        final ByteArray ba;
        if (text!=null) ba=text;
        else {
            final int fLen=sze(proteinFile);
            ba= fLen>0 ? readAndDecompress(proteinFile,baFromPool(METHOD_PARSE,fLen)) : null;
            if (ba!=null) ba.insidePreTags();
        }
        return sze(ba)>0 ? parse(ba,StrapAlign.getProteinParsers(),mode) : false;
    }
    private static long parserOptions;
    public static long getParserOptions() { return parserOptions;}
    public static void setParserOptions(long options) { parserOptions=options; }
    /* <<< Parser <<< */
    /* ---------------------------------------- */
    private final static Map<Image,Image> mapImageSmall=new WeakHashMap();
    /* >>> Protein Icon >>> */
    /** Get the icon of the protein or null */
    public Image getImage(Component observer) { return getIconImage();}
    private javax.swing.ImageIcon icon;
    public javax.swing.ImageIcon getIcon() {
        final Image im=getIconImage();
        if (im==null) return null;
        if (icon==null) {
            Image imSmall=mapImageSmall.get(im);
            if (imSmall==null)
                mapImageSmall.put(im, imSmall=ChIcon.getScaledInstance(im,ICON_HEIGHT, ICON_HEIGHT, dotSuffixLC(iconFile),null));
            if (imSmall!=null) icon=new javax.swing.ImageIcon(imSmall);
        }
        return icon;
    }
    @Override public Image getIconImage() {
        Image im=iconImage;
        if (im==null) {
            final javax.swing.ImageIcon ic=hasCalpha() ?  iicon(IC_STRUCTURE_BIG) : getNucleotides()!=null ? iicon(IC_DNA) : null;
            if (ic!=null) im=ic.getImage();
        }
        return im;
    }
    //    private int imageHeight;     void setImageHeight(int h) { imageHeight=h;}

    public void setIconImage(String url) {

        if (sze(url)==0) delFile(iconLink()); else wrte(iconLink(), url);
        deleteOldIcons();
        findIconImage();
    }
    private Image iconImage;
    private File iconFile;
    private File dirAnnotations;
    private void findIconImageOld() {
        final File dir=dirAnnotations==null ? dirAnnotations=file(DIR_ANNOTATIONS) : dirAnnotations;
        Image im=null;
        nextType:
        for(String ext : JPG_PNG_GIF_BMP) {
            final File f=new File(dir,getName()+ext);
            if (sze(f)>0) {
                try {
                    im=javax.imageio.ImageIO.read(f);
                    iconImage=im;
                    break nextType;
                } catch(Exception e){}
            }
        }
    }
    private String imageId, iconURL;
    public String getIconUrl() { return iconURL;}
    public String getImageId() { return imageId;}
     void findIconImage() {
        this.icon=null;
        this.iconImage=null;
        this.imageId=this.iconURL=null;
        findIconImageOld();
        final String link=readStrg(iconLink());

        if (sze(link)>0) {
            final String trim=this.iconURL=link.trim();
            this.iconFile=file(trim);
            requestImage(this.imageId=trim,(ChRunnable)this);
        }
    }
     File iconLink() { return file(DIR_ANNOTATIONS+"/"+getNameURLE()+".icon");}
    private void deleteOldIcons() {
        this.icon=null;
        this.iconImage=null;
        for(String ext : JPG_PNG_GIF_BMP) delFile(file(DIR_ANNOTATIONS+"/"+getNameURLE()+ext));
    }

    /* <<< Protein Icon <<< */
    /* ---------------------------------------- */
    /* >>> balloon text >>> */
    private String balloonText, infoPdb;
    private long balloonTextLastModified;
    public void setInfoAssociatedPdb(String info) { infoPdb=info;}
    @Override protected StringBuffer getTooltipWithoutHtmlTags(boolean shift) {
        final StringBuffer sb=strBuf(SB_TIP);
        {
            final File f=getFile();
            if (f!=null) sb.append(filepathUnixStyle(f)).append("<br>");
        }
        final File fIcon=iconFile;
        final String sFile=toStrg(fIcon);
        final Image im=sFile!=null && !sFile.endsWith(".bmp") && !sFile.endsWith(".BMP")  ? iconImage : null;
        if (im!=null) {
            if (shift) return sb.append("<img src=\"").append(fIcon.toURI()).append("\">");
            sb.append("<table><tr><td><img src=\"").append(fIcon.toURI()).append("\" ");
            //if (scaleDown) sb.append(" height=\"70\" width=\"70\"");
            sb.append(">\n</td><td>");
        }
        {
            String h=getHeader(), c=getTitle(), o=getOrganism(), sc=getOrganismScientific();
            if (c==null) c=getCompound();
            if (sze(h)>0)  sb.append(rplcToStrg("|","<br>",htmlEncod(h, null))).append("<br>");
            if (sze(c)>0)  htmlEncod(c,sb).append("<br>");
            if (sze(sc)>0) htmlEncod(sc,sb).append("<br>");
            if (sze(o)>0 && !o.equals(sc)) htmlEncod(o,sb).append("<br>");
        }
        join(getSequenceRefs(), " ", 5, 60,"<br>",sb.append("IDs: ")).append("<br>");
        if (getAccessionID()!=null) sb.append(" Acc=").append(getAccessionID());
        if (getPdbID()!=null)       sb.append(" PDB=").append(getPdbID());
        if (getUniprotID()!=null)   sb.append(" UniProt=").append(getUniprotID());
        final File fTip=strapFile(null,this,mTIPS);
        if (sze(fTip)>0) {
            if (balloonTextLastModified!=fTip.lastModified()) {
                balloonTextLastModified=fTip.lastModified();
                balloonText=readStrg(fTip);
            }
            htmlEncod(balloonText, sb.append("<pre style=\"background-color: white;\">")).append("</pre>");
        }
        sb.append("<br>");
        if (getNucleotides()!=null) sb.append("<br>").append(isReverse() ? " REVERSE":"").append(isComplement() ? " COMPLEMENT":"");
        final String unip[]=charite.christo.protein.FindUniprotID.ids(true,this);
        if (sze(unip)>0) {
            sb.append("<br>Uniprot by sequ identity:");
            for(String s : unip) sb.append(' ').append(delPfx("UNIPROT:",s));
            sb.append("<br>");
        }
        if (!Float.isNaN(getResolutionAnstroms())) sb.append("<br>Resolution ").append(getResolutionAnstroms()).append(" Anstroms<br>");
        if (im!=null) {
            //if (scaleDown) sb.append("<br><font color=\"FF0000\"><sub>shift key for full size image</sub></font>");
            sb.append("</td></tr></table>");
        }
        final int[] firstLast=getFirstAndLastResidueIndex();
        if (firstLast[0]>0) sb.append("<br>").append(firstLast[0]+1).append("...").append(firstLast[1]+1).append("<br");
        if (getDnaAndRnaStructures().length>0) {
            sb.append("<font color=\"#00FF00\" ><b>|</b></font> DNA/RNA compound structures # ").append(getDnaAndRnaStructures().length).append(" &nbsp; ");
            for(HeteroCompound h : getDnaAndRnaStructures()) sb.append(' ').append(h.getNameForPymol());
            sb.append("<br>");
        }
        if (getHeteroCompounds().length>0) {
            sb.append("<font color=\"#FF0000\" ><b>|</b></font>");
            int sbL0=sze(sb);
            sb.append("Hetero compound structures # ").append(getHeteroCompounds().length).append(" &nbsp; ");
            for(HeteroCompound h : getHeteroCompounds()) {
                sb.append(' ').append(h.getNameForPymol());
                if (sze(sb)-sbL0>60) sbL0=sze(sb.append("<br>"));
            }
            sb.append("<br>");
        }
        if (infoPdb!=null && hasCalpha()) sb.append("Associated PDB <pre>").append(infoPdb).append("</pre>");
        /*
        if (isMyComputer()) sb.append("O_WEB_TOK1 ").append(OBJECTS[O_WEB_TOK1]).append("<br>")
                                .append("getCanonicalPath()="+getFilePath()).append("<br>")
                                .append("getFileWithAllChains()=").append(getFileWithAllChains());
        */
        return sb;
    }
    /* <<< Balloon text <<< */
    /* ---------------------------------------- */
    /* >>> Renderer >>> */
    private static ProteinLabel pLabel, pLabelV;

    final long hashCodeShift=hashCode()<<32;
    public long getRendererModiCount() {
        return hashCodeShift+MODI_COUNTS[MC_PROTEIN_LABELS]+MODI_COUNTS[MC_CURSOR_PROTEIN];
    }
    public javax.swing.JComponent getRendererComponent() {
        if (pLabel==null) pLabel=new ProteinLabel(0L,null);
        pLabel.setProtein(this); return pLabel;
    }
    public javax.swing.JComponent getVerticalRendererComponent() {
        if (pLabelV==null) pLabelV=new ProteinLabel(ProteinLabel.VERTICAL,null);
        pLabelV.setProtein(this);
        return pLabelV;
    }
    final String getRendTxt() {
        final Object[] cached=cached();
        getNucleotides();
        String s=(String)fromCache(cached, REND_TEXT,   MODI_COUNTS[MC_INFERRED_PDB]+MODI_COUNTS[MC_NUCLEOTIDE_TYPE]+MODI_COUNTS[MC_PROTEIN_LABELS]);
        if (s==null) {
            final StringBuffer sb=strBuf(SB_REND_COMP).append(this);
            if (getNucleotides()!=null) sb.append(isComplement() && isReverse() ? " RC" : isComplement() ? " C" : isReverse() ? " R" : " F");
            if (getInferredPdbID()!=null) sb.append("  (").append(getInferredPdbID()).append(')');
            cached[REND_TEXT]=s=sb.toString();
        }
        return s;
    }
    private final Rectangle labelSize=new Rectangle();
    private final Object getLabelSize_depend[]=new Object[3];
     final Rectangle getLabelSize(Font font,Rectangle charBounds) {
        final Image image=getIconImage();
        final String name=getName();
        if (getLabelSize_depend[0]!=name || getLabelSize_depend[1]!=image || !font.equals(getLabelSize_depend[2])) {
            getLabelSize_depend[0]=name;
            getLabelSize_depend[1]=image;
            getLabelSize_depend[2]=font;
            final int imHeight=ChIcon.getHeight(image);
            labelSize.x=imHeight>0 ?  ChIcon.getWidth(image)*charBounds.height/imHeight : 0;
            labelSize.width=labelSize.x+charBounds.width+ strgWidth(font,name);
            labelSize.height=charBounds.height;
        }
        return labelSize;
    }
    /* <<< Renderer <<< */
    // --------------------------------------
    /* >>> Residue Selection >>> */
    /**
       Returns an array telling what residues are selected by at least one residue selection.
       0 element means not selected any other number means selected.
    */
    public byte[] getSelectedAminoacids() {
        final Object[]c=cached();
        final int mc=
            +vResidueAnnotation.getModificationCount()
            +vResidueSelection.getModificationCount()
            +vResidueSelectionHidden.getModificationCount()
            +MODI_COUNTS[MC_RESIDUE_SELECTIONS];

        final byte[] cSelA= (byte[]) c[SEL_AMIN], cSelN=(byte[])c[SEL_NUC];
        if (cSelA==null  || cSelN==null  || setModificationCount(MC_SELECTED_AA, mc)) {
            byte[] bbbA=NO_BYTE, bbbN=NO_BYTE;
            for(ResidueSelection s: getAllResidueSelections()) {
                if (ResSelUtils.isShownIn(s,VisibleIn123.SEQUENCE)==null) continue;
                final SelectorOfNucleotides son=s instanceof SelectorOfNucleotides ? (SelectorOfNucleotides)s : null;
                if (son!=null) {
                    final boolean bbN[]=son.getSelectedNucleotides();
                    final int nN=countNucleotides();
                    final int selOffsetN=son.getSelectedNucleotidesOffset();
                    final int lastTrue=lastTrue(bbN);
                    if (lastTrue>=0) {
                        if (bbbN==NO_BYTE) {
                            if (cSelN!=null && nN<=cSelN.length) Arrays.fill(bbbN=cSelN,(byte)0);
                            else bbbN=new byte[nN+9];
                        }
                        final int iTo=mini(lastTrue+1, nN-selOffsetN);
                        for(int i=maxi(0,-selOffsetN); i<iTo; i++) {
                            if (bbN[i] && bbbN[i+selOffsetN]<Byte.MAX_VALUE) bbbN[i+selOffsetN]++;
                        }
                    }
                }
                final int nR=countResidues();
                final boolean bbA[]=ResidueAnnotation.getSelectedAminoacidsDisplayed(s);
                final int selOffsetA=s.getSelectedAminoacidsOffset();
                final int lastTrue=lastTrue(bbA);
                if (lastTrue>=0) {
                    if (bbbA==NO_BYTE) {
                        if (cSelA!=null && nR<=cSelA.length) Arrays.fill(bbbA=cSelA,(byte)0);
                        else bbbA=new byte[nR+9];
                    }
                    final int iTo=mini(lastTrue+1, nR-selOffsetA);
                    for(int i=maxi(0,-selOffsetA); i<iTo; i++) {
                        if (bbA[i] && bbbA[i+selOffsetA]<Byte.MAX_VALUE) bbbA[i+selOffsetA]++;
                    }
                }
            }
            c[SEL_NUC]=bbbN;
            c[SEL_AMIN]=bbbA;
        }
        return (byte[])c[SEL_AMIN];
    }
    public byte[] getSelectedNucleotides() {
        Object[]c=cached();
        getSelectedAminoacids();
        final byte nn[]=(byte[])c[SEL_NUC];
        return nn==null ? NO_BYTE:nn;
    }

    @Override public ResidueSelection[] getAllResidueSelections() {
        final Object[]cached=cached();
        final ResidueSelection[]
            ss1=getResidueSelections(),
            ss2=getResidueAnnotations(),
            ss3=vResidueSelectionHidden.asArray();
        final int countAll=ss1.length+ss2.length+ss3.length;
        if (ss1.length==countAll) return ss1;
        if (ss2.length==countAll) return ss2;
        if (ss3.length==countAll) return ss3;

        final int mc=vResidueAnnotation.getModificationCount()+vResidueSelection.getModificationCount()+vResidueSelectionHidden.getModificationCount();
        ResidueSelection ss[]=(ResidueSelection[])fromCache(cached, A_SEL, mc);
        if (ss==null) {
            ss=new ResidueSelection[countAll];
            System.arraycopy(ss1,0, ss,0                    , ss1.length);
            System.arraycopy(ss2,0, ss,ss1.length           , ss2.length);
            System.arraycopy(ss3,0, ss,ss1.length+ss2.length, ss3.length);
            ss=rmNullA(ss,ResidueSelection.class);
            ResSelUtils.sortResidueSelectionsByVisibility(ss);
            cached[A_SEL]=ss;
        }
        return ss;
    }
    /**  Remove all selection of residues that are instances of c  */
    public void removeResidueSelection(Class c) {
        if (c==null) return;
        for(ResidueSelection s : getAllResidueSelections()) {
            if (isAssignblFrm(c,s))
                removeResidueSelection(s);
        }
    }
    @Override public boolean addResidueSelection(ResidueSelection s) {
        if (s==null) return false;
        final ResidueAnnotation a=s instanceof ResidueAnnotation ? (ResidueAnnotation)s : null;
        final boolean changed, isCursor=s instanceof HasName && ((HasName)s).getName()==ResidueSelection.CURSOR;
        if (a!=null) {
            changed=vResidueAnnotation.add(a);
            if (a.getFeatureName()!=null) Protein.incrementModificationCount(MCA_SEQUENCE_FEATURES_V, this);
            if (changed) Protein.incrementModificationCount(MC_RESIDUE_ANNOTATIONS_V,this);
        } else if (isCursor) {
            changed=vResidueSelectionHidden.add(s);
            Protein.incrementModificationCount(MC_RESIDUE_SELECTIONS_V,this);
        } else changed=super.addResidueSelection(s);
        if (changed) {
            if (s.getProtein()!=this) s.setProtein(this);
            Protein.incrementModificationCount(MC_RESIDUE_SELECTIONS,this);
            if (!isCursor ) updateTree(a!=null ? a : vResidueSelection,null);
        }
        return changed;
    }
    @Override public boolean removeResidueSelection(ResidueSelection s) {
        boolean changed=super.removeResidueSelection(s);
        changed|=vResidueAnnotation.remove(s);
        changed|=vSequenceFeatures.remove(s);
        changed|=vResidueSelectionHidden.remove(s);
        if (changed) {
            StrapAlign.getSelectedObjectsV().remove(s);
            if (s instanceof ResidueAnnotation && ((ResidueAnnotation)s).getFeatureName()!=null) {
                Protein.incrementModificationCount(MCA_SEQUENCE_FEATURES_V, this);
            }
            Protein.incrementModificationCount(MC_RESIDUE_SELECTIONS,this);
        }
        return changed;

    }
    @Override public ByteArray posAsString(int type, int ias) {
        final ByteArray sb=super.posAsString(type,  ias);
        if (getNucleotides()!=null) {
            sb.append(" Nucleotide ").append(aminoAcidIndex2nucleotideIndex(ias)+1);
            final byte trip[]=getResidueTriplet(ias,TRIPLET);
            if (trip!=null){
                sb.append(' ').append((char)trip[0]).append((char)trip[1]).append((char)trip[2]);
            }
        }
        //if (isMyComputer()) sb.append(" sumGaps="+sumUpGaps(ias)+"  ");
        if (type==POS_PRINT_COLUMN) sb.append(" Column ").append(getResidueColumn(ias)+1);
        return sb;
    }
  @Override public ResidueSelection[] getResidueSelectionsAt(long options, int iA_from, int iA_to, int where) {
    final int nRes=countResidues();
    if (iA_from>=iA_to || iA_to<1 || iA_from >=nRes) return ResidueSelection.NONE;
    List<ResidueSelection> v=null;
         for(ResidueSelection s: getAllResidueSelections()) {
           if (s==null) continue;
             if (s instanceof VisibleIn123) {
                 final VisibleIn123 a=(VisibleIn123)s;
                 if ( (a.getVisibleWhere()&where)==0 || a.getStyle()==VisibleIn123.CURSOR) continue;
             }
             final boolean bb[]=
               0!=(options&VisibleIn123.ARROW_HEADS)  ? ResidueAnnotation.getSelectedAminoacidsDisplayed(s) :
               s.getSelectedAminoacids();
             if (firstTrue(bb)<0) continue;
             final int selOffset=s.getSelectedAminoacidsOffset();
             final int to=mini(nRes,bb.length+selOffset, iA_to);
             for (int iA=maxi(selOffset,iA_from);  iA<to; iA++) {
                 if (bb[iA-selOffset]) {
                     (v==null ? v=new UniqueList(ResidueSelection.class) : v).add(s);
                     break;
                 }
             }
         }
         return v!=null ?  toArry(v, ResidueSelection.class) : ResidueSelection.NONE;
     }

    /* <<< Residue Selections <<< /*
    /* ---------------------------------------- */
    /* >>> ResidueAnnotation >>> */
    private final UniqueList<ResidueAnnotation>
        vResidueAnnotation=new UniqueList(ResidueAnnotation.class).t("Residue annotations and sequence features"),
        vResidueAnnotationNF=new UniqueList(ResidueAnnotation.class).t("Residue annotations"),
        vSequenceFeatures=new UniqueList(ResidueAnnotation.class).t("Sequence features").addHook(UniqueList.HOOK_ACCESS,this,"SORT_RA");
    private final UniqueList<ResidueSelection>
        vResidueSelectionHidden=new UniqueList(ResidueSelection.class).t("Hidden");

    public ResidueAnnotation getResidueAnnotationWithName(CharSequence name) {
        if (name==null) return null;
        final String trim=name.toString().trim();
        long created=0;
        ResidueAnnotation a=null;
        for(ResidueAnnotation ra : getResidueAnnotations()) {
            String n=ra.getValue(ResidueAnnotation.NAME);
            if (n.indexOf(trim)>=0 && trim.equals(ra.getValue(ResidueAnnotation.NAME).trim())) {
                final long cr=ra.getWhenCreated();
                if (cr>created) { a=ra; created=cr;}
            }
        }
        return a;
    }
     ChMap<String,UniqueList<ResidueAnnotation>> getResidueAnnotationsByGroup() {
        final Object[] cached=cached();
        final ResidueAnnotation aa[]=getResidueAnnotations();
        final int mc=MODI_COUNTS[MC_RESIDUE_ANNOTATION_GROUPS]+(vResidueAnnotation!=null ? vResidueAnnotation.getModificationCount() : 0);
        ChMap<String,UniqueList<ResidueAnnotation>> mapGroups=(ChMap)cached[V_GROUPS];
        if (mapGroups==null || cacheEmpty(cached, V_GROUPS, mc)) {
            if (mapGroups==null) {
                mapGroups=new ChMap(0L, String.class, UniqueList.class,16).i(IC_ANNO).t("Residue Annotations");
                cached[V_GROUPS]=mapGroups;
            }
            for(UniqueList v : mapGroups.vArray()) if (v!=null) v.clear();
            vSequenceFeatures.clear();
            vResidueAnnotationNF.clear();
            for(ResidueAnnotation a :aa) {
                if (a.getFeatureName()==null) {
                    vResidueAnnotationNF.add(a);
                    final String group=a.getValue(ResidueAnnotation.GROUP);
                    UniqueList<ResidueAnnotation> v=mapGroups.get(group);
                    if (v==null) mapGroups.put(group,v=new UniqueList(ResidueAnnotation.class).t(group).i(iicon(IC_ANNO)));
                    v.add(a);
                } else if (a.isEnabled()) vSequenceFeatures.add(a);
            }
            for(UniqueList v : mapGroups.vArray()) {
                if (v.size()==0) mapGroups.remove(v.getText());
            }
            Collections.sort(vSequenceFeatures, comparator(COMPARE_RENDERER_TEXT));
        }
        return mapGroups;
    }
    /**    @return All annotations of the protein  */
    public ResidueAnnotation[] getResidueAnnotations() {  return vResidueAnnotation.asArray(); }
    public ResidueAnnotation[] getSequenceFeatures() {  return vSequenceFeatures.asArray(); }
    public ResidueAnnotation[] getResidueAnnotationsNF() {  return vResidueAnnotationNF.asArray(); }
    /* <<< ResidueAnnotation <<< /*
    /* ---------------------------------------- */
    /* >>> SequenceFeatures >>> */

    public ReferenceSequence getReferenceSequence(String otherId, ChRunnable log, boolean background, Runnable runWhenFinished ) {
            getResidueType();
            int mc=getModificationCount(ProteinModiCount.MC_RESIDUE_TYPE);
            if (mapReferenceSequence_mc!=mc || mapReferenceSequence==null) {
                mapReferenceSequence_mc=mc;
                mapReferenceSequence=new HashMap();
            }
            ReferenceSequence uniprot=mapReferenceSequence.get(otherId);
            if (uniprot==null) mapReferenceSequence.put(otherId,uniprot=new ReferenceSequence(otherId,this,log, background, runWhenFinished));
            return uniprot.getSuccess() || background ? uniprot : null;

    }
    private Map<String,ReferenceSequence> mapReferenceSequence;
    private int mapReferenceSequence_mc;
    private Set<CharSequence> vInvalidFeature;
     void addInvalidFeatureLine(CharSequence s) { if (vInvalidFeature==null) vInvalidFeature=new HashSet();  vInvalidFeature.add(s);   }
     boolean containsInvalidFeatureLine(CharSequence s) { return s!=null && vInvalidFeature!=null && vInvalidFeature.contains(s);}
    /* <<< SequenceFeatures <<< */
    /* ---------------------------------------- */
    /* >>> Threading >>> */
    private static Object haystacks1[]=new Object[9];
    @Override public Object run(String id, Object arg) {

        if (id==PROVIDE_HIGHLIGHTS) return getName();
        super.run(id,arg);
        final Object argv[]=arg instanceof Object[] ? (Object[])arg : null;
        if (id=="SORT_RA") getResidueAnnotationsByGroup();

        if (id=="H_CHANGED") {
            updateTree(vDna,vHeteroDna);
        }

        if (id=="V3D_CHANGED") updateTree(vProteinViewers,null);
        if (id==ChRunnable.SET_ICON_IMAGE) {
            final Image im=(Image)arg;
            if (this.iconImage!=im) {
                this.iconImage=im;
                icon=null;
                StrapAlign.dispatchLater(StrapEvent.PROTEIN_RENAMED,444);
                Protein.incrementModificationCount(MC_PROTEIN_LABELS,this);
            }
        }
        if (id==DialogStringMatch.PROVIDE_HAYSTACK || id==DialogStringMatch.PROVIDE_STRINGS || id==PROVIDE_WORD_COMPLETION_LIST) {
            final int iRadio=StrapTree.idxRadioSearch();
            if (id==DialogStringMatch.PROVIDE_HAYSTACK) {
                return iRadio==2 ? getResidueAnnotations() : null;
            } else {
                Object hh[]=haystacks1;
                if (iRadio==1) {
                    hh[0]=getOrganism();
                    hh[1]=getCompound();
                    hh[2]=getUniprotID();
                    hh[3]=getPdbID();
                    hh[4]=getPdbRef();
                    hh[5]=getSequenceRefs();
                    hh[6]=getAccessionID();
                    hh[7]=getHeader();
                } else {
                    for(int i=hh.length;--i>=2;) hh[i]=null;
                    hh[0]=getName();
                    hh[1]=getFile();
                }
                return hh;
            }
        }
        if (id==PROVIDE_JTREE_CHILDS) {
            final Object[] cached=cached();
            final ChMap grouped=getResidueAnnotationsByGroup();
            final int mc=vResidueAnnotation.getModificationCount()
                +vResidueSelection.getModificationCount()
                +vProteinViewers.getModificationCount()+
                vHeteroDna.getModificationCount();
            getDnaAndRnaStructures(); /* see getD_R_H */
            Object[] cc=(Object[])fromCache(cached, CHILDS, mc);
            if (cc==null) {
                Object[] buf=(Object[])cached[CHILDS_BUF];
                if (buf==null) cached[CHILDS_BUF]=buf=new Object[6];
                buf[0]=vResidueSelection;
                buf[1]=vProteinViewers;
                buf[2]=grouped.vArray().length>0? grouped : null;
                buf[3]=vSequenceFeatures;
                buf[4]=vDna;
                buf[5]=vHetero;
                for(int i=buf.length; --i>=0;) if (sze(buf[i])==0) buf[i]=null;
                cached[CHILDS]=cc=rmNullA(buf);
            }
            return cc;
        }
        return null;
    }
    /* <<< Threading  <<< */
    /* ---------------------------------------- */
    /* >>> JTree >>> */
    private ChJTree jTree; // JTree
    public ChJTree getJTree() {
        if (jTree==null) {
            jTree=new ChJTree(ChJTable.DEFAULT_RENDERER|ChJTable.DRAG_ENABLED, new ChTreeModel(0L,this));
            jTree.setRootVisible(false);
            jTree.setShowsRootHandles(true);
            jTree.setRowHeight(ICON_HEIGHT);
            StrapAlign.newDropTarget(jTree,false);
            StrapAlign.addListeners(jTree);
            pcp(KEY_DROP_TARGET_REDIRECT,WEAK_REF, jTree);
            ChButton hide=toggl(" ",ChButton.NO_FILL|ChButton.NO_BORDER).s(true);
            ChRenderer.setSmallButtonXc(0,hide);
            final Object msg=vBhB(
                                  hB("Note: Objects may have  a context menu and may support Drag-and-Drop  ","#",hide),
                                  hBL(WATCH_MOVIE+MOVIE_Context_Menu," ",WATCH_MOVIE+MOVIE_Sequence_Features_in_3D));
            hide.actionCollapse(false,msg);
            final javax.swing.JComponent pan=newCNSEW(scrllpn(0L,jTree, dim(222,222)),null,msg,null,null);
            pcp(KEY_PANEL,pan,jTree);
        }
        return jTree;
    }
    public void expandTreeNode(Class c) {
        final ChJTree jT=this.jTree;
        if (jT==null) return;
        for(int pass=2; --pass>=0;) {
            final Object object=
                c==ResidueAnnotation.class ? getResidueAnnotationsByGroup() :
                c==SequenceFeatures.class ? vSequenceFeatures :
                c==ResidueSelection.class ? vResidueSelection :
                c==HeteroCompound.class ? (pass==0 ? vDna : vHetero ) : null;
            if (object !=null) ChJTree.expandAllNodes(object,jT);
        }
    }
    private void updateTree(Object expand1, Object expand2) {
        //putln(this+" updateTree ");
        final ChJTree jT=this.jTree;
        final Window w=jT==null ? null : parentWndw(jT);
        if (w instanceof Frame && ((Frame)w).getState()==Frame.NORMAL) {
            Object ex1=expand1, ex2=expand2;
            if (ex1 instanceof ResidueAnnotation) {
                final ResidueAnnotation a=(ResidueAnnotation)ex1;
                ex1=getResidueAnnotationsByGroup();
                ex2=getResidueAnnotationsByGroup().get(a.getValue(ResidueAnnotation.GROUP));
            }
            if (ex1!=null || ex2!=null) pcp(ChJTree.KEY_EXPAND, new Object[]{ex1,ex2},jT);
            jT.updateKeepStateLater(0L, 99);
        }
    }
    /* <<< JTree <<< */
    /* ---------------------------------------- */
 /* >>> Static Utilities >>> */
    public static StrapProtein[] loadProteinsInList(File f,long options) { return loadProteinsInList(readBytes(f),options);}
    public static StrapProtein[] loadProteinsInList(ByteArray ba,long options) {
        if (ba==null) return NONE;
        final ChTokenizer T=new ChTokenizer().setText(ba);
        UniqueList<StrapProtein>v=null;
        while(T.nextToken()) {
            final File f=file(T.asString());
            if (sze(f)==0) continue;
            final StrapProtein p=StrapProtein.newInstance(f,options);
            if (p.countResidues()>0) (v==null?v=new UniqueList(StrapProtein.class):v).add(p);
        }
        return v==null? NONE : v.asArray();
    }
    public final static StrapProtein[] getProteinsWithNames(ByteArray ba, StrapProtein pp[]) {
        if (ba==null) return NONE;
        final ChTokenizer T=new ChTokenizer().setText(ba);
        UniqueList<StrapProtein>v=null;
        while(T.nextToken()) {
            final StrapProtein p=getProteinWithName(T.asString(),pp);
            if (p!=null) (v==null?v=new UniqueList(StrapProtein.class):v).add(p);
        }
        return v==null? NONE : v.asArray();
    }
    /**  returns the 1st protein whose name equals to name.*/
    public static StrapProtein getProteinWithName(String name,Protein pp0[]) {
        final StrapAlign a=StrapAlign.getInstance();
        final Protein pp[]=pp0!=null ? pp0 : a!=null ?  a.getProteins() : null;
        if (name!=null && pp!=null) {
            final int hcName=name.hashCode();
            for(Protein p:pp) {
                if  (p==null) continue;
                final Object wt=((StrapProtein)p).OBJECTS[O_WEB_TOK1];
                if (hcName==p.getNameHC() && name.equals(p.getName()) || wt!=null && wt.equals(name)) return (StrapProtein)p;
            }
        }
        return null;
    }

    public static StrapProtein getProteinWithNameAndFile(String n, File file, Object pp[]) {
        final String fn=file!=null ? file.getName() : null;
        //String canonicalPath=null;        try { canonicalPath=file.getCanonicalPath();} catch(Exception ex){stckTrc(ex);}

        if (pp!=null && fn!=null) {
            final int fnHC=fn.hashCode(), nHC=n!=null?n.hashCode():0;
            for(Object o:pp) {
                final StrapProtein p= o instanceof StrapProtein ? (StrapProtein)o:null;
                if (p==null || n!=null && !(nHC==p.getNameHC() && n.equals(p.getName()))) continue;
                final File f=p.getFile();
                if (f!=null && fnHC==p.getFileNameHC()  && f.getName()==p.getFileName()/* && (canonicalPath==null||canonicalPath.equals(p.getFilePath()))*/) return p;
            }
        }
        return null;
    }
    public static int getMaxColumn(StrapProtein pp[]) {
        int max=0;
        for(StrapProtein p : pp) if (p!=null) max=maxi(max,p.getMaxColumn());
        return max;
    }
    public static StrapProtein[] getProteinsOfObjects(Object oo[]) {
        ArrayList<StrapProtein> vP=null;
        if (sze(oo)>0) {
            for(Object o : oo) {
                final StrapProtein p=(StrapProtein)(o instanceof StrapProtein ? o : o instanceof ResidueSelection ? ((ResidueSelection)o).getProtein() : null);
                if (p!=null) adUniq(p, vP==null?vP=new ArrayList() : vP);
            }
        }
        return vP==null ? StrapProtein.NONE : toArry(vP,StrapProtein.class);
    }

 /* <<< Static Utilities <<< */

}
